/*
 Niels Elken Snderby 2003
 NesQuant - http://www.nielses.dk/quantlib/nesquant
 NesQuant is an extension to QuantLib - http://quantlib.org/
*/

/*! \file FlatPlainOption.cpp
    \brief Flat plain (no dividends, no barriers) option on a single asset

    \fullpath
    nq/Instruments/%FlatPlainOption.cpp
*/

// $Id: FlatPlainOption.cpp,v 1.11 2002/03/05 16:58:02 lballabio Exp $

#include "FlatPlainOption.hpp"
#include <ql/Solvers1D/brent.hpp>

namespace NesQuant {

    namespace Instruments {

        FlatPlainOption::FlatPlainOption(Option::Type type,
                    double underlying,
                    double strike,
                    Spread dividendYield,
                    Rate riskFreeRate,
                    Time residualTime,
                    double volatility,
                    const std::string& isinCode,
                    const std::string& description)
        : Option(Handle<EuropeanEngine>(new EuropeanEngine()) , isinCode, description), type_(type),
          underlying_(underlying), strike_(strike),
          dividendYield_(dividendYield), riskFreeRate_(riskFreeRate),
          residualTime_(residualTime), volatility_(volatility) {
        }

        double FlatPlainOption::delta() const {
            calculate();
            QL_REQUIRE(delta_ != Null<double>(),
                       "delta calculation failed");
            return delta_;
        }

        double FlatPlainOption::gamma() const {
            calculate();
            QL_REQUIRE(gamma_ != Null<double>(),
                       "gamma calculation failed");
            return gamma_;
        }

        double FlatPlainOption::theta() const {
            calculate();
            QL_REQUIRE(theta_ != Null<double>(),
                       "theta calculation failed");
            return theta_;
        }

        double FlatPlainOption::vega() const {
            calculate();
            QL_REQUIRE(vega_ != Null<double>(),
                       "vega calculation failed");
            return vega_;
        }

        double FlatPlainOption::rho() const {
            calculate();
            QL_REQUIRE(rho_ != Null<double>(),
                       "rho calculation failed");
            return rho_;
        }

        double FlatPlainOption::dividendRho() const {
            calculate();
            QL_REQUIRE(dividendRho_ != Null<double>(),
                       "dividend rho calculation failed");
            return dividendRho_;
        }

        double FlatPlainOption::impliedVolatility(double targetValue,
          double accuracy, Size maxEvaluations,
          double minVol, double maxVol) const {
            double value = NPV(), vol = volatility_;
            QL_REQUIRE(!isExpired_, "option expired");
            if (value == targetValue) {
                return vol;
            } else {
                ImpliedVolHelper f(engine_,targetValue);
                Solvers1D::Brent solver;
                solver.setMaxEvaluations(maxEvaluations);
                return solver.solve(f,accuracy,vol,minVol,maxVol);
            }
        }

        void FlatPlainOption::setupEngine() const {
            PlainOptionParameters* parameters =
                dynamic_cast<PlainOptionParameters*>(
                    engine_->parameters());
            QL_REQUIRE(parameters != 0,
                       "pricing engine does not supply needed parameters");

            parameters->type = type_;

            QL_REQUIRE(!(underlying_ == Null<double>()), "null underlying price given");
            parameters->underlying = underlying_;

            parameters->strike = strike_;

            if (dividendYield_ == Null<double>())
                parameters->dividendYield = 0.0;
            else
                parameters->dividendYield = dividendYield_;

            QL_REQUIRE(!(riskFreeRate_ == Null<double>()), "null risk free rate given");
            parameters->riskFreeRate = riskFreeRate_;

            parameters->residualTime = residualTime_;

            QL_REQUIRE(!(volatility_ == Null<double>()), "null volatility given");
            parameters->volatility = volatility_;
        }

        void FlatPlainOption::performCalculations() const {
            if (residualTime_ <= 0) {
                isExpired_ = true;
                NPV_ = delta_ = gamma_ = theta_ =
                    vega_ = rho_ = dividendRho_ = 0.0;
            } else {
                isExpired_ = false;
                Option::performCalculations();
                const OptionGreeks* results =
                    dynamic_cast<const OptionGreeks*>(engine_->results());
                QL_ENSURE(results != 0,
                          "no greeks returned from option pricer");
                /* no check on null values - just copy.
                   this allows:
                   a) to decide in derived options what to do when null
                      results are returned (throw? numerical calculation?)
                   b) to implement slim engines which only calculate the
                      value---of course care must be taken not to call
                      the greeks methods when using these.
                */
                delta_       = results->delta;
                gamma_       = results->gamma;
                theta_       = results->theta;
                vega_        = results->vega;
                rho_         = results->rho;
                dividendRho_ = results->dividendRho;
            }
            QL_ENSURE(isExpired_ || NPV_ != Null<double>(),
                      "null value returned from option pricer");
        }


        FlatPlainOption::ImpliedVolHelper::ImpliedVolHelper(
            const Handle<OptionPricingEngine>& engine, double targetValue)
        : engine_(engine), targetValue_(targetValue) {
            parameters_ = dynamic_cast<PlainOptionParameters*>(
                engine_->parameters());
            QL_REQUIRE(parameters_ != 0,
                       "pricing engine does not supply needed parameters");
            results_ = dynamic_cast<const OptionValue*>(
                engine_->results());
            QL_REQUIRE(results_ != 0,
                       "pricing engine does not supply needed results");
        }

        double FlatPlainOption::ImpliedVolHelper::operator()(double x) const {
            parameters_->volatility = x;
            engine_->calculate();
            return results_->value-targetValue_;
        }

    }

}

