/*
 Niels Elken Snderby 2003
 NesQuant - http://www.nielses.dk/quantlib/nesquant
 NesQuant is an extension to QuantLib - http://quantlib.org/
*/

/*! \file FlatPlainOption.hpp
    \brief Flat plain (no dividends, no barriers) option on a single asset

    \fullpath
    nq/Instruments/%flatFlatPlainOption.hpp
*/

#ifndef nesquant_flatplain_option_h
#define nesquant_flatplain_option_h

#include <ql/option.hpp>
#include <ql/Instruments/plainoption.hpp>
#include <ql/Pricers/europeanengine.hpp>
#include <ql/termstructure.hpp>
#include <ql/solver1d.hpp>

using namespace QuantLib;
using QuantLib::Option;
using QuantLib::Instruments::PlainOptionParameters;
using QuantLib::Pricers::EuropeanEngine;

namespace NesQuant {

    namespace Instruments {

    class FlatPlainOptionParameters;

        //! Plain (no dividends, no barriers) option on a single asset
        class FlatPlainOption : public Option {
          public:
            FlatPlainOption(Option::Type type,
                        double underlying,
                        double strike,
                        Spread dividendYield,
                        Rate riskFreeRate,
                        Time residualTime,
                        double volatility,
                        const std::string& isinCode = "",
                        const std::string& description = "");
            //! \name greeks
            //@{
            double delta() const;
            double gamma() const;
            double theta() const;
            double vega() const;
            double rho() const;
            double dividendRho() const;
            //@}
            /*! \warning Options with a gamma that changes sign have values
                that are <b>not</b> monotonic in the volatility, e.g binary
                options. In these cases impliedVolatility can fail and in
                any case it is almost meaningless.
                Another possible source of failure is to have a
                targetValue that is not attainable with any volatility, e.g.
                a targetValue lower than the intrinsic value in the case of
                American options. */
            double impliedVolatility(double targetValue,
                                     double accuracy = 1.0e-4,
                                     Size maxEvaluations = 100,
                                     double minVol = 1.0e-4,
                                     double maxVol = 4.0) const;
          protected:
            void setupEngine() const;
            void performCalculations() const;
          private:
            // parameters
            Option::Type type_;
            double underlying_;
            double strike_;
            Spread dividendYield_;
            Rate riskFreeRate_;
            Time residualTime_;
            double volatility_;
            // results
            mutable double delta_, gamma_, theta_, vega_, rho_, dividendRho_;
            // helper class for implied volatility calculation
            class ImpliedVolHelper : public ObjectiveFunction {
              public:
                ImpliedVolHelper(const Handle<OptionPricingEngine>& engine,
                                 double targetValue);
                double operator()(double x) const;
              private:
                Handle<OptionPricingEngine> engine_;
                double targetValue_;
                PlainOptionParameters* parameters_;
                const OptionValue* results_;
            };
        };

    }

}


#endif

