(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    250120,       6602]*)
(*NotebookOutlinePosition[    250801,       6626]*)
(*  CellTagsIndexPosition[    250757,       6622]*)
(*WindowFrame->Normal*)



Notebook[{
Cell[BoxData[
    StyleBox[\(QuantLib\ for\ Mathematica\ demonstration\),
      "Title"]], "Input"],

Cell[TextData[{
  "Niels Elken S\[OSlash]nderby, May 2003",
  ButtonBox["\nhttp://www.nielses.dk/quantlib/mma",
    ButtonData:>{
      URL[ "http://www.nielses.dk/quantlib/mma"], None},
    ButtonStyle->"Hyperlink"],
  "\n",
  ButtonBox["http://quantlib.org/",
    ButtonData:>{
      URL[ "http://quantlib.org/"], None},
    ButtonStyle->"Hyperlink"]
}], "Text"],

Cell[CellGroupData[{

Cell["Installation", "Section"],

Cell["\<\
At this point only binaries for Microsoft Windows are available. It should be \
quite easy to make QuantLib for Mathematica work on other platforms, but it \
requires that you know how to compile the C++ files (see Readme.txt for \
details).\
\>", "Text"],

Cell[TextData[{
  "To install QuantLib for Mathematica, please extract all the files from the \
distribution ZIP-file (",
  ButtonBox["download",
    ButtonData:>{
      URL[ "http://www.nielses/quantlib/mma/download"], None},
    ButtonStyle->"Hyperlink"],
  ") and copy the folder QuantLib into your Mathematica application \
directory. In version 4 on a Microsoft Windows system this is most often \
c:\\Program Files\\Wolfram Research\\Mathematica\\4.0\\AddOns\\Applications"
}], "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Loading the package", "Section"],

Cell[TextData[{
  "After installing QuantLib for ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " on your system, you can load the package and start the external program \
by writing:"
}], "Text"],

Cell[BoxData[
    \(<< QuantLib`\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?QuantLib\)\)], "Input"],

Cell[BoxData[
    \("QuantLib gives access to various financial functions from the QuantLib \
C++ library. More info: http://www.nielses.dk/quantlib/mma"\)], "Print"]
}, Open  ]],

Cell["Now you have access to a number of new financial functions.", "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Design principles and the naming of things", "Section"],

Cell[TextData[{
  "One of the principles applied in the development of QuantLib for ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " is that it should give ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " users easy access to the functionality in QuantLib rather than giving \
QuantLib users and developers a front-end for QuantLib. This means that the \
names of the function are following the ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " naming conventions rather than reflecting the QuantLib class hierarchy \
names. I.e. BusinessDayQ instead of isBusinessDay, MonteCarloPath instead of \
PathGenerator.next()."
}], "Text"],

Cell[TextData[{
  "It is the goal to make QuantLib for ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " have the look and feel of a standard ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " package, and still expose all the QuantLib functionality with things like \
instruments, term structures, calendars and finite difference pricing engines \
in a powerful way. The decision to make the things \"the ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " way\" also means that in principle another back-end than QuantLib could \
be used. The most obvious example would be to code it all in ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " itself, but other C++ toolkits could be used as well."
}], "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Valuing options", "Section"],

Cell["\<\
Let's try to price a European option. First we define the option.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?EuropeanOption\)\)], "Input"],

Cell[BoxData[
    \("EuropeanOption[strike, residualtime, type] represents a european \
option of the given type (CallOption or PutOption), exercise price (strike) \
and time to maturity (residualtime)."\)], "Print"]
}, Open  ]],

Cell[BoxData[
    \(\(instr = EuropeanOption[110, 0.25, CallOption];\)\)], "Input"],

Cell["Then we define the model under which it should be priced:", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?BlackScholesModel\)\)], "Input"],

Cell[BoxData[
    \("BlackScholesModel[stockprice, riskfreerate, dividend, vol] represents \
a model where the underlying asset is worth stockprice at time zero and the \
risk free rate, dividend yield and volatility are as given."\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(model = BlackScholesModel[100, 0.10, 0.01, 0.15]\)], "Input"],

Cell[BoxData[
    \(BlackScholesModel[100, 0.1`, 0.01`, 0.15`]\)], "Output"]
}, Open  ]],

Cell["Now we can calculate keyfigures for this instrument.", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Keyfigures[instr, model]\)], "Input"],

Cell[BoxData[
    \({Value \[Rule] 0.6828593225781923`, Delta \[Rule] 0.17489412560177403`, 
      Gamma \[Rule] 0.03432544935471783`, 
      Theta \[Rule] \(-5.367374250563904`\), 
      Vega \[Rule] 12.872043508019189`, Rho \[Rule] 4.201638309399803`, 
      DividendRho \[Rule] \(-4.372353140044351`\)}\)], "Output"]
}, Open  ]],

Cell["A short cut for having the value is:", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Value[instr, model]\)], "Input"],

Cell[BoxData[
    \(0.6828593225781923`\)], "Output"]
}, Open  ]],

Cell[TextData[{
  "For the other keyfigures we must use the ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " syntax \"/.\" (ReplaceAll):"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Delta\  /. \ Keyfigures[instr, model]\)], "Input"],

Cell[BoxData[
    \(0.17489412560177403`\)], "Output"]
}, Open  ]],

Cell["\<\
We could now make a function, which gives the price given a certain spot \
price:\
\>", "Text"],

Cell[BoxData[
    \(optval[s_] := 
      Value[instr, BlackScholesModel[s, 0.10, 0.01, 0.15]]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Plot[optval[s], {s, 80, 120}]\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-1.88095 0.0238095 0.0147151 0.0463879 [
[.2619 .00222 -6 -9 ]
[.2619 .00222 6 0 ]
[.5 .00222 -9 -9 ]
[.5 .00222 9 0 ]
[.7381 .00222 -9 -9 ]
[.7381 .00222 9 0 ]
[.97619 .00222 -9 -9 ]
[.97619 .00222 9 0 ]
[.01131 .10749 -6 -4.5 ]
[.01131 .10749 0 4.5 ]
[.01131 .20027 -6 -4.5 ]
[.01131 .20027 0 4.5 ]
[.01131 .29304 -6 -4.5 ]
[.01131 .29304 0 4.5 ]
[.01131 .38582 -6 -4.5 ]
[.01131 .38582 0 4.5 ]
[.01131 .47859 -12 -4.5 ]
[.01131 .47859 0 4.5 ]
[.01131 .57137 -12 -4.5 ]
[.01131 .57137 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.2619 .01472 m
.2619 .02097 L
s
[(90)] .2619 .00222 0 1 Mshowa
.5 .01472 m
.5 .02097 L
s
[(100)] .5 .00222 0 1 Mshowa
.7381 .01472 m
.7381 .02097 L
s
[(110)] .7381 .00222 0 1 Mshowa
.97619 .01472 m
.97619 .02097 L
s
[(120)] .97619 .00222 0 1 Mshowa
.125 Mabswid
.07143 .01472 m
.07143 .01847 L
s
.11905 .01472 m
.11905 .01847 L
s
.16667 .01472 m
.16667 .01847 L
s
.21429 .01472 m
.21429 .01847 L
s
.30952 .01472 m
.30952 .01847 L
s
.35714 .01472 m
.35714 .01847 L
s
.40476 .01472 m
.40476 .01847 L
s
.45238 .01472 m
.45238 .01847 L
s
.54762 .01472 m
.54762 .01847 L
s
.59524 .01472 m
.59524 .01847 L
s
.64286 .01472 m
.64286 .01847 L
s
.69048 .01472 m
.69048 .01847 L
s
.78571 .01472 m
.78571 .01847 L
s
.83333 .01472 m
.83333 .01847 L
s
.88095 .01472 m
.88095 .01847 L
s
.92857 .01472 m
.92857 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.02381 .10749 m
.03006 .10749 L
s
[(2)] .01131 .10749 1 0 Mshowa
.02381 .20027 m
.03006 .20027 L
s
[(4)] .01131 .20027 1 0 Mshowa
.02381 .29304 m
.03006 .29304 L
s
[(6)] .01131 .29304 1 0 Mshowa
.02381 .38582 m
.03006 .38582 L
s
[(8)] .01131 .38582 1 0 Mshowa
.02381 .47859 m
.03006 .47859 L
s
[(10)] .01131 .47859 1 0 Mshowa
.02381 .57137 m
.03006 .57137 L
s
[(12)] .01131 .57137 1 0 Mshowa
.125 Mabswid
.02381 .03791 m
.02756 .03791 L
s
.02381 .0611 m
.02756 .0611 L
s
.02381 .0843 m
.02756 .0843 L
s
.02381 .13068 m
.02756 .13068 L
s
.02381 .15388 m
.02756 .15388 L
s
.02381 .17707 m
.02756 .17707 L
s
.02381 .22346 m
.02756 .22346 L
s
.02381 .24665 m
.02756 .24665 L
s
.02381 .26985 m
.02756 .26985 L
s
.02381 .31624 m
.02756 .31624 L
s
.02381 .33943 m
.02756 .33943 L
s
.02381 .36262 m
.02756 .36262 L
s
.02381 .40901 m
.02756 .40901 L
s
.02381 .43221 m
.02756 .43221 L
s
.02381 .4554 m
.02756 .4554 L
s
.02381 .50179 m
.02756 .50179 L
s
.02381 .52498 m
.02756 .52498 L
s
.02381 .54818 m
.02756 .54818 L
s
.02381 .59456 m
.02756 .59456 L
s
.02381 .61776 m
.02756 .61776 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .01472 m
.03279 .01472 L
.04262 .01472 L
.05288 .01472 L
.06244 .01472 L
.0731 .01473 L
.0828 .01473 L
.094 .01474 L
.10458 .01474 L
.11428 .01475 L
.12455 .01476 L
.14299 .01479 L
.1524 .01481 L
.16253 .01484 L
.18083 .0149 L
.19093 .01495 L
.20196 .01501 L
.2125 .01509 L
.22205 .01517 L
.24173 .01539 L
.25253 .01555 L
.2627 .01573 L
.28278 .01618 L
.30125 .01675 L
.32007 .01751 L
.34074 .01862 L
.36025 .01999 L
.38116 .02189 L
.4024 .02438 L
.42252 .02735 L
.46085 .03507 L
.48063 .0403 L
.50163 .04695 L
.53938 .06212 L
.57959 .08347 L
.61828 .10966 L
.65942 .144 L
.69904 .18356 L
.73715 .22746 L
.77771 .28013 L
.81675 .33605 L
.85825 .40036 L
.89823 .46631 L
.9367 .53273 L
.97619 .60332 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/A000`40O003h00Oogoo8Goo001<Ool3
0004Ool2000eOol50002Ool20004Ool2000bOol500000goo000000030002Ool2000bOol500000goo
000000020003Ool20005Ool004ioo`05001oogooOol00002Ool00`00Oomoo`0dOol01@00Oomoogoo
00000Woo00@007ooOol0009oo`03001oogoo035oo`03001oogoo00=oo`05001oogooOol00002Ool0
0`00Oomoo`0aOol01@00Oomoogoo00001Goo00@007ooOol000Aoo`00CGoo0`000Woo00@007ooOol0
03Ioo`05001oogooOol00002Ool01000Oomoo`000Woo00<007ooOol0<Goo00<007ooOol00goo00D0
07ooOomoo`00009oo`03001oogoo035oo`03001oogoo009oo`03001oogoo009oo`04001oogoo0004
Ool004aoo`04001oogoo0002Ool01000Oomoo`00=Woo00D007ooOomoo`00009oo`04001oogoo0002
Ool00`00Oomoo`0aOol00`00Oomoo`03Ool01@00Oomoogoo00000Woo00<007ooOol0<Goo00<007oo
Ool00goo00D007ooOomoo`00009oo`03001oogoo009oo`00C7oo00@007ooOol0009oo`04001oogoo
000fOol01@00Oomoogoo00000Woo00@007ooOol0009oo`03001oogoo035oo`03001oogoo00=oo`05
001oogooOol00002Ool00`00Oomoo`0aOol01@00Oomoogoo00000Woo00@007ooOol0009oo`03001o
ogoo009oo`00CGoo0`000goo0P00=Woo0P0017oo0P0017oo0P00<goo0P0017oo0P0017oo0P00<goo
0P0017oo0P0017oo0P001Goo003oOolQOol0011oo`03001oogoo0?moo`ioo`0047oo00<007ooOol0
ogoo3Woo000@Ool00`00Oomoo`3oOol>Ool0011oo`03001oogoo0?moo`ioo`002Wooo`004`0017oo
000@Ool00`00Oomoo`0:Ool00`00Oomoo`0:Ool00`00Oomoo`0:Ool00`00Oomoo`0:Ool00`00Oomo
o`0:Ool00`00Oomoo`06OolE0009Ool00`00Oomoo`0:Ool00`00Oomoo`0:Ool00`00Oomoo`0:Ool0
0`00Oomoo`0:Ool00`00Oomoo`0:Ool00`00Oomoo`0:Ool00`00Oomoo`0:Ool00`00Oomoo`0:Ool0
0`00Oomoo`0:Ool00`00Oomoo`0:Ool00`00Oomoo`0:Ool00`00Oomoo`0:Ool00`00Oomoo`09Ool0
011oo`03001oogoo03ioo`03001oogoo01]oo`H001eoo`03001oogoo03ioo`03001oogoo03ioo`03
001oogoo00Uoo`0047oo00<007ooOol0HWoo2P00XGoo000@Ool00`00Oomoo`1/Ool4002MOol0011o
o`03001oogoo071oo`@009Uoo`0047oo0P00MGoo1000UGoo000@Ool00`00Oomoo`1hOol4002AOol0
011oo`03001oogoo07aoo`<008ioo`0047oo00<007ooOol0Ogoo0`00Rgoo000@Ool00`00Oomoo`22
Ool20029Ool0011oo`03001oogoo08Aoo`<008Ioo`0047oo00<007ooOol0Qgoo0P00Q7oo000@Ool2
002:Ool20022Ool0011oo`03001oogoo08]oo`80081oo`0047oo00<007ooOol0SGoo0P00OWoo000@
Ool00`00Oomoo`2?Ool2001lOol0011oo`03001oogoo095oo`8007Yoo`0047oo00<007ooOol0Tgoo
0P00N7oo000@Ool2002FOol00`00Oomoo`1eOol0011oo`03001oogoo09Ioo`8007Eoo`0047oo00<0
07ooOol0V7oo00<007ooOol0LWoo000@Ool00`00Oomoo`2IOol00`00Oomoo`1aOol000Qoo`@000Ao
o`03001oogoo09Yoo`80075oo`0027oo00<007ooOol01Goo00<007ooOol0W7oo00<007ooOol0KWoo
0009Ool00`00Oomoo`04Ool3002MOol2001^Ool000Yoo`03001oogoo00=oo`03001oogoo09moo`03
001oogoo06]oo`0027oo00@007ooOol000Aoo`03001oogoo0:1oo`03001oogoo06Yoo`002Goo0P00
1Goo00<007ooOol0XGoo0P00JWoo000@Ool00`00Oomoo`2SOol00`00Oomoo`1WOol0011oo`03001o
ogoo0:Aoo`03001oogoo06Ioo`0047oo00<007ooOol0YGoo0P00IWoo000@Ool2002XOol00`00Oomo
o`1SOol0011oo`03001oogoo0:Qoo`03001oogoo069oo`0047oo00<007ooOol0ZGoo0P00HWoo000@
Ool00`00Oomoo`2[Ool00`00Oomoo`1OOol0011oo`03001oogoo0:aoo`03001oogoo05ioo`0047oo
00<007ooOol0[Goo00<007ooOol0GGoo000@Ool2002_Ool00`00Oomoo`1LOol0011oo`03001oogoo
0:moo`03001oogoo05]oo`0047oo00<007ooOol0/7oo00<007ooOol0FWoo000@Ool00`00Oomoo`2`
Ool00`00Oomoo`1JOol0011oo`03001oogoo0;5oo`03001oogoo05Uoo`0047oo00<007ooOol0/Woo
00<007ooOol0F7oo000@Ool2002dOol00`00Oomoo`1GOol0011oo`03001oogoo0;Aoo`03001oogoo
05Ioo`0047oo00<007ooOol0]Goo00<007ooOol0EGoo000@Ool00`00Oomoo`2fOol00`00Oomoo`1D
Ool0011oo`03001oogoo0;Moo`03001oogoo05=oo`002Woo0`000goo00<007ooOol0^7oo00<007oo
Ool0DWoo000;Ool00`00Oomoo`02Ool00`00Oomoo`2iOol00`00Oomoo`1AOol000Qoo`D000=oo`<0
0;Yoo`03001oogoo051oo`0027oo00@007ooOol000Aoo`03001oogoo0;Yoo`03001oogoo051oo`00
2Goo00<007oo000017oo00<007ooOol0^goo00<007ooOol0Cgoo000:Ool20004Ool00`00Oomoo`2l
Ool00`00Oomoo`1>Ool0011oo`03001oogoo0;eoo`03001oogoo04eoo`0047oo00<007ooOol0_Woo
00<007ooOol0C7oo000@Ool20030Ool00`00Oomoo`1;Ool0011oo`03001oogoo0<1oo`03001oogoo
04Yoo`0047oo00<007ooOol0`Goo00<007ooOol0BGoo000@Ool00`00Oomoo`32Ool00`00Oomoo`18
Ool0011oo`03001oogoo0<9oo`03001oogoo04Qoo`0047oo00<007ooOol0`goo00<007ooOol0Agoo
000@Ool20035Ool00`00Oomoo`16Ool0011oo`03001oogoo0<Eoo`03001oogoo04Eoo`0047oo00<0
07ooOol0aGoo00<007ooOol0AGoo000@Ool00`00Oomoo`36Ool00`00Oomoo`14Ool0011oo`03001o
ogoo0<Moo`03001oogoo04=oo`0047oo00<007ooOol0b7oo00<007ooOol0@Woo000@Ool00`00Oomo
o`39Ool00`00Oomoo`11Ool0011oo`800<Yoo`03001oogoo045oo`0047oo00<007ooOol0bWoo00<0
07ooOol0@7oo000@Ool00`00Oomoo`3;Ool00`00Oomoo`0oOol0011oo`03001oogoo0<aoo`03001o
ogoo03ioo`0027oo0`001Goo00<007ooOol0c7oo00<007ooOol0?Woo0008Ool01000Oomoo`0017oo
00<007ooOol0cGoo00<007ooOol0?Goo0008Ool01000Oomoo`0017oo0`00cWoo00<007ooOol0?7oo
0008Ool30005Ool00`00Oomoo`3>Ool00`00Oomoo`0lOol000Uoo`03001oogoo00Aoo`03001oogoo
0<moo`03001oogoo03]oo`002Goo0`0017oo00<007ooOol0d7oo00<007ooOol0>Woo000@Ool00`00
Oomoo`3@Ool00`00Oomoo`0jOol0011oo`03001oogoo0=5oo`03001oogoo03Uoo`0047oo0P00dgoo
00<007ooOol0>7oo000@Ool00`00Oomoo`3COol00`00Oomoo`0gOol0011oo`03001oogoo0==oo`03
001oogoo03Moo`0047oo00<007ooOol0e7oo00<007ooOol0=Woo000@Ool00`00Oomoo`3EOol00`00
Oomoo`0eOol0011oo`03001oogoo0=Eoo`03001oogoo03Eoo`0047oo00<007ooOol0eWoo00<007oo
Ool0=7oo000@Ool2003HOol00`00Oomoo`0cOol0011oo`03001oogoo0=Moo`03001oogoo03=oo`00
47oo00<007ooOol0f7oo00<007ooOol0<Woo000@Ool00`00Oomoo`3IOol00`00Oomoo`0aOol0011o
o`03001oogoo0=Uoo`03001oogoo035oo`0047oo00<007ooOol0fWoo00<007ooOol0<7oo000@Ool2
003LOol00`00Oomoo`0_Ool0011oo`03001oogoo0=]oo`03001oogoo02moo`0047oo00<007ooOol0
g7oo00<007ooOol0;Woo000@Ool00`00Oomoo`3LOol00`00Oomoo`0^Ool000Uoo`8000Eoo`03001o
ogoo0=eoo`03001oogoo02eoo`0027oo00@007ooOol000Aoo`03001oogoo0=ioo`03001oogoo02ao
o`0027oo00@007ooOol000Aoo`<00=ioo`03001oogoo02aoo`002Goo0P001Goo00<007ooOol0ggoo
00<007ooOol0:goo0008Ool01000Oomoo`0017oo00<007ooOol0h7oo00<007ooOol0:Woo0009Ool2
0005Ool00`00Oomoo`3POol00`00Oomoo`0ZOol0011oo`03001oogoo0>5oo`03001oogoo02Uoo`00
47oo00<007ooOol0hWoo00<007ooOol0:7oo000@Ool2003SOol00`00Oomoo`0XOol0011oo`03001o
ogoo0>=oo`03001oogoo02Moo`0047oo00<007ooOol0hgoo00<007ooOol09goo000@Ool00`00Oomo
o`3TOol00`00Oomoo`0VOol0011oo`03001oogoo0>Eoo`03001oogoo02Eoo`0047oo00<007ooOol0
iGoo00<007ooOol09Goo000@Ool00`00Oomoo`3VOol00`00Oomoo`0TOol0011oo`800>Moo`03001o
ogoo02Aoo`0047oo00<007ooOol0igoo00<007ooOol08goo000@Ool00`00Oomoo`3XOol00`00Oomo
o`0ROol0011oo`03001oogoo0>Qoo`03001oogoo029oo`0047oo00<007ooOol0jGoo00<007ooOol0
8Goo000@Ool00`00Oomoo`3ZOol00`00Oomoo`0POol0011oo`800>]oo`03001oogoo021oo`0047oo
00<007ooOol0jgoo00<007ooOol07goo000@Ool00`00Oomoo`3[Ool00`00Oomoo`0OOol0011oo`03
001oogoo0>aoo`03001oogoo01ioo`000Woo1@000Woo0P001Goo00<007ooOol0kGoo00<007ooOol0
7Goo0004Ool01@00Oomoogoo00000Woo00<007ooOol00Woo00<007ooOol0kGoo00<007ooOol07Goo
0004Ool01@00Oomoogoo00000Woo00<007ooOol00Woo0`00kWoo00<007ooOol077oo0004Ool01@00
Oomoogoo00000Woo00<007ooOol00Woo00<007ooOol0kWoo00<007ooOol077oo0004Ool01@00Oomo
ogoo00000Woo00<007ooOol00Woo00<007ooOol0kgoo00<007ooOol06goo0003Ool20004Ool20005
Ool00`00Oomoo`3_Ool00`00Oomoo`0KOol0011oo`03001oogoo0?1oo`03001oogoo01Yoo`0047oo
00<007ooOol0l7oo00<007ooOol06Woo000@Ool00`00Oomoo`3aOol00`00Oomoo`0IOol0011oo`80
0?=oo`03001oogoo01Qoo`0047oo00<007ooOol0lWoo00<007ooOol067oo000@Ool00`00Oomoo`3c
Ool00`00Oomoo`0GOol0011oo`03001oogoo0?=oo`03001oogoo01Moo`0047oo00<007ooOol0m7oo
00<007ooOol05Woo000@Ool00`00Oomoo`3dOol00`00Oomoo`0FOol0011oo`800?Ioo`03001oogoo
01Eoo`0047oo00<007ooOol0mGoo00<007ooOol05Goo000@Ool00`00Oomoo`3fOol00`00Oomoo`0D
Ool0011oo`03001oogoo0?Moo`03001oogoo01=oo`0047oo00<007ooOol0mgoo00<007ooOol04goo
000@Ool00`00Oomoo`3hOol00`00Oomoo`0BOol0011oo`800?Uoo`03001oogoo019oo`0047oo00<0
07ooOol0nGoo00<007ooOol04Goo000@Ool00`00Oomoo`3iOol00`00Oomoo`0AOol0011oo`03001o
ogoo0?Yoo`03001oogoo011oo`0047oo00<007ooOol0ngoo00<007ooOol03goo0002Ool500000goo
000000020004Ool00`00Oomoo`3kOol00`00Oomoo`0?Ool000Aoo`05001oogooOol00007Ool00`00
Oomoo`3lOol00`00Oomoo`0>Ool000Aoo`03001oogoo009oo`03001oogoo00Aoo`<00?aoo`03001o
ogoo00ioo`0017oo00<007ooOol00goo00<007ooOol00goo00<007ooOol0oGoo00<007ooOol03Goo
0004Ool01@00Oomoogoo00000Woo00<007ooOol00Woo00<007ooOol0oWoo00<007ooOol037oo0003
Ool20004Ool20005Ool00`00Oomoo`3nOol00`00Oomoo`0<Ool0011oo`03001oogoo0?moo`03001o
ogoo00]oo`0047oo00<007ooOol0ogoo00<007ooOol02goo000@Ool2003oOol2Ool00`00Oomoo`0:
Ool0011oo`03001oogoo0?moo`5oo`03001oogoo00Yoo`0047oo00<007ooOol0ogoo0Woo00<007oo
Ool02Goo000@Ool00`00Oomoo`3oOol>Ool0011oo`03001oogoo0?moo`ioo`0047oo00<007ooOol0
ogoo3Woo000@Ool2003oOol?Ool00001\
\>"],
  ImageRangeCache->{{{0, 287}, {176.938, 0}} -> {77.4058, -0.980872, \
0.154011, 0.0790494}}],

Cell[BoxData[
    TagBox[\(\[SkeletonIndicator]  Graphics  \[SkeletonIndicator]\),
      False,
      Editable->False]], "Output"]
}, Open  ]],

Cell["Or make it depend on both time to maturity and spot price:", "Text"],

Cell[BoxData[
    \(optval[s_, t_] := 
      Value[EuropeanOption[110, t, PutOption], 
        BlackScholesModel[s, 0.10, 0.01, 0.15]]\)], "Input"],

Cell["And make a 3D plot:", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Plot3D[optval[s, t], {s, 70, 140}, {t, 0.01, 1}]\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .81114 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% SurfaceGraphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
5.55112e-017 1.04977 -0.0679587 1.04977 [
[.11158 .21312 -12 -8.95035 ]
[.11158 .21312 0 .04965 ]
[.28836 .14356 -17.0282 -9 ]
[.28836 .14356 .97183 0 ]
[.47788 .06904 -15.9559 -9 ]
[.47788 .06904 2.04408 0 ]
[.68156 -0.01098 -14.8837 -9 ]
[.68156 -0.01098 3.11632 0 ]
[.76488 .0931 0 -6.1657 ]
[.76488 .0931 18 2.8343 ]
[.82636 .18723 0 -6.07503 ]
[.82636 .18723 18 2.92497 ]
[.88261 .27338 0 -5.99372 ]
[.88261 .27338 18 3.00628 ]
[.93428 .35251 0 -5.9204 ]
[.93428 .35251 18 3.0796 ]
[.98191 .42546 0 -5.85393 ]
[.98191 .42546 6 3.14607 ]
[.02411 .26511 -6 -2.74232 ]
[.02411 .26511 0 6.25768 ]
[.01566 .31809 -12 -2.78711 ]
[.01566 .31809 0 6.21289 ]
[.00691 .37292 -12 -2.83364 ]
[.00691 .37292 0 6.16636 ]
[-0.00215 .42972 -12 -2.88199 ]
[-0.00215 .42972 0 6.11801 ]
[-0.01154 .48857 -12 -2.93229 ]
[-0.01154 .48857 0 6.06771 ]
[ 0 0 0 0 ]
[ 1 .81114 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.03716 .25514 m
.68874 0 L
s
.12091 .22234 m
.12558 .22696 L
s
[(80)] .11158 .21312 1 .98897 Mshowa
.2971 .15335 m
.30147 .15825 L
s
[(100)] .28836 .14356 .89202 1 Mshowa
.4859 .07942 m
.48992 .08461 L
s
[(120)] .47788 .06904 .77288 1 Mshowa
.68874 0 m
.69233 .00549 L
s
[(140)] .68156 -0.01098 .65374 1 Mshowa
.125 Mabswid
.16384 .20553 m
.1666 .20834 L
s
.20751 .18843 m
.21022 .19129 L
s
.25192 .17104 m
.25459 .17394 L
s
.34306 .13535 m
.34564 .13834 L
s
.38984 .11704 m
.39236 .12006 L
s
.43745 .0984 m
.43991 .10147 L
s
.53524 .06011 m
.53759 .06327 L
s
.58547 .04044 m
.58776 .04364 L
s
.63663 .0204 m
.63885 .02365 L
s
.07869 .23888 m
.08153 .2416 L
s
.25 Mabswid
.68874 0 m
.96935 .42924 L
s
.75258 .09765 m
.74642 .09993 L
s
[(0.2)] .76488 .0931 -1 .37015 Mshowa
.81397 .19156 m
.80778 .19373 L
s
[(0.4)] .82636 .18723 -1 .35001 Mshowa
.87016 .27751 m
.86393 .27958 L
s
[(0.6)] .88261 .27338 -1 .33194 Mshowa
.92177 .35646 m
.91551 .35844 L
s
[(0.8)] .93428 .35251 -1 .31564 Mshowa
.96935 .42924 m
.96306 .43113 L
s
[(1)] .98191 .42546 -1 .30087 Mshowa
.125 Mabswid
.70266 .0213 m
.69899 .02272 L
s
.71969 .04735 m
.71602 .04876 L
s
.73633 .0728 m
.73264 .07418 L
s
.76845 .12194 m
.76475 .12329 L
s
.78397 .14567 m
.78026 .147 L
s
.79914 .16888 m
.79543 .17019 L
s
.82848 .21376 m
.82476 .21504 L
s
.84267 .23547 m
.83895 .23674 L
s
.85656 .25671 m
.85283 .25797 L
s
.88347 .29787 m
.87973 .2991 L
s
.8965 .31781 m
.89275 .31902 L
s
.90926 .33733 m
.90551 .33853 L
s
.93402 .37521 m
.93027 .37638 L
s
.94603 .39358 m
.94227 .39473 L
s
.95781 .41158 m
.95404 .41273 L
s
.25 Mabswid
.03716 .25514 m
0 .48963 L
s
.03634 .26033 m
.04245 .25795 L
s
[(0)] .02411 .26511 1 -0.3906 Mshowa
.02792 .31342 m
.03406 .31109 L
s
[(10)] .01566 .31809 1 -0.38064 Mshowa
.01922 .36837 m
.02537 .36609 L
s
[(20)] .00691 .37292 1 -0.3703 Mshowa
.0102 .42528 m
.01637 .42306 L
s
[(30)] -0.00215 .42972 1 -0.35956 Mshowa
.00085 .48425 m
.00705 .48209 L
s
[(40)] -0.01154 .48857 1 -0.34838 Mshowa
.125 Mabswid
.03468 .27081 m
.03835 .26938 L
s
.03301 .28135 m
.03668 .27993 L
s
.03132 .29197 m
.035 .29055 L
s
.02963 .30266 m
.03331 .30125 L
s
.02621 .32426 m
.02989 .32286 L
s
.02448 .33517 m
.02816 .33378 L
s
.02274 .34616 m
.02642 .34478 L
s
.02098 .35723 m
.02467 .35585 L
s
.01744 .37959 m
.02113 .37823 L
s
.01565 .39089 m
.01934 .38954 L
s
.01384 .40227 m
.01754 .40092 L
s
.01203 .41373 m
.01573 .41239 L
s
.00836 .4369 m
.01206 .43558 L
s
.0065 .44861 m
.01021 .44729 L
s
.00463 .4604 m
.00834 .45909 L
s
.00275 .47228 m
.00646 .47098 L
s
.25 Mabswid
.03716 .25514 m
0 .48963 L
s
0 .48963 m
.39787 .81114 L
s
.39787 .81114 m
.40529 .59895 L
s
.40529 .59895 m
.03716 .25514 L
s
.68874 0 m
.96935 .42924 L
s
.96935 .42924 m
1 .6535 L
s
1 .6535 m
.70298 .24544 L
s
.70298 .24544 m
.68874 0 L
s
.03716 .25514 m
0 .48963 L
s
0 .48963 m
.70298 .24544 L
s
.70298 .24544 m
.68874 0 L
s
.68874 0 m
.03716 .25514 L
s
.40529 .59895 m
.96935 .42924 L
s
.96935 .42924 m
1 .6535 L
s
1 .6535 m
.39787 .81114 L
s
.39787 .81114 m
.40529 .59895 L
s
0 0 m
1 0 L
1 .81114 L
0 .81114 L
closepath
clip
newpath
.5 Mabswid
.82 .799 .832 r
.37665 .73883 .39984 .75482 .43882 .71912 .41618 .70247 Metetra
.817 .798 .834 r
.41618 .70247 .43882 .71912 .47778 .68421 .4557 .66678 Metetra
.811 .797 .838 r
.4557 .66678 .47778 .68421 .51676 .65075 .49522 .6324 Metetra
.801 .793 .845 r
.49522 .6324 .51676 .65075 .55581 .6195 .53479 .60013 Metetra
.786 .788 .854 r
.53479 .60013 .55581 .6195 .59502 .59107 .57449 .57067 Metetra
.767 .781 .864 r
.57449 .57067 .59502 .59107 .63451 .56577 .61444 .54442 Metetra
.746 .773 .875 r
.61444 .54442 .63451 .56577 .67439 .54355 .65476 .52139 Metetra
.725 .764 .884 r
.65476 .52139 .67439 .54355 .71476 .52403 .69557 .50123 Metetra
.706 .757 .891 r
.69557 .50123 .71476 .52403 .7557 .50668 .73695 .48337 Metetra
.692 .751 .897 r
.73695 .48337 .7557 .50668 .79726 .49091 .77895 .46719 Metetra
.681 .747 .901 r
.77895 .46719 .79726 .49091 .83947 .47619 .82161 .45211 Metetra
.674 .744 .904 r
.82161 .45211 .83947 .47619 .88234 .46209 .86494 .43768 Metetra
.669 .743 .905 r
.86494 .43768 .88234 .46209 .92587 .44832 .90896 .42358 Metetra
.667 .742 .906 r
.90896 .42358 .92587 .44832 .97009 .43466 .95367 .40957 Metetra
.82 .799 .832 r
.35279 .72241 .37665 .73883 .41618 .70247 .3929 .68541 Metetra
.818 .799 .833 r
.3929 .68541 .41618 .70247 .4557 .66678 .43299 .64894 Metetra
.813 .797 .837 r
.43299 .64894 .4557 .66678 .49522 .6324 .47307 .61364 Metetra
.804 .794 .843 r
.47307 .61364 .49522 .6324 .53479 .60013 .51317 .58031 Metetra
.789 .789 .852 r
.51317 .58031 .53479 .60013 .57449 .57067 .55339 .54976 Metetra
.771 .782 .862 r
.55339 .54976 .57449 .57067 .61444 .54442 .59382 .5225 Metetra
.749 .774 .873 r
.59382 .5225 .61444 .54442 .65476 .52139 .6346 .49862 Metetra
.727 .765 .883 r
.6346 .49862 .65476 .52139 .69557 .50123 .67585 .47778 Metetra
.707 .757 .891 r
.67585 .47778 .69557 .50123 .73695 .48337 .71767 .4594 Metetra
.692 .751 .897 r
.71767 .4594 .73695 .48337 .77895 .46719 .76013 .4428 Metetra
.681 .746 .901 r
.76013 .4428 .77895 .46719 .82161 .45211 .80325 .42735 Metetra
.673 .744 .904 r
.80325 .42735 .82161 .45211 .86494 .43768 .84706 .41258 Metetra
.669 .742 .905 r
.84706 .41258 .86494 .43768 .90896 .42358 .89157 .39813 Metetra
.666 .742 .906 r
.89157 .39813 .90896 .42358 .95367 .40957 .93679 .38376 Metetra
.821 .8 .832 r
.32822 .70556 .35279 .72241 .3929 .68541 .36893 .66791 Metetra
.819 .799 .833 r
.36893 .66791 .3929 .68541 .43299 .64894 .40961 .63068 Metetra
.815 .798 .836 r
.40961 .63068 .43299 .64894 .47307 .61364 .45027 .59445 Metetra
.806 .795 .841 r
.45027 .59445 .47307 .61364 .51317 .58031 .49094 .56004 Metetra
.793 .79 .85 r
.49094 .56004 .51317 .58031 .55339 .54976 .53169 .52835 Metetra
.774 .783 .86 r
.53169 .52835 .55339 .54976 .59382 .5225 .57262 .50001 Metetra
.752 .774 .871 r
.57262 .50001 .59382 .5225 .6346 .49862 .61387 .47521 Metetra
.729 .765 .882 r
.61387 .47521 .6346 .49862 .67585 .47778 .65559 .45365 Metetra
.708 .757 .89 r
.65559 .45365 .67585 .47778 .71767 .4594 .69786 .43472 Metetra
.692 .751 .896 r
.69786 .43472 .71767 .4594 .76013 .4428 .74078 .41769 Metetra
.68 .746 .901 r
.74078 .41769 .76013 .4428 .80325 .42735 .78437 .40188 Metetra
.673 .744 .904 r
.78437 .40188 .80325 .42735 .84706 .41258 .82867 .38676 Metetra
.668 .742 .905 r
.82867 .38676 .84706 .41258 .89157 .39813 .87368 .37195 Metetra
.666 .742 .907 r
.87368 .37195 .89157 .39813 .93679 .38376 .91942 .35721 Metetra
.821 .8 .831 r
.30293 .68824 .32822 .70556 .36893 .66791 .34424 .64995 Metetra
.82 .799 .832 r
.34424 .64995 .36893 .66791 .40961 .63068 .38554 .61197 Metetra
.817 .798 .835 r
.38554 .61197 .40961 .63068 .45027 .59445 .42681 .57481 Metetra
.809 .796 .84 r
.42681 .57481 .45027 .59445 .49094 .56004 .46806 .5393 Metetra
.796 .791 .848 r
.46806 .5393 .49094 .56004 .53169 .52835 .50937 .50641 Metetra
.778 .784 .858 r
.50937 .50641 .53169 .52835 .57262 .50001 .55082 .47692 Metetra
.755 .775 .87 r
.55082 .47692 .57262 .50001 .61387 .47521 .59257 .45113 Metetra
.731 .766 .88 r
.59257 .45113 .61387 .47521 .65559 .45365 .63475 .42881 Metetra
.71 .757 .889 r
.63475 .42881 .65559 .45365 .69786 .43472 .67749 .40931 Metetra
.692 .75 .896 r
.67749 .40931 .69786 .43472 .74078 .41769 .72088 .39184 Metetra
.68 .746 .901 r
.72088 .39184 .74078 .41769 .78437 .40188 .76496 .37567 Metetra
.672 .743 .904 r
.76496 .37567 .78437 .40188 .82867 .38676 .80975 .36019 Metetra
.668 .742 .906 r
.80975 .36019 .82867 .38676 .87368 .37195 .85528 .34501 Metetra
.666 .741 .907 r
.85528 .34501 .87368 .37195 .91942 .35721 .90154 .32988 Metetra
.822 .8 .831 r
.27686 .67043 .30293 .68824 .34424 .64995 .31881 .6315 Metetra
.821 .8 .832 r
.31881 .6315 .34424 .64995 .38554 .61197 .36074 .59279 Metetra
.818 .799 .834 r
.36074 .59279 .38554 .61197 .42681 .57481 .40264 .55471 Metetra
.812 .797 .838 r
.40264 .55471 .42681 .57481 .46806 .5393 .44451 .51808 Metetra
.8 .792 .845 r
.44451 .51808 .46806 .5393 .50937 .50641 .4864 .48393 Metetra
.782 .785 .856 r
.4864 .48393 .50937 .50641 .55082 .47692 .52839 .4532 Metetra
.759 .776 .867 r
.52839 .4532 .55082 .47692 .59257 .45113 .57065 .42635 Metetra
.734 .766 .879 r
.57065 .42635 .59257 .45113 .63475 .42881 .61332 .40321 Metetra
.711 .757 .889 r
.61332 .40321 .63475 .42881 .67749 .40931 .65654 .38313 Metetra
.692 .75 .896 r
.65654 .38313 .67749 .40931 .72088 .39184 .70041 .36522 Metetra
.68 .745 .901 r
.70041 .36522 .72088 .39184 .76496 .37567 .74498 .34867 Metetra
.672 .743 .904 r
.74498 .34867 .76496 .37567 .80975 .36019 .79028 .33283 Metetra
.668 .742 .906 r
.79028 .33283 .80975 .36019 .85528 .34501 .83633 .31728 Metetra
.665 .741 .907 r
.83633 .31728 .85528 .34501 .90154 .32988 .88313 .30174 Metetra
.822 .8 .831 r
.24999 .65212 .27686 .67043 .31881 .6315 .2926 .61254 Metetra
.821 .8 .832 r
.2926 .61254 .31881 .6315 .36074 .59279 .33518 .5731 Metetra
.819 .799 .833 r
.33518 .5731 .36074 .59279 .40264 .55471 .37773 .53414 Metetra
.814 .797 .836 r
.37773 .53414 .40264 .55471 .44451 .51808 .42024 .49638 Metetra
.803 .794 .843 r
.42024 .49638 .44451 .51808 .4864 .48393 .46274 .46091 Metetra
.786 .787 .853 r
.46274 .46091 .4864 .48393 .52839 .4532 .50531 .42884 Metetra
.763 .777 .865 r
.50531 .42884 .52839 .4532 .57065 .42635 .5481 .40083 Metetra
.736 .767 .877 r
.5481 .40083 .57065 .42635 .61332 .40321 .59127 .37682 Metetra
.712 .757 .888 r
.59127 .37682 .61332 .40321 .65654 .38313 .63498 .35613 Metetra
.692 .749 .895 r
.63498 .35613 .65654 .38313 .70041 .36522 .67934 .33778 Metetra
.679 .745 .901 r
.67934 .33778 .70041 .36522 .74498 .34867 .72441 .32087 Metetra
.671 .742 .904 r
.72441 .32087 .74498 .34867 .79028 .33283 .77023 .30467 Metetra
.667 .742 .906 r
.77023 .30467 .79028 .33283 .83633 .31728 .81681 .28872 Metetra
.665 .741 .907 r
.81681 .28872 .83633 .31728 .88313 .30174 .86417 .27276 Metetra
.822 .801 .831 r
.22228 .63326 .24999 .65212 .2926 .61254 .26556 .59303 Metetra
.822 .8 .831 r
.26556 .59303 .2926 .61254 .33518 .5731 .30881 .55289 Metetra
.82 .8 .832 r
.30881 .55289 .33518 .5731 .37773 .53414 .35204 .51306 Metetra
.816 .798 .835 r
.35204 .51306 .37773 .53414 .42024 .49638 .39522 .47417 Metetra
.807 .795 .841 r
.39522 .47417 .42024 .49638 .46274 .46091 .43837 .43733 Metetra
.791 .788 .85 r
.43837 .43733 .46274 .46091 .50531 .42884 .48155 .40382 Metetra
.767 .778 .863 r
.48155 .40382 .50531 .42884 .5481 .40083 .52489 .37455 Metetra
.739 .767 .876 r
.52489 .37455 .5481 .40083 .59127 .37682 .56857 .34959 Metetra
.713 .757 .887 r
.56857 .34959 .59127 .37682 .63498 .35613 .61278 .32826 Metetra
.692 .749 .895 r
.61278 .32826 .63498 .35613 .67934 .33778 .65764 .30949 Metetra
.678 .744 .901 r
.65764 .30949 .67934 .33778 .72441 .32087 .70323 .29221 Metetra
.67 .742 .904 r
.70323 .29221 .72441 .32087 .77023 .30467 .74958 .27565 Metetra
.666 .741 .906 r
.74958 .27565 .77023 .30467 .81681 .28872 .7967 .2593 Metetra
.665 .741 .907 r
.7967 .2593 .81681 .28872 .86417 .27276 .84463 .2429 Metetra
.822 .801 .831 r
.19369 .61383 .22228 .63326 .26556 .59303 .23766 .57294 Metetra
.822 .801 .831 r
.23766 .57294 .26556 .59303 .30881 .55289 .28161 .5321 Metetra
.821 .8 .832 r
.28161 .5321 .30881 .55289 .35204 .51306 .32554 .49144 Metetra
.818 .799 .833 r
.32554 .49144 .35204 .51306 .39522 .47417 .36942 .45146 Metetra
.811 .796 .838 r
.36942 .45146 .39522 .47417 .43837 .43733 .41325 .41321 Metetra
.795 .79 .847 r
.41325 .41321 .43837 .43733 .48155 .40382 .45706 .37813 Metetra
.772 .78 .86 r
.45706 .37813 .48155 .40382 .52489 .37455 .50098 .34745 Metetra
.742 .768 .873 r
.50098 .34745 .52489 .37455 .56857 .34959 .54521 .32146 Metetra
.714 .756 .886 r
.54521 .32146 .56857 .34959 .61278 .32826 .58993 .29949 Metetra
.691 .748 .895 r
.58993 .29949 .61278 .32826 .65764 .30949 .6353 .2803 Metetra
.677 .743 .901 r
.6353 .2803 .65764 .30949 .70323 .29221 .68141 .26268 Metetra
.669 .741 .904 r
.68141 .26268 .70323 .29221 .74958 .27565 .72829 .24575 Metetra
.666 .741 .906 r
.72829 .24575 .74958 .27565 .7967 .2593 .77598 .22899 Metetra
.664 .741 .907 r
.77598 .22899 .7967 .2593 .84463 .2429 .82447 .21212 Metetra
.822 .801 .831 r
.16417 .59381 .19369 .61383 .23766 .57294 .20886 .55224 Metetra
.822 .801 .831 r
.20886 .55224 .23766 .57294 .28161 .5321 .25353 .5107 Metetra
.822 .801 .831 r
.25353 .5107 .28161 .5321 .32554 .49144 .29817 .46925 Metetra
.82 .8 .832 r
.29817 .46925 .32554 .49144 .36942 .45146 .34278 .42823 Metetra
.814 .797 .836 r
.34278 .42823 .36942 .45146 .41325 .41321 .38733 .38854 Metetra
.801 .791 .844 r
.38733 .38854 .41325 .41321 .45706 .37813 .43182 .35177 Metetra
.777 .781 .856 r
.43182 .35177 .45706 .37813 .50098 .34745 .47636 .3195 Metetra
.746 .768 .871 r
.47636 .3195 .50098 .34745 .54521 .32146 .52114 .29237 Metetra
.715 .756 .884 r
.52114 .29237 .54521 .32146 .58993 .29949 .56638 .26974 Metetra
.69 .747 .895 r
.56638 .26974 .58993 .29949 .6353 .2803 .61227 .25017 Metetra
.675 .742 .901 r
.61227 .25017 .6353 .2803 .68141 .26268 .65892 .23223 Metetra
.668 .741 .905 r
.65892 .23223 .68141 .26268 .72829 .24575 .70635 .21494 Metetra
.665 .741 .907 r
.70635 .21494 .72829 .24575 .77598 .22899 .7546 .19774 Metetra
.664 .741 .907 r
.7546 .19774 .77598 .22899 .82447 .21212 .80369 .18037 Metetra
.822 .801 .831 r
.13368 .57316 .16417 .59381 .20886 .55224 .17911 .5309 Metetra
.822 .801 .831 r
.17911 .5309 .20886 .55224 .25353 .5107 .22452 .48865 Metetra
.822 .801 .831 r
.22452 .48865 .25353 .5107 .29817 .46925 .26991 .44644 Metetra
.821 .8 .832 r
.26991 .44644 .29817 .46925 .34278 .42823 .31527 .40445 Metetra
.818 .799 .834 r
.31527 .40445 .34278 .42823 .38733 .38854 .36057 .36335 Metetra
.806 .793 .84 r
.36057 .36335 .38733 .38854 .43182 .35177 .40579 .32474 Metetra
.783 .783 .853 r
.40579 .32474 .43182 .35177 .47636 .3195 .45098 .29065 Metetra
.75 .769 .868 r
.45098 .29065 .47636 .3195 .52114 .29237 .49634 .26224 Metetra
.715 .755 .883 r
.49634 .26224 .52114 .29237 .56638 .26974 .54212 .23894 Metetra
.689 .745 .894 r
.54212 .23894 .56638 .26974 .61227 .25017 .58854 .21905 Metetra
.673 .741 .901 r
.58854 .21905 .61227 .25017 .65892 .23223 .63572 .20084 Metetra
.667 .74 .905 r
.63572 .20084 .65892 .23223 .70635 .21494 .68372 .18318 Metetra
.664 .741 .907 r
.68372 .18318 .70635 .21494 .7546 .19774 .73255 .16551 Metetra
.664 .741 .908 r
.73255 .16551 .7546 .19774 .80369 .18037 .78223 .14762 Metetra
.823 .801 .831 r
.10217 .55185 .13368 .57316 .17911 .5309 .14836 .50887 Metetra
.823 .801 .831 r
.14836 .50887 .17911 .5309 .22452 .48865 .19454 .4659 Metetra
.822 .801 .831 r
.19454 .4659 .22452 .48865 .26991 .44644 .2407 .42295 Metetra
.822 .801 .831 r
.2407 .42295 .26991 .44644 .31527 .40445 .28683 .38008 Metetra
.82 .8 .832 r
.28683 .38008 .31527 .40445 .36057 .36335 .33292 .33767 Metetra
.812 .796 .837 r
.33292 .33767 .36057 .36335 .40579 .32474 .37891 .2971 Metetra
.791 .786 .848 r
.37891 .2971 .40579 .32474 .45098 .29065 .42481 .26086 Metetra
.755 .77 .865 r
.42481 .26086 .45098 .29065 .49634 .26224 .47078 .23093 Metetra
.715 .753 .882 r
.47078 .23093 .49634 .26224 .54212 .23894 .51711 .207 Metetra
.686 .743 .894 r
.51711 .207 .54212 .23894 .58854 .21905 .56407 .18691 Metetra
.671 .74 .902 r
.56407 .18691 .58854 .21905 .63572 .20084 .6118 .16848 Metetra
.665 .74 .906 r
.6118 .16848 .63572 .20084 .68372 .18318 .66036 .15043 Metetra
.664 .741 .907 r
.66036 .15043 .68372 .18318 .73255 .16551 .70978 .13226 Metetra
.664 .741 .908 r
.70978 .13226 .73255 .16551 .78223 .14762 .76008 .1138 Metetra
.823 .801 .831 r
.06958 .52986 .10217 .55185 .14836 .50887 .11657 .48612 Metetra
.823 .801 .831 r
.11657 .48612 .14836 .50887 .19454 .4659 .16354 .44241 Metetra
.823 .801 .831 r
.16354 .44241 .19454 .4659 .2407 .42295 .21049 .39871 Metetra
.823 .801 .831 r
.21049 .39871 .2407 .42295 .28683 .38008 .25742 .35504 Metetra
.822 .801 .831 r
.25742 .35504 .28683 .38008 .33292 .33767 .30432 .31151 Metetra
.818 .798 .833 r
.30432 .31151 .33292 .33767 .37891 .2971 .35114 .26896 Metetra
.8 .789 .842 r
.35114 .26896 .37891 .2971 .42481 .26086 .39782 .23008 Metetra
.762 .771 .86 r
.39782 .23008 .42481 .26086 .47078 .23093 .44444 .19827 Metetra
.714 .75 .88 r
.44444 .19827 .47078 .23093 .51711 .207 .49132 .17381 Metetra
.681 .739 .895 r
.49132 .17381 .51711 .207 .56407 .18691 .53882 .15375 Metetra
.667 .739 .903 r
.53882 .15375 .56407 .18691 .6118 .16848 .58711 .13515 Metetra
.664 .74 .907 r
.58711 .13515 .6118 .16848 .66036 .15043 .63625 .11667 Metetra
.664 .741 .908 r
.63625 .11667 .66036 .15043 .70978 .13226 .68627 .09793 Metetra
.664 .741 .908 r
.68627 .09793 .70978 .13226 .76008 .1138 .73719 .07886 Metetra
.823 .802 .831 r
.03586 .50713 .06958 .52986 .11657 .48612 .08368 .46262 Metetra
.823 .801 .831 r
.08368 .46262 .11657 .48612 .16354 .44241 .13147 .41814 Metetra
.823 .801 .831 r
.13147 .41814 .16354 .44241 .21049 .39871 .17924 .37367 Metetra
.823 .801 .831 r
.17924 .37367 .21049 .39871 .25742 .35504 .22698 .32923 Metetra
.823 .801 .831 r
.22698 .32923 .25742 .35504 .30432 .31151 .27471 .28481 Metetra
.822 .8 .831 r
.27471 .28481 .30432 .31151 .35114 .26896 .3224 .24058 Metetra
.812 .794 .836 r
.3224 .24058 .35114 .26896 .39782 .23008 .36997 .1984 Metetra
.773 .773 .853 r
.36997 .1984 .39782 .23008 .44444 .19827 .41731 .16383 Metetra
.71 .744 .878 r
.41731 .16383 .44444 .19827 .49132 .17381 .46473 .13925 Metetra
.673 .734 .896 r
.46473 .13925 .49132 .17381 .53882 .15375 .51275 .11963 Metetra
.664 .738 .905 r
.51275 .11963 .53882 .15375 .58711 .13515 .56161 .10083 Metetra
.664 .741 .907 r
.56161 .10083 .58711 .13515 .63625 .11667 .61134 .08181 Metetra
.664 .741 .908 r
.61134 .08181 .63625 .11667 .68627 .09793 .66197 .06246 Metetra
.664 .741 .908 r
.66197 .06246 .68627 .09793 .73719 .07886 .71353 .04274 Metetra
.823 .802 .831 r
.00095 .48363 .03586 .50713 .08368 .46262 .04962 .43833 Metetra
.823 .802 .831 r
.04962 .43833 .08368 .46262 .13147 .41814 .09826 .39304 Metetra
.823 .802 .831 r
.09826 .39304 .13147 .41814 .17924 .37367 .14688 .34778 Metetra
.823 .801 .831 r
.14688 .34778 .17924 .37367 .22698 .32923 .19548 .30255 Metetra
.823 .801 .831 r
.19548 .30255 .22698 .32923 .27471 .28481 .24405 .25733 Metetra
.823 .801 .831 r
.24405 .25733 .27471 .28481 .3224 .24058 .2926 .21214 Metetra
.822 .8 .831 r
.2926 .21214 .3224 .24058 .36997 .1984 .34112 .16697 Metetra
.803 .785 .837 r
.34112 .16697 .36997 .1984 .41731 .16383 .38945 .12561 Metetra
.684 .718 .875 r
.38945 .12561 .41731 .16383 .46473 .13925 .43729 .10373 Metetra
.663 .731 .901 r
.43729 .10373 .46473 .13925 .51275 .11963 .48583 .08476 Metetra
.663 .74 .907 r
.48583 .08476 .51275 .11963 .56161 .10083 .53526 .06546 Metetra
.664 .741 .908 r
.53526 .06546 .56161 .10083 .61134 .08181 .58558 .0458 Metetra
.664 .741 .908 r
.58558 .0458 .61134 .08181 .66197 .06246 .63684 .02578 Metetra
.664 .741 .908 r
.63684 .02578 .66197 .06246 .71353 .04274 .68905 .00539 Metetra
0 g
.25 Mabswid
.68874 0 m
.96935 .42924 L
s
.96935 .42924 m
1 .6535 L
s
1 .6535 m
.70298 .24544 L
s
.70298 .24544 m
.68874 0 L
s
.03716 .25514 m
0 .48963 L
s
0 .48963 m
.70298 .24544 L
s
.70298 .24544 m
.68874 0 L
s
.68874 0 m
.03716 .25514 L
s
.03716 .25514 m
.68874 0 L
s
.12091 .22234 m
.12558 .22696 L
s
[(80)] .11158 .21312 1 .98897 Mshowa
.2971 .15335 m
.30147 .15825 L
s
[(100)] .28836 .14356 .89202 1 Mshowa
.4859 .07942 m
.48992 .08461 L
s
[(120)] .47788 .06904 .77288 1 Mshowa
.68874 0 m
.69233 .00549 L
s
[(140)] .68156 -0.01098 .65374 1 Mshowa
.125 Mabswid
.16384 .20553 m
.1666 .20834 L
s
.20751 .18843 m
.21022 .19129 L
s
.25192 .17104 m
.25459 .17394 L
s
.34306 .13535 m
.34564 .13834 L
s
.38984 .11704 m
.39236 .12006 L
s
.43745 .0984 m
.43991 .10147 L
s
.53524 .06011 m
.53759 .06327 L
s
.58547 .04044 m
.58776 .04364 L
s
.63663 .0204 m
.63885 .02365 L
s
.07869 .23888 m
.08153 .2416 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 233.563},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000jA000`40O003h00Oogoo8Goo002iOol5
0003Ool30002Ool2001HOol00;]oo`03001oogoo00Aoo`04001oogoo0002Ool00`00Oomoo`1EOol0
0;]oo`04001oogooOol500001Goo001oogoo0000Egoo002kOol01@00Oomoogoo00000Woo00@007oo
Ool0009oo`03001oogoo05Eoo`00^goo00<007ooOol00Woo00<007oo00000Woo00@007ooOol005Mo
o`00^Woo0P001Goo0P000goo0P00F7oo003oOolQOol00?moob5oo`00ogoo8Goo003oOolQOol00?mo
ob5oo`00b7oo0P00EWoo0036Ool200000goo0000001EOol00<=oo`<0009oo`<005Eoo`00`Goo0P00
0Woo0`0000AFo@00000005Aoo`00_Woo0`000Woo0P0015Km0P0000=oo`00Ool0DWoo002kOol30002
Ool30006E_d01000E_d00000Dgoo002iOol3000017oo000000002EKm00<005KmE_d00P00DWoo002f
Ool3000017oo0000000035Km00H005KmE_d007oo001AOol00;Aoo`80009oo`80009Fo@03001FoEKm
00YFo@04001FoEKmE_d2001AOol00;5oo`<0009oo`8000AFo@03001FoEKm00]Fo@04001FoEKmE_d2
001@Ool00:moo`80009oo`<000MFo@03001FoEKm00YFo@03001FoEKm009Fo@03001oogoo04ioo`00
PGoo1@0000=oo`0000000P000goo0P0077oo0`000Woo0P002eKm00<005KmE_d02EKm00<005KmE_d0
0UKm0P00Cgoo0023Ool01@00Oomoogoo00001Goo00@007ooOol001Uoo`80009oo`<000eFo@03001F
oEKm00UFo@03001FoEKm009Fo@<004ioo`00Pgoo00<007ooOol00Woo00<007ooOol00Woo00@007oo
Ool001Ioo`<0009oo`<0011Fo@03001FoEKm00QFo@03001Fo@000080009Fo@03001oogoo04aoo`00
Pgoo00<007ooOol00goo00D007ooOomoo`00009oo`03001oogoo019oo`8000=oo`80009Fo@03001F
oEKm00mFo@03001FoEKm00MFo@8000EFo@8004eoo`00Pgoo00D007ooOomoo`00009oo`04001oogoo
0002Ool00`00Oomoo`0?Ool30002Ool30004E_d00`00E_eFo@0@E_d00`00E_eFo@03E_d40007E_d2
001<Ool0089oo`8000Aoo`8000Aoo`8000moo`<0009oo`<000QFo@03001FoEKm00mFo@03001FoEKm
00<000=Fo@03001FoEKm00EFo@8004aoo`00WGoo0P0000=oo`00Ool00P0035Km00<005KmE_d03UKm
0`001UKm00<005KmE_d01UKm0P00Bgoo002JOol30002Ool3000>E_d00`00E_eFo@0<E_d20002E_d0
0`00E_eFo@04E_d00`00E_eFo@07E_d2001:Ool009Qoo`80009oo`<0009Fo@03001FoEKm00eFo@03
001FoEKm00UFo@8000AFo@03001FoEKm00AFo@03001FoEKm00MFo@8004Yoo`00UGoo0`000Woo0P00
1EKm00<005KmE_d03UKm00<005KmE_d01EKm0`001eKm00<005KmE_d00eKm00<005KmE_d025Km0P00
BGoo002COol20002Ool30008E_d00`00E_eFo@0>E_d00`00E_eFo@02E_d2000:E_d00`00E_eFo@03
E_d00`00E_eFo@08E_d00`00Ool00008Ool20004Ool20003Ool4000aOol0091oo`<00004Ool00000
000<E_d00`00E_eFo@0=E_d00`00E_eFo@02000=E_d00`00E_eFo@02E_d00`00E_eFo@08E_d30007
Ool01000Oomoo`0027oo00<007ooOol0<Woo002>Ool2000017oo000000003eKm00<005KmE_d03EKm
0`0045Km00D005KmE_eFo@0000MFo@<0009Fo@8000Ioo`04001oogoo0009Ool00`00Oomoo`0aOol0
08]oo`<00003Ool00000009Fo003001FoEKm00iFo@03001FoEKm00YFo@80009Fo@03001FoEKm00iF
o@05001FoEKmE_d00005E_d20005E_d00`00Ool00005Ool01000Oomoo`002Woo00<007ooOol0<7oo
0028Ool3000017oo0000000015Kl00<005KmE_d03eKm00<005KmE_d01UKm0`001EKm00<005KmE_d0
3UKm00@005KmE_d0009Fo@<000QFo@8000Eoo`04001oogoo0008Ool01000Oomoo`00<Goo0026Ool2
000017oo0000000025Kl00<005KmE_d03eKm00<005KmE_d00UKm0`0025Km00<005KmE_d03UKm00<0
05KmE_d00`0035Km0P001Goo0P002Woo0P00<Woo0023Ool300000goo0000000<E_`00`00E_eFo@0>
E_d00`00E_eFo@02000<E_d00`00E_eFo@0=E_d300000eKm001Fo@0<E_d30014Ool0085oo`800004
Ool00000000>E_`00`00E_eFo@0>E_d3000?E_d00`00E_eFo@0:E_d20002E_d00`00E_d0000>E_d2
0014Ool007ioo`<00003Ool00000009Ff`03001Fo5Kl00eFo003001FoEKm00YFo@<0009Fo003001F
oEKm00iFo@03001FoEKm00IFo@<000EFo@8000mFo@8004=oo`00O7oo0P000Woo0P0015KK00<005Kl
E_`03UKl00<005KmE_d01eKm0P001UKl00<005KmE_d03EKm00<005KmE_d00eKm0`0025Km0P003eKm
0P00@goo001=Ool50002Ool20004Ool2000MOol30002Ool20007E]/00`00E_aFo00>E_`00`00E_eF
o@03E_d30008E_`00`00E_eFo@0>E_d00`00E_eFo@02000<E_d00`00E_eFo@0=E_d30012Ool004mo
o`05001oogooOol00002Ool01000Oomoo`000Woo00<007ooOol067oo0P000goo0P002UKK00<005Kl
E_`03EKl00<005KmE_d00`0035Kl00<005KmE_d03EKm0`003UKm0P0035Km0P000UKm0P00@Goo001?
Ool01@00Oomoogoo00000Woo00@007ooOol0009oo`03001oogoo01Eoo`<0009oo`<000aFf`03001F
o5Kl00eFo0<0011Fo003001FoEKm00YFo@80009Fo@03001FoEKm00aFo@03001Fo@0000QFo@<000AF
o@<0041oo`00Cgoo00D007ooOomoo`00009oo`04001oogoo0002Ool00`00Oomoo`0COol20004Ool2
000?E]/00`00E_aFo00:E_`20002E_`00`00E_aFo00?E_`00`00E_eFo@06E_d30004E_d00`00E_eF
o@0<E_d00`00E_d00005E_d30008E_d20010Ool004moo`05001oogooOol00002Ool01000Oomoo`00
0Woo00<007ooOol047oo0`001Goo00<006Lj000045KK00<005KlE_`01UKl0`001EKl00<005KlE_`0
3UKl00<005KmE_d00eKm0`0025Km00<005KmE_d02eKm00@005KmE_d0009Fo@8000aFo@8003moo`00
CWoo0P0017oo0P0017oo0P0047oo0`001goo00@006LjIcX0015Ff`03001Fo5Kl009Fo0<000UFo003
001Fo5Kl00iFo003001FoEKm008000aFo@03001FoEKm00YFo@03001FoEKm00<000iFo@8003moo`00
JWoo0`0027oo00<006LjIcX00VLj00<005KKE]/03UKK00<005KlE_`00P003EKl00<005KlE_`03EKl
0`003UKm00<005KmE_d02UKm0`000UKm00<005KmE_d03EKm0P00?Woo001WOol30002Ool00`00Oomo
o`05Ool00`00IcYW>P04IcX00`00E]]Ff`0>E]/2000?E_`00`00E_aFo00;E_`20002E_`00`00E_eF
o@0=E_d00`00E_eFo@07E_d30004E_d00`00E_eFo@0>E_d2000mOol006Eoo`8000aoo`03001W>VLj
00IW>P03001FfeKK00]Ff`800003F_`005Kl00mFo003001Fo5Kl00MFo0<000AFo003001FoEKm00iF
o@03001FoEKm00=Fo@<0009Fo@03001FoEKm00=Fo@03001FoEKm00eFo@8003eoo`00HWoo0`0037oo
0P002FLj00<005KKE]/02EKK0P0015[l00<005KlE_`03UKl00<005KlE_`00eKl0`0025Kl00<005Km
E_d03EKm00<005KmE_d00`001EKm00<005KmE_d015Km00<005KmE_d035Km0`00?7oo001POol2000>
Ool00`00IcYW>P0:IcX00`00E]]Ff`07E]/00`00F_aJo005F_`00`00E_aFo00>E_`00`00E_aFo002
000<E_`00`00E_eFo@0<E_d30008E_d00`00E_eFo@04E_d00`00E_eFo@0:E_d20002E_d2000kOol0
05eoo`<000moo`03001W>VLj00aW>P03001FfeKK00AFf`8000UJo003001Fo5Kl00eFo0<000iFo003
001FoEKm00YFo@80009Fo003001FoEKm00MFo@03001FoEKm00AFo@03001FoEKm00IFo@<000AFo@80
03]oo`00Fgoo0P000goo00<007ooOol02goo00<006LjIcX03VLj00@005KKE]]Ff`8000]Jo003001F
o5Kl00]Fo080009Fo003001Fo5Kl00eFo003001FoEKm00IFo@<000EFo003001FoEKm00IFo@03001F
oEKm00EFo@03001FoEKm009Fo@<000QFo@8003Yoo`00F7oo0`004Woo00<006LjIcX03fLj00@005KK
000000iJo003001Fo5Kl00MFo0<000AFo003001Fo5Kl00iFo003001FoEKm009Fo@<000QFo003001F
oEKm00IFo@03001FoEKm00EFo@03001FoEKm008000aFo@8003Uoo`00EGoo0`0057oo00<006LjIcX0
4FLj00<005[lF_`03e[l00<005KlE_`00eKl0`0025Kl00<005KlE_`03EKl00<005KmE_d00P0035Kl
00<005KmE_d01EKm00<005KmE_d01EKm0`003UKm0P00>Goo001COol2000FOol00`00IcYW>P0@IcX2
00000f<K001Jo00@F_`00`00E_aFo002000<E_`00`00E_aFo00<E_`3000?E_`00`00E_eFo@04E_d0
0`00E_eFo@03E_d20002E_d00`00E_eFo@0=E_d2000hOol0051oo`D001Eoo`03001[>P00011W>P03
001S6f<K009S6`03001Jo5[l00mJo08000iFo003001Fo5Kl00YFo080009Fo003001Fo5Kl00iFo003
001FoEKm00=Fo@03001FoEKm00<000AFo@03001FoEKm00iFo@8000Moo`8000Aoo`8000Eoo`<0021o
o`00CWoo0P006Goo00D006/jJcY[>P0000eW>P8000ES6`03001Jo5[l00eJo0800003F_`005Kl00iF
o003001Fo5Kl00IFo0<000EFo003001Fo5Kl00eFo003001FoEKm00=Fo@<000QFo@03001FoEKm00eF
o@8000Ioo`04001oogoo000;Ool00`00Oomoo`0OOol004]oo`<001Yoo`03001[>V/j00=[>P03001W
>VLj00UW>P03001S6f<K00IS6`03001Jo5[l00YJo08000AJo003001Fo5Kl00eFo003001Fo5Kl009F
o0<000QFo003001Fo5Kl00iFo003001FoEKm00<000]Fo@03001FoEKm00aFo@<000Eoo`04001oogoo
0008Ool5000POol002=oo`8000Aoo`8001ioo`8001aoo`03001[>V/j00E[>P03001W>VLj00MW>P03
001S6f<K00QS6`03001Jo5[l00MJo08000IJo003001Fo5Kl00eFo003001Fo5Kl008000aFo003001F
o5Kl00eFo0<0009Fo@03001FoEKm00UFo@03001FoEKm00YFo@80009Fo@03001oogoo00=oo`04001o
ogoo0008Ool01000Oomoo`008Goo000ROol01000Oomoo`000Woo00@007ooOol001Yoo`<001eoo`03
001[>V/j00M[>P03001W>VLj00AW>P8000aS6`03001Jo5[l00AJo08000UJo003001Fo5Kl00aFo0<0
00mFo003001Fo5Kl00YFo080009Fo004001FoEKm000<E_d00`00E_eFo@06E_d30004E_d20004Ool0
1000Oomoo`002Goo00<007oo00008Goo000ROol01000Oomoo`000Woo00@007ooOol001Qoo`80009o
o`03001oogoo01Uoo`8000][>P03001W>VLj009W>P03001S6f<K00aS6`03001Jo5[l009Jo08000aJ
o003001Fo5Kl00UFo080009Fo003001Fo5Kl00iFo003001Fo5Kl00IFo0<000AFo004001FoEKm000=
E_d00`00E_eFo@02E_d30008E_d20004Ool2000;Ool2000QOol002=oo`8000=oo`04001oogoo000E
Ool3000OOol00`00JcY[>P0<JcX01000IcX0000046<K00@005[l000000mJo003001Fo5Kl00IFo080
00EFo003001Fo5Kl00eFo003001Fo5Kl00=Fo0<000QFo003001Fo@0000eFo@03001FoEKm008000aF
o@03001oogoo039oo`008Woo00@007ooOol0009oo`04001oogoo000BOol3000QOol00`00JcY[>P0>
JcX00`00Ha]S6`0AHa/00`00F_aJo00?F_`00`00E_aFo003E_`30007E_`00`00E_aFo00>E_`00`00
E_aFo002000<E_`2000=E_d3000>E_d2000cOol002=oo`8000Aoo`80015oo`8002=oo`03001[>V/j
00i[>P03001W>P00011S6`800003Ha/005[l011Jo003001Fo5Kl008000]Fo003001Fo5Kl00eFo0<0
00iFo08000]Fo@80009Fo@03001FoEKm00eFo@80039oo`00>Goo0`0097oo00<006/jJcX03V/j00D0
06LjIcYW>P0000iS6`03001S6f<K009S6`03001Jo5[l00mJo08000iFo003001Fo5Kl00YFo080009F
o003001Fo5Kl00eFo003001FoEKm00IFo@<000AFo@03001FoEKm00iFo@03001oogoo031oo`00=goo
0P000goo00<007ooOol087oo00<006/jJcX03V/j00<006LjIcX00VLj00<006<KHa/02V<K0P001F<K
00<005[lF_`03E[l0P0000=K7000E_`03UKl00<005KlE_`01UKl0`0015Kl00<005KlE_`03EKl0P00
15Km0`0025Km00<005KmE_d03EKm0P00<Goo000dOol3000WOol00`00JcY[>P0>JcX00`00IcYW>P04
IcX00`00Ha]S6`08Ha/00`00Ha]S6`06Ha/00`00F_aJo00:F_`20003Fa`00`00E_aFo00=E_`00`00
E_aFo003E_`30008E_`00`00E_aFo00<E_`20002E_d2000<E_d00`00E_eFo@0<E_d3000`Ool0039o
o`8002Uoo`03001[>P0000i[>P03001W>VLj00IW>P03001S6f<K00IS6`03001S6f<K00QS6`03001J
o5[l00MJo08000IK7003001Fo5Kl00eFo003001Fo5Kl008000aFo003001Fo5Kl00aFo0<000mFo@03
001FoEKm00QFo@<0009Fo@03001oogoo02ioo`00;goo0`00:Woo00D006/jJcY[>P0000][>P8000YW
>P03001S6f<K00=S6`8000aS6`03001Jo5[l00AJo08000UK7003001Fo5Kl00aFo0<000mFo003001F
o5Kl00QFo0@00003E_`005Km00iFo@03001FoEKm00AFo@@000EFo@8002moo`00;Goo0`00:goo00<0
06/jJcX00f/j00<006/jJcX01f/j00<006LjIcX02VLj00<006<KHa/00V<K00<006<KHa/036<K00<0
05[lF_`00U[l0P002e/L00<005KlE_`02UKl0P000U[l00<005KlE_`03EKl00<005KlE_`015Kl1000
0eKl00<005Kl000045Km00<005KmE_d00`002UKm0P00;Woo000ZOol30003Ool2000XOol00`00JcY[
>P05JcX00`00JcY[>P05JcX00`00IcYW>P0<IcX01000Ha/0000046<K00@005[l000000iK7003001F
o5Kl00IFo0<000AJo003001Fo5Kl00iFo003001Fo5Kl00<000MFo004001Fo5Kl000?E_d3000=E_d2
000^Ool002Qoo`8002moo`03001[>V/j00M[>P03001[>V/j00=[>P03001W>VLj00iW>P03001S6f<K
015S6`03001K75/L00mK7003001Fo5Kl009Fo0<000QJo003001Fo5Kl00eFo0<000YFo005001Fo5Kl
E_`0000<E_d20002E_`00`00E_eFo@0<E_d2000]Ool002Eoo`<002moo`8000][>P05001[>V/jJcX0
000@IcX00`00IcX0000@Ha/200000f<K001K700?Fa`00`00E_aFo002000<F_`00`00E_aFo00:E_`2
0002E_`00`00E_aFo008E_`00`00E_aFo002E_`00`00E_eFo@06E_d30004E_`00`00E_eFo@0=E_d2
000/Ool0029oo`<0035oo`03001[>V/j00a[>P03001[>P00011W>P05001W>VLjIcX0000>Ha/00`00
Ha]S6`02Ha/00`00FaaK700>Fa`2000?F_`00`00E_aFo006E_`30004E_`00`00E_aFo008E_`00`00
E_aFo002E_`00`00E_eFo@03E_d30008E_`00`00E_eFo@0<E_d2000/Ool0021oo`8000=oo`03001o
ogoo02eoo`03001[>V/j00i[>P03001W>VLj00eW>P8000EW>P03001S6f<K00US6`8000IS6`03001K
75/L00]K70800003Fa`005[l00iJo003001Fo5Kl00=Fo0<000QFo003001Fo5Kl00MFo003001Fo5Kl
00=Fo003001FoEKm008000]Fo003001FoEKm00aFo@<002]oo`007Goo0`00=7oo00<006/jJcX03V/j
00<006/j00003FLj00<006LjIcX01VLj00<006<KHa/01f<K00<006<KHa/01V<K00<005/LFa`02E/L
0P0015/L00<005[lF_`03E[l00<005KlE_`00P0035Kl00<005KlE_`01EKl0P001EKl0`003UKl00<0
05KmE_d02EKm0P000UKm0P00:Woo000KOol2000fOol00`00JcY[>P0>JcX01@00JcY[>V/j00002fLj
00<006LjIcX026Lj00<006<KHa/016<K0P002V<K00<005/LFa`01U/L0P001e/L00<005[lF_`035[l
0`003UKl00<005KlE_`00UKl0`000UKl00<005KlE_`00P000UKl00<005KlE_`035Kl00<005KmE_d0
1UKm0`0015Km0P00:Woo000IOol2000gOol00`00JcY[>P0>JcX00`00JcY[>P03JcX00`00IcYW>P07
IcX00`00IcYW>P0:IcX00`00Ha]S6`02Ha/00`00Ha]S6`0;Ha/00`00FaaK7003Fa`20009Fa`00`00
F_aJo00:F_`20002F_`00`00E_aFo00=E_`40005E_`30004E_`00`00E_aFo00=E_`00`00E_eFo@02
E_d30008E_d2000YOol001Uoo`8003Ioo`03001[>V/j00i[>P03001[>V/j00E[>P8000IW>P03001W
>VLj00aW>P04001S6`00000?Ha/00`00FaaK7002000<Fa`00`00F_aJo007F_`20005F_`00`00E_aF
o00:E_`40004E_`300000eKl001Fo005E_`00`00E_aFo00<E_`00`00E_eFo@02000<E_d20008Ool2
0004Ool20003Ool3000BOol001Qoo`800003Ool007oo03=oo`03001[>P0000i[>P03001[>V/j00Q[
>P04001W>VLjIcX2000@IcX00`00Ha]S6`0@Ha/2000?Fa`00`00F_aJo003F_`30008F_`00`00E_aF
o007E_`20003E_`00`00E_aFo0020005E_`00`00E_aFo004E_`00`00E_aFo00;E_`3000?E_d00`00
Oomoo`05Ool01000Oomoo`0027oo00@007ooOol0015oo`006Goo00@007ooOol0039oo`05001[>V/j
JcX0000;JcX2000<JcX00`00IcX0000AIcX00`00Ic/0000?Ha/200000elK001K700?Fa`00`00F_aJ
o002000;F_`00`00E_aFo004E_`30005E_`30008E_`00`00E_aFo004E_`00`00E_aFo007E_`30002
E_`00`00E_eFo@0=E_d20006Ool01000Oomoo`0027oo00@007ooOol0015oo`004Goo0P001Goo0P00
0goo00<007ooOol0;Woo00<006/jJcX00f/j00<006/jJcX01f/j00<006/jJcX03F/j00<006LjIcX0
3VLj0P000fLk00<006<KHa/02f<K00<005lKGa/00UlK00<005/LFa`03U/L0P003U[l00<005KlE_`0
0`001EKl0`000UKl00<005KlE_`01UKl00<005KlE_`015Kl00<005KlE_`015Kl0`001EKl00<005Km
E_d03UKm00<007ooOol017oo00@007ooOol000Qoo`<0019oo`0047oo00@007ooOol000Aoo`03001o
ogoo00=oo`03001oogoo02aoo`03001[>V/j00E[>P03001[>V/j00E[>P03001[>V/j00e[>P03001[
>P0000iW>P03001W>fLk00AW>`03001S6f<K00QS6`8000EO6`03001K75/L00aK70800003Fa`005[l
00eJo0<000EFo0<000EFo003001Fo5Kl00MFo003001Fo5Kl00AFo003001Fo5Kl00<000UFo003001F
oEKm00eFo@8000Eoo`04001oogoo0009Ool00`00Oomoo`0AOol0011oo`04001oogoo0004Ool00`00
Oomoo`04Ool00`00Oomoo`0ZOol00`00JcY[>P07JcX00`00JcY[>P03JcX00`00JcY[>P0=JcX01@00
JcY[>V/j000036Lj00<006LkIc/01FLk00<006<KHa/01f<K00<005lKGa/01UlK00<005/LFa`02E/L
0P000e/L00<005[lF_`02E[l0`000U[l00<005KlE_`00`002EKl00<005KlE_`01eKl00<005KlE_`0
0eKl0`003EKl00<005KmE_d02UKm0P000UKm00<007ooOol00goo0P002Woo0`004Goo000@Ool01000
Oomoo`0017oo0P001Woo00<007ooOol09goo0P002f/j00D006/jJcY[>P0000m[>P03001[>V/j00=[
>P03001W>VLj00MW>P8000UW>`03001S6f<K00AS6`8000YO6`03001K75/L00MK7003001K75/L00AK
7003001Jo5[l00EJo0<000EJo0<000eFo003001Fo5Kl00MFo003001Fo5Kl0080009Fo003001Fo5Kl
00aFo003001FoEKm00IFo@<000AFo@03001oogoo02=oo`0047oo00@007ooOol000Aoo`03001oogoo
00Ioo`03001oogoo02Eoo`03001[>V/j00a[>P03001[>P0000m[>P03001[>V/j00A[>P03001W>VLj
00IW>P03001W>fLk00YW>`03001S6f<K009S6`03001O6elK00]O6`03001K75/L00AK708000QK7004
001Jo5[lF_`30006F_`20002F_`00`00E_aFo00<E_`00`00E_aFo006E_`30004E_`00`00E_aFo00<
E_`00`00E_eFo@03E_d30008E_d00`00Oomoo`0ROol0015oo`8000Eoo`03001oogoo00Moo`8002Ao
o`03001[>V/j00i[>P03001[>V/j00e[>P03001[>V/j00I[>P03001W>VLj00AW>P03001W>fLk00aW
>`04001S6`00000>Ga/00`00FaaK7002Fa`2000;Fa`30006F_`30004F_`00`00E_aFo00<E_`00`00
E_aFo003E_`300000eKl001Fo005E_`00`00E_aFo00<E_`00`00E_eFo@02000;E_d2000SOol001Mo
o`8000]oo`03001oogoo021oo`03001[>V/j00i[>P03001[>P0000a[>P8000Y[>P04001W>VLjIcX2
000@Ic/00`00Ga]O6`0?Ga/01000Fa`000002e/L0`001E[l0`0025[l00<005KlE_`035Kl00<005Kl
E_`00P001EKl00<005KlE_`00eKl00<005KlE_`035Kl0`003UKm00<007ooOol08Goo000GOol00`00
Oomoo`0;Ool00`00Oomoo`0NOol00`00JcY[>P0>JcX01@00JcY[>V/j00002V/j00<006/jJcX02f/j
00<006Lj00004FLk00<006Lk00004ElK00<005/LFa`025/L0`000e/L00<005[lF_`00P0035[l00<0
05KlE_`02eKl0`001eKl00<005KlE_`015Kl00<005KlE_`025Kl0`000UKl00<005KmE_d03EKm00<0
07ooOol087oo000GOol00`00Oomoo`0<Ool00`00Oomoo`0LOol00`00JcY[>P0>JcX00`00JcY[>P03
JcX00`00JcY[>P06JcX00`00JcY[>P0=JcX00`00Ic]W>`0>Ic/20003Ic/00`00Ga]O6`0<Ga/20000
0elK001K7005Fa`30007Fa`2000?F_`00`00E_aFo007E_`30002E_`00`00E_aFo006E_`00`00E_aF
o004E_`00`00E_aFo003E_`40005E_`00`00E_eFo@0=E_d2000QOol001Moo`8000ioo`03001oogoo
01Yoo`03001[>V/j00i[>P03001[>V/j00E[>P8000E[>P03001[>V/j00e[>P03001W>P0000iW>`03
001W>fLk00AW>`03001O6elK00UO6`8000AO6`03001K75/L00<000QK70800003Fa`005[l00iJo003
001Fo5Kl00AFo0<000IFo003001Fo5Kl00IFo003001Fo5Kl00=Fo003001Fo5Kl00<000YFo003001F
oEKm00aFo@80025oo`005goo00<007ooOol03Woo00<007ooOol067oo00<006/j00003V/j00<006/j
JcX026/j00D006/jJcY[>P0000m[>P05001W>VLjIcX0000<Ic/00`00Ic]W>`05Ic/00`00Ga]O6`08
Ga/00`00Ga]O6`04Ga/30009Fa`20004Fa`00`00F_aJo00=F_`00`00E_aFo003000:E_`00`00E_aF
o006E_`00`00E_aFo002E_`3000>E_`00`00E_eFo@08E_d30002E_d00`00Oomoo`0NOol001Moo`03
001oogoo00moo`03001oogoo01Ioo`05001[>V/jJcX0000;JcX2000<JcX00`00JcX0000?JcX00`00
IcYW>P03IcX00`00Ic]W>`07Ic/20009Ic/00`00Ga]O6`05Ga/20004Ga/300000elK001K7007Fa`2
0007Fa`00`00F_aJo00<F_`3000=E_`00`00E_aFo006E_`00`00E_`000020002E_`00`00E_aFo00=
E_`00`00E_eFo@03E_d40005E_d2000OOol001Ioo`<0015oo`03001oogoo01Aoo`03001[>V/j00=[
>P03001[>V/j00M[>P03001[>V/j00e[>P03001[>V/j00e[>P03001W>VLj00EW>P8000IW>`03001W
>fLk00YW>`03001O6elK009O6`8000=O6`<000EO6`03001K75/L00=K708000UK7003001Jo5[l00YJ
o080009Jo003001Fo5Kl00aFo003001Fo5Kl00AFo0<000EFo003001Fo5Kl00eFo003001FoEKm00<0
00YFo@03001oogoo01eoo`005Woo00<007ooOol04Woo00<007ooOol04Woo00<006/jJcX01F/j00<0
06/jJcX01F/j00<006/jJcX03F/j00<006/j00003F/j00<006LjIcX026Lj00<006LkIc/00VLk00<0
06LkIc/036Lk00@005lK0000009O6`<000UO6`03001K75/L008000aK7003001Jo5[l00MJo08000AJ
o003001Fo5Kl00eFo003001Fo5Kl00<000=Fo003001Fo5Kl00=Fo003001Fo5Kl00aFo0<000eFo@80
01ioo`0027oo1@000Woo0P001Goo00<007ooOol04goo00<007ooOol047oo00<006/jJcX01f/j00<0
06/jJcX00f/j00<006/jJcX03F/j00D006/jJcY[>P0000Y[>P8000aW>P04001W>`00000@Ic/00`00
Ga/00002000=Ga/2000?Fa`00`00F_aJo003F_`30007F_`00`00E_aFo00<E_`30007E_`00`00E_aF
o003E_`00`00E_aFo009E_`20002E_`00`00E_eFo@0<E_d2000MOol000Yoo`05001oogooOol00002
Ool00`00Oomoo`02Ool2000EOol00`00Oomoo`0>Ool00`00JcY[>P09JcX01@00JcY[>V/j00003f/j
00<006/jJcX00f/j00<006/jJcX01V/j00<006LjIcX03FLj00<006LkIc/03fLk0`003UlK0P0000=O
6`00Fa`03e/L00<005[lF_`00P002e[l00<005KlE_`02EKl0P000UKl00<005KlE_`01EKl00<005Kl
E_`015Kl00<005KlE_`01UKl0P0015Kl00<005KmE_d03EKm00<007ooOol06goo000:Ool01@00Oomo
ogoo00000Woo00<007ooOol00Woo00<007ooOol05Goo00<007ooOol02goo0P003F/j00<006/j0000
3f/j00<006/jJcX01F/j0P001F/j00<006LjIcX03FLj00<006Lj00003VLk0`000V<K00<005lKGa/0
2ElK0P0015lK00<005/LFa`03U/L0P003U[l00<005KlE_`01UKl0P0015Kl00<005KlE_`01UKl00<0
05KlE_`00eKl00<005KlE_`00eKl0`001eKl00<005KmE_d035Km0P0077oo000:Ool01@00Oomoogoo
00000Woo00<007ooOol00Woo00<007ooOol05Woo00<007ooOol02Goo00<006/jJcX03V/j00<006/j
JcX03F/j00<006/jJcX026/j00D006/jJcY[>P0000mW>P05001W>VLjIcX0000:Ic/30006Ha/00`00
Ga]O6`07Ga/00`00Ga]O6`05Ga/00`00FaaK700;Fa`200000e/L001Jo00=F_`00`00E_aFo003E_`3
0007E_`00`00E_aFo006E_`00`00E_aFo003E_`00`00E_aFo002000:E_`00`00E_eFo@0<E_d2000L
Ool000Yoo`05001oogooOol00002Ool01000Oomoogoo0P006Goo00<007ooOol01goo00<006/jJcX0
3V/j00<006/j000036/j0P0036/j00<006/j00003VLj0P001FLj00<006LkIc/016Lk0`0000=W>`00
Ha/01V<K00<005lKGa/01ElK0P0025lK00<005/LFa`02E/L0P000e/L00<005[lF_`03E[l00<005Kl
E_`00P002UKl00<005KlE_`01UKl00<005KlE_`00eKl0`003EKl00<005KmE_d025Km0`000UKm00<0
07ooOol06Goo0009Ool20004Ool20004Ool00`00Oomoo`0IOol00`00Oomoo`05Ool00`00JcY[>P0>
JcX01@00JcY[>V/j00002V/j00<006/jJcX03F/j00<006LjIcX036Lj00<006LjIcX01FLj00@006Lk
Ic]W>`<0009W>`8000US6`03001O6elK009O6`8000]O6`03001K75/L00IK708000IK7003001Jo5[l
00aJo0<000eFo003001Fo5Kl00IFo003001Fo0000080009Fo003001Fo5Kl00aFo003001FoEKm00=F
o@@000EFo@8000Qoo`8000Aoo`8000Aoo`8000Aoo`005Goo00<007ooOol06Woo00<007ooOol00goo
00<006/jJcX03V/j00<006/jJcX00f/j00<006/jJcX01V/j00<006/jJcX03F/j00<006/j000036Lj
00<006LjIcX01fLj0`0016Lk00<006<KHa/02V<K00@005lK000000iO6`03001K75/L00=K708000UK
7003001Jo5[l00UJo080009Jo003001Fo5Kl00aFo003001Fo5Kl00=Fo0@000EFo003001Fo5Kl00aF
o003001FoEKm00<000YFo@03001oogoo00Eoo`04001oogoo0008Ool01000Oomoo`000goo000EOol2
000LOol01@00Oomoogoo000046/j00<006/jJcX01F/j0P001F/j00<006/jJcX03F/j00D006/jJcY[
>P0000YW>P03001W>VLj00EW>P<00003IcX006Lk00=W>`03001S6f<K00aS6`03001O6elK00mO6`03
001K75/L008000aK7003001Jo5[l00IJo08000AJo003001Fo5Kl00aFo003001Fo5Kl00<000AFo003
001Fo5Kl00=Fo003001Fo5Kl00]Fo0<000iFo@03001oogoo00Aoo`04001oogoo0008Ool01000Oomo
o`000goo000EOol00`00Oomoo`0LOol00`00Ool0000@JcX00`00JcY[>P08JcX01@00JcY[>V/j0000
3V/j0P001F/j00<006LjIcX01VLj00<006LjIcX00fLj0`001FLj00@006Lk000000iS6`03001S6`00
015O6`8000iK7003001Jo5[l00=Jo0<000MJo003001Fo5Kl00]Fo0<000MFo003001Fo5Kl00AFo003
001Fo5Kl00MFo0<0009Fo003001FoEKm00eFo@03001oogoo00=oo`04001oogoo0009Ool20004Ool0
01Eoo`03001oogoo01eoo`03001[>P0000e[>P8000a[>P03001[>P0000i[>P03001[>V/j00I[>P03
001W>VLj00=W>P8000AW>P8000UW>P03001S6f<K00aS6`8000=S6`03001O6elK00aO6`800003Ga`0
05/L00iK7003001Jo5[l008000]Jo003001Fo5Kl00MFo0<0009Fo003001Fo5Kl00IFo003001Fo5Kl
00AFo003001Fo5Kl00=Fo0<000EFo003001FoEKm00eFo@03001oogoo00=oo`04001oogoo0008Ool0
1000Oomoo`000goo000DOol2000NOol01000JcY[>V/j0P002V/j00<006/jJcX03F/j00<006/jJcX0
36/j00<006/jJcX026/j00D006LjIcYW>P0000=W>P<000YW>P03001W>`0000aS6`03001S6f<K00AS
6`03001O6elK00UO6`8000AO7003001K75/L00aK70<000iJo003001Fo5Kl00=Fo0<000IFo003001F
o5Kl00IFo003001Fo5Kl00=Fo003001Fo5Kl00<000UFo003001FoEKm00aFo@8000Eoo`8000Yoo`80
00Aoo`0057oo00<007ooOol077oo00<006/jJcX016/j00<006/jJcX01V/j00<006/jJcX03F/j00<0
06/j000036/j00<006/jJcX02V/j00@006Lj001W>P<000aW>P05001W>fLkIc/0000:Ha/00`00Ha]S
6`05Ha/00`00Ga]O6`08Ga/00`00GaaO7005Ga`00`00FaaK7009Fa`20002Fa`00`00F_aJo00<F_`0
0`00E_aFo003000:E_`00`00E_aFo006E_`00`00E_aFo002E_`3000=E_`00`00E_eFo@08E_d30002
E_d00`00Oomoo`0DOol001Aoo`03001oogoo01]oo`03001[>V/j00I[>P03001[>V/j00A[>P03001[
>V/j00e[>P04001[>V/jJcX20009JcX00`00JcY[>P0<JcX3000=IcX20005Ic/00`00Ha]S6`05Ha/2
0009Ha/00`00Ga]O6`05Ga/20008Ga`00`00FaaK7007Fa`20004Fa`00`00F_aJo00<F_`3000=E_`0
0`00E_aFo006E_`00`00E_`000020002E_`00`00E_aFo00<E_`00`00E_eFo@03E_d40005E_d00`00
Oomoo`0DOol001Aoo`8001]oo`03001[>V/j00Q[>P8000=[>P03001[>V/j00e[>P03001[>V/j00A[
>P03001[>V/j00A[>P8000a[>P<00003JcX006Lj00aW>P03001W>fLk00EW>`03001S6f<K00AS6`03
001S6f<K00YS6`03001O6elK009O6`8000]O7003001K75/L00=K70<000MK7003001Jo5[l00UJo080
009Jo003001Fo5Kl00aFo003001Fo5Kl00AFo0<000EFo003001Fo5Kl00aFo003001FoEKm00<000YF
o@03001oogoo01=oo`0057oo00<007ooOol067oo0P003F/j00<006/j00003f/j00<006/jJcX01V/j
00<006/jJcX00V/j00<006/jJcX02F/j0`0000=[>P00JcX00V/j00<006LjIcX02FLj00<006LkIc/0
1fLk00<006<KHa/00V<K00<006<KHa/036<K00@005lK000000iO7003001K75/L008000]K7003001J
o5[l00IJo08000EJo003001Fo5Kl00aFo003001Fo5Kl00<000=Fo003001Fo5Kl00=Fo003001Fo5Kl
00]Fo0<000eFo@8001Aoo`001Goo10000goo0P001Woo00<007ooOol05goo00<006/jJcX03V/j00<0
06/jJcX03F/j00<006/jJcX026/j0P0000=[>P00JcX026/j0`000f/j00<006/jJcX00f/j00<006Lj
IcX01fLj00<006LkIc/02FLk00@006<K0000011S6`03001O75lL00mO708000iK7003001Jo5[l009J
o0<000QJo003001Fo5Kl00]Fo0<000IFo003001Fo5Kl00AFo003001Fo5Kl00MFo0<0009Fo003001F
oEKm00aFo@8001=oo`001Goo00<007ooOol00goo00@007ooOol000Aoo`8001Qoo`03001[>V/j00i[
>P04001[>P00000;JcX2000=JcX00`00JcY[>P05JcX30005JcX00`00JcY[>P05JcX20006IcX00`00
Ic]W>`0;Ic/00`00Ha]S6`0>Ha/200000f<K001O700=Ga`200000elL001K700=Fa`00`00F_aJo002
000;F_`00`00E_aFo008E_`30002E_`00`00E_aFo005E_`00`00E_aFo004E_`00`00E_aFo003E_`3
0005E_`00`00E_eFo@0<E_d2000COol000Ioo`03001oogoo009oo`04001oogoo0005Ool00`00Oomo
o`0EOol00`00JcY[>P0>JcX00`00JcY[>P02JcX00`00JcY[>P07JcX00`00JcY[>P0<JcX00`00JcX0
0004JcX20007JcX00`00JcY[>P08JcX01000IcYW>VLj0P003FLk00<006Lk00003V<K00<006<KHa/0
0V<K00<005lLGa`02ElL0P000elL00<005/LFa`03E/L0P003U[l00<005KlE_`015Kl0`001EKl00<0
05KlE_`01UKl00<005KlE_`00eKl00<005KlE_`00`002EKl00<005KmE_d02eKm0`004Woo0007Ool0
1@00Oomoogoo00000Woo00<007ooOol00Woo00<007ooOol05Goo00<006/jJcX03V/j00<006/jJcX0
16/j0P001V/j00<006/jJcX036/j00@006/jJcY[>P@000M[>P8000a[>P03001W>P0000eW>`8000=W
>`03001S6f<K00US6`8000IS6`03001O75lL00MO7003001O75lL00AO7003001K75/L00YK70800003
Fa`005[l00iJo003001Fo5Kl00<000UFo003001Fo5Kl00IFo003001Fo5Kl009Fo0<000eFo003001F
oEKm00MFo@<0009Fo003001oogoo011oo`001Goo00@007ooOol0009oo`04001oogoo0004Ool2000E
Ool00`00JcY[>P0>JcX00`00JcY[>P07JcX00`00JcY[>P02JcX00`00JcY[>P0<JcX00`00JcY[>P04
0008JcX00`00JcY[>P0=JcX00`00Ic]W>`0;Ic/00`00Ic]W>`04Ic/00`00Ha]S6`07Ha/00`00Ha]S
6`06Ha/00`00GaaO7005Ga`20008Ga`00`00FaaK7007Fa`20004Fa`00`00F_aJo00<F_`3000=E_`0
0`00E_aFo005E_`00`00E_`000020002E_`00`00E_aFo00<E_`00`00E_eFo@03E_d30005E_`2000A
Ool000Ioo`8000Aoo`8000Eoo`03001oogoo01=oo`03001[>V/j00i[>P03001[>V/j00U[>P800003
JcX006/j00e[>P@000A[>P8000E[>P03001[>V/j00e[>P03001[>P0000]W>`03001W>fLk00IW>`03
001S6f<K00AS6`8000YS6`03001O75lL009O708000]O7003001K75/L00AK708000MK7003001Jo5[l
00UJo080009Jo003001Fo5Kl00aFo003001Fo5Kl00=Fo0<000EFo003001Fo5Kl00aFo003001FoEKm
00<000UFo080011oo`004goo00<007ooOol04Woo00<006/j00003F/j0P003V/j00<006/jJcX02f/j
0`002F/j00D006/jJcY[>P0000i[>P8000=[>P03001W>fLk00MW>`03001W>fLk00QW>`03001S6f<K
009S6`03001S6f<K00]S6`04001O7000000=Ga`00`00FaaK7002Fa`20009Fa`00`00F_aJo007F_`2
0004F_`00`00E_aFo00<E_`00`00E_aFo0030003E_`00`00E_aFo003E_`00`00E_aFo00;E_`3000=
E_`00`00Oomoo`0>Ool0019oo`8001=oo`04001[>V/jJcX2000:JcX00`00JcY[>P0=JcX00`00JcX0
0009JcX4000<JcX00`00JcX0000>JcX00`00JcY[>P04JcX20006Ic/00`00Ic]W>`0:Ic/01000Ha/0
00003f<K00<005lLGa`03UlL00@005/L000000aK7003001Jo5[l00=Jo0<000MJo003001Fo5Kl00]F
o0<000MFo003001Fo5Kl00=Fo003001Fo5Kl00MFo0<0009Fo003001Fo5Kl00aFo003001oogoo00eo
o`004goo00<007ooOol047oo00<006/jJcX016/j00<006/jJcX01V/j00<006/jJcX03F/j00D006/j
JcY[>P0000E[>P<0009[>P03001[>V/j00a[>P03001[>V/j00a[>P03001[>V/j00M[>P04001W>fLk
Ic/2000>Ic/00`00Ha]S6`0=Ha/200000f<K001O700?Ga`00`00FaaK700=Fa`00`00F_aJo002000;
F_`00`00E_aFo007E_`30002E_`00`00E_aFo005E_`00`00E_aFo004E_`00`00E_aFo003E_`30005
E_`00`00E_aFo00<E_`00`00Oomoo`0=Ool0019oo`03001oogoo011oo`03001[>V/j00I[>P03001[
>V/j00A[>P03001[>V/j00e[>P03001[>V/j00=[>P03001[>P00008000A[>P03001[>V/j00a[>P03
001[>P0000][>P8000][>P03001W>`0000mW>`03001W>`0000eS6`03001S6f<K009S6`03001O75lL
00]O70800003Ga`005/L00eK70<000iJo003001Fo5Kl00=Fo0<000EFo003001Fo5Kl00IFo003001F
o5Kl00=Fo003001Fo5Kl00<000UFo003001Fo5Kl00]Fo08000ioo`004Woo0P0047oo00<006/jJcX0
26/j0P000f/j00<006/jJcX03F/j00<006/jJcX016/j0`001F/j00<006/jJcX036/j00D006/jJcY[
>P0000U[>P03001[>V/j00a[>P03001W>fLk00aW>`8000=W>`03001S6f<K00QS6`8000IS6`03001O
75lL00QO708000AO7003001K75/L00UK7080009K7003001Jo5[l00aJo003001Fo5Kl00<000UFo003
001Fo5Kl00IFo003001Fo5Kl009Fo0<000eFo003001Fo5Kl00MFo0<0009Fo003001oogoo00Eoo`D0
005oo`004Woo00<007ooOol03Goo0P003F/j00<006/j00003f/j00<006/jJcX00V/j0`000f/j00D0
06/jJcY[>P0000i[>P03001[>V/j00=[>P8000I[>P03001[>V/j00a[>P03001W>P0000aW>`03001W
>fLk00AW>`03001S6f<K00IS6`03001S6f<K00IS6`03001O75lL00IO708000IO7003001K75/L00MK
708000EK7003001Jo5[l00]Jo0<000eFo003001Fo5Kl00EFo003001Fo0000080009Fo003001Fo5Kl
00aFo003001Fo5Kl00=Fo0<000EFo003001oogoo00Moo`03001oogoo005oo`004Woo00<007ooOol0
37oo00<006/jJcX03V/j00<006/jJcX036/j0P000V/j0`001f/j00<006/j00003V/j00<006/jJcX0
1V/j00@006/jJcY[>P8000i[>P05001W>VLjIcX0000:Ic/00`00Ic]W>`05Ic/00`00Ha]S6`04Ha/2
000:Ha/00`00GaaO7003Ga`20009Ga`00`00FaaK7003Fa`30008Fa`00`00F_aJo008F_`20002Fa`0
0`00E_aFo00<E_`00`00E_aFo003E_`30005E_`00`00E_aFo00<E_`00`00E_aFo0030009E_`00`00
Oomoo`06Ool00`00Oomoo`01Ool0019oo`8000aoo`03001[>V/j00i[>P04001[>P00000;JcX00`00
JcX00002000;JcX00`00JcY[>P0<JcX00`00JcY[>P08JcX00`00JcX0000?JcX00`00IcYW>P03IcX0
0`00Ic]W>`05Ic/20009Ic/00`00Ha]S6`02Ha/00`00Ha]S6`0;Ha/00`00GaaO7002000<Ga`00`00
FaaK7002000;Fa`00`00F_aJo006F_`20004Fa`00`00E_aFo00<E_`00`00E_aFo0030003E_`00`00
E_aFo003E_`00`00E_aFo00;E_`3000<E_`20007Ool00`00Oomoo`01Ool0015oo`8000aoo`03001[
>V/j00i[>P03001[>V/j009[>P03001[>V/j00M[>P<000e[>P03001[>P0000][>P8000a[>P03001[
>V/j00e[>P03001W>VLj00AW>P03001W>fLk00AW>`03001W>fLk00YW>`04001S6`00000?Ha/2000?
Ga`2000>Fa`00`00F_aJo003F_`20007Fa`00`00E_aFo00;E_`30007E_`00`00E_aFo003E_`00`00
E_aFo007E_`30002E_`00`00E_aFo00;E_`00`00Oomoo`05Ool00`00Oomoo`01Ool000Aoo`8000Ao
o`8000Eoo`03001oogoo00Yoo`03001[>V/j00i[>P03001[>V/j00A[>P8000A[>P<000m[>P04001[
>V/jJcX20008JcX00`00JcY[>P0;JcX00`00JcX0000=JcX00`00IcYW>P06IcX00`00Ic]W>`02Ic/0
0`00Ic]W>`0<Ic/00`00Ha]S6`0=Ha/200000f<K001O700<Ga`200000elL001K700>Fa`00`00F_aJ
o002000:Fa`00`00E_aFo007E_`30002E_`00`00E_aFo005E_`00`00E_aFo004E_`00`00E_aFo003
E_`30005E_`00`00E_aFo00;E_`20005Ool20003Ool000=oo`04001oogoo0002Ool01000Oomoo`00
17oo0P002Woo00<006/jJcX03V/j00<006/jJcX01f/j100000=[>P00JcX03V/j00<006/jJcX016/j
00<006/jJcX016/j00<006/jJcX02f/j00D006/jJcY[>P0000Y[>P8000YW>P04001W>`00000=Ic/2
00000fLk001S6`0<Ha/00`00Ha]S6`02Ha/00`00GaaO7009Ga`00`00GaaO7002Ga`00`00FaaK700<
Fa`3000=Fa`00`00E_aFo003E_`30005E_`00`00E_aFo006E_`00`00E_aFo003E_`00`00E_aFo003
0009E_`00`00E_aFo00:E_`2000:Ool000Eoo`05001oogooOol00002Ool00`00Oomoo`02Ool00`00
Oomoo`08Ool00`00JcY[>P0>JcX00`00JcY[>P06JcX500000f/j001[>P0>JcX00`00JcY[>P06JcX0
0`00JcY[>P02JcX00`00JcY[>P0;JcX00`00JcY[>P03JcX00`00JcY[>P06JcX00`00IcYW>P0;IcX0
0`00Ic]W>`0<Ic/00`00Ic]W>`02Ic/00`00Ha]S6`08Ha/20006Ha/00`00GaaO7006Ga`20006Ga`0
0`00FaaK7009Fa`20002Fa`00`00FaaK700;Fa`00`00E_aFo0030009E_`00`00E_aFo006E_`00`00
E_aFo002E_`3000<E_`00`00E_aFo007E_`300000goo001oo`09Ool000Ioo`04001oogoo0002Ool0
0`00Oomoo`02Ool00`00Oomoo`07Ool00`00JcX0000=JcX20006JcX30005JcX00`00JcY[>P0=JcX0
0`00JcY[>P08JcX200000f/j001[>P0<JcX00`00JcY[>P05JcX20005JcX00`00IcYW>P0;IcX00`00
IcX0000<Ic/00`00Ic]W>`04Ic/20006Ha/20008Ha/00`00GaaO7004Ga`20008Ga`00`00FaaK7007
Fa`20004Fa`00`00FaaK700;Fa`3000=E_`00`00E_aFo006E_`40002E_`00`00E_aFo00;E_`00`00
E_aFo003E_`30004Ool00`00Oomoo`08Ool000=oo`04001oogoo0002Ool01000Oomoo`0017oo00<0
07ooOol01Woo00@006/jJcY[>P8000Y[>P03001[>V/j00A[>P8000M[>P03001[>P0000a[>P8000e[
>P03001[>V/j00][>P03001[>V/j00Q[>P05001[>V/jJcX0000=IcX01@00IcYW>VLj00002FLk0P00
2FLk00<006<KHa/00V<K00<006<KHa/02F<K00<005lLGa`00UlL00<005lLGa`02ElL00<005/LFa`0
0e/L0`001e/L00<005/LFa`025/L0P000U/L00<005KlE_`035Kl00<005KlE_`00eKl0`001EKl00<0
05KlE_`02eKl00<005KlE_`00`0027oo00<007ooOol01goo0004Ool20004Ool20004Ool20007Ool0
0`00JcY[>P04JcX00`00JcY[>P06JcX00`00JcY[>P02JcX30008JcX01@00JcY[>V/j00002V/j00<0
06/jJcX036/j00<006/j00002V/j0P0036/j00<006/j00003FLj00<006LjIcX00fLj0P001VLk00<0
06LkIc/02VLk00@006<K000000eS6`04001O7000000=Ga`00`00FaaK7002000;Fa`00`00FaaK7005
Fa`20005Fa`00`00E_aFo00;E_`00`00E_aFo0030003E_`00`00E_aFo003E_`00`00E_aFo00:E_`3
000;Ool00`00Oomoo`07Ool0015oo`03001oogoo00Aoo`03001[>V/j00I[>P03001[>V/j00A[>P03
001[>V/j00<000Y[>P03001[>V/j00=[>P03001[>V/j00I[>P03001[>V/j00a[>P05001[>V/jJcX0
0008JcX00`00JcY[>P0=JcX00`00IcYW>P0;IcX00`00IcYW>P06IcX00`00Ic]W>`02Ic/00`00Ic]W
>`0<Ic/00`00Ha]S6`0>Ha/00`00GaaO700>Ga`2000>Fa`00`00FaaK7002Fa`20008Fa`00`00E_aF
o00:E_`30007E_`00`00E_aFo003E_`00`00E_aFo006E_`3000>Ool00`00Oomoo`07Ool0011oo`03
001oogoo00Aoo`03001[>V/j00Q[>P8000=[>P@000a[>P03001[>V/j00E[>P8000E[>P03001[>V/j
00a[>P03001[>V/j00=[>P8000E[>P03001[>V/j00e[>P03001[>P0000YW>P8000YW>P04001W>`00
000=Ic/200000fLk001S6`0<Ha/200000f<K001O700<Ga`200000elL001K700=Fa`00`00FaaK7002
000:Fa`00`00E_aFo007E_`30002E_`00`00E_aFo005E_`00`00E_aFo004E_`00`00E_aFo003E_`2
000AOol00`00Oomoo`07Ool0011oo`8000=oo`8000e[>P@000i[>P03001[>V/j00Q[>P05001[>V/j
JcX0000>JcX00`00JcY[>P06JcX01@00JcY[>V/j00003f/j00D006/jJcY[>P0000QW>P03001W>VLj
00]W>P03001W>fLk00aW>`03001W>fLk009W>`03001S6f<K00US6`03001S6f<K009S6`03001O75lL
00QO708000AO7003001K75/L00]K70<000eK7003001Fo5Kl00AFo08000IFo003001Fo5Kl00EFo003
001Fo5Kl00=Fo003001Fo5Kl00<001=oo`03001oogoo00Moo`0047oo00D007ooOomoo`0000e[>P@0
00i[>P8000a[>P03001[>P0000i[>P03001[>V/j00Q[>P03001[>P0000i[>P8000E[>P8000EW>P03
001W>VLj00]W>P03001W>P0000]W>`8000IW>`03001S6f<K00IS6`8000IS6`03001O75lL00EO7080
00MO7003001K75/L00QK7080009K7003001K75/L00aK7003001Fo5Kl00<000QFo003001Fo5Kl00IF
o003001Fo5Kl009Fo0<001Ioo`03001oogoo00Moo`0047oo00@007ooOol000][>P<0009[>P04001[
>P00000;JcX00`00JcY[>P0=JcX00`00JcY[>P0;JcX2000<JcX00`00JcY[>P0<JcX00`00JcY[>P07
JcX01@00IcYW>VLj00003FLj00D006LjIcYW>P0000UW>`03001W>fLk00IW>`03001S6f<K00AS6`80
00US6`03001O75lL009O708000UO7003001K75/L00IK708000EK7003001K75/L00]K70<000aFo003
001Fo5Kl00IFo0@001Uoo`03001oogoo00Moo`003goo0P0000=oo`00JcX026/j0`0016/j00<006/j
JcX00V/j00<006/jJcX01f/j00<006/jJcX03F/j00<006/j00002f/j00<006/jJcX02f/j00<006/j
000036/j00<006/jJcX02F/j00<006Lj000036Lj0P001FLj00<006LkIc/016Lk0P002VLk00<006<K
Ha/00V<K00<006<KHa/02V<K00@005lL000000aO7003001K75/L00=K708000MK7003001K75/L00QK
70<0009K7003001Fo5Kl00]Fo004001Fo5KlE_`5000LOol00`00Oomoo`07Ool0011oo`8000Q[>P80
00I[>P03001[>V/j00A[>P8000I[>P03001[>V/j00e[>P04001[>V/jJcX20008JcX00`00JcY[>P0;
JcX01000JcY[>V/j0P002F/j00<006/jJcX02f/j00<006LjIcX02VLj00<006LjIcX01VLj00<006Lk
Ic/00VLk00<006LkIc/02fLk00@006<K000000iS6`03001O75lL00eO7003001K75/L008000YK7003
001K75/L00AK70<000IK7003001Fo5Kl00YFo0@000Eoo`03001oogoo01Yoo`03001oogoo00Ioo`00
3goo0P001V/j0`001f/j00<006/jJcX01f/j00<006/jJcX00V/j00<006/jJcX03F/j00<006/jJcX0
16/j00<006/jJcX016/j00<006/jJcX02f/j00<006/jJcX016/j00<006/jJcX01F/j00<006/jJcX0
2f/j00<006/j00002VLj00<006LjIcX026Lj00@006Lk000000mW>`03001S6f<K00aS6`800003Ha/0
05lL00eO70<000eK7003001K75/L00<000UK7003001Fo5Kl00MFo0<000Yoo`03001oogoo01Uoo`03
001oogoo00Ioo`003goo00<006/jJcX00V/j0`002F/j00<006/jJcX02F/j0P0000=[>P00JcX03V/j
00<006/jJcX01V/j00<006/jJcX00V/j00<006/jJcX02f/j00<006/jJcX01V/j00<006/jJcX00V/j
0P003F/j00D006/jJcY[>P0000QW>P03001W>VLj00YW>P03001W>fLk00eW>`800003Hc/006<K00]S
6`03001S6f<K009S6`03001O75lL00UO7080009O7003001K75/L00]K70<000eK7003001Fo5Kl00AF
o08000ioo`03001oogoo01Qoo`03001oogoo00Ioo`003goo1@002V/j0P003V/j00<006/jJcX03F/j
00<006/jJcX026/j0P0000=[>P00JcX036/j00<006/jJcX026/j0P0000=[>P00JcX036/j0P001F/j
00<006LjIcX016Lj00<006LjIcX02VLj00<006Lk00003FLk00<006<kHc/00V<k00<006<KHa/01f<K
0P001F<K00<005lLGa`01elL0P001ElL00<005/LFa`025/L0P000U/L00<005/LFa`035/L00<005Kl
E_`00`0047oo00<007ooOol067oo00<007ooOol01Woo0003Ool30002Ool20005Ool4000:JcX00`00
JcY[>P0=JcX01000JcX000002f/j0P003F/j00<006/jJcX02f/j00<006/jJcX02f/j00<006/jJcX0
2f/j00<006/jJcX01V/j00@006LjIcYW>P8000aW>P05001W>fLkIc/0000;Ic/00`00Hc]S>`04Hc/0
0`00Ha]S6`04Ha/20008Ha/00`00GaaO7003Ga`30007Ga`00`00FaaK7006Fa`20004Fa`00`00FaaK
700<Fa`3000DOol00`00Oomoo`0GOol00`00Oomoo`06Ool000Aoo`04001oogoo0002Ool00`00Oomo
o`03Ool01000Oomoo`0026/j00<006/jJcX03F/j00<006/jJcX00V/j00<006/jJcX01f/j00<006/j
JcX036/j00<006/j00002f/j00<006/jJcX02f/j00<006/j00002f/j00<006/jJcX026/j00<006Lj
00003FLj00<006LkIc/00fLk00<006LkIc/01VLk0P0026<k00<006<KHa/00V<K00<006<KHa/02F<K
00<005lLGa`00P002elL00<005/LFa`00e/L0P001e/L00<005/LFa`02E/L0P0067oo00<007ooOol0
5Woo00<007ooOol01Woo00000goo0000000300001Goo001oogoo00001Woo0P0000=oo`00JcX01F/j
00<006/jJcX03F/j00<006/jJcX016/j0P001V/j00<006/jJcX036/j00@006/jJcY[>P8000Q[>P03
001[>V/j00][>P05001[>V/jJcX00009JcX00`00JcY[>P0:JcX00`00IcYW>P0;IcX00`00Ic]W>`05
Ic/00`00Ic]W>`04Ic/00`00Hc]S>`09Hc/01000Ha/000003F<K0P003UlL00<005/LFa`00P002U/L
00<005/LFa`01U/L0P006Woo00<007ooOol05Woo00<007ooOol01Woo00001Goo001oogoo00000Woo
00@007ooOol000Qoo`04001oo`000003JcX00`00JcY[>P0=JcX00`00JcY[>P07JcX00`00JcY[>P02
JcX00`00JcY[>P0<JcX00`00JcY[>P04JcX00`00JcY[>P04JcX00`00JcY[>P0;JcX00`00JcY[>P03
JcX20006JcX00`00JcY[>P0:JcX00`00JcX0000:IcX20009Ic/00`00Ic]W>`02Ic/00`00Hc]S>`0;
Hc/00`00Ha]S6`0;Ha/200000elK001O700>Ga`2000=Fa`00`00FaaK7002Fa`3000MOol00`00Oomo
o`0FOol00`00Oomoo`05Ool0009oo`03001oo`00009oo`04001oogoo0009Ool01P00Oomoo`00JcX0
00m[>P03001[>V/j00U[>P800003JcX006/j00e[>P03001[>V/j00I[>P03001[>V/j009[>P03001[
>V/j00][>P03001[>V/j00I[>P04001[>V/jJcX2000<JcX01@00JcY[>V/j000026Lj00<006LkIc/0
2VLk00@006Lk000000aS>`800003Hc/006<K00YS6`03001O6elK009O6`03001O75lL00YO70800003
Ga`005/L00aK7003001K75/L0080025oo`03001oogoo01Eoo`03001oogoo00Eoo`000goo0P000goo
0P002goo00@007ooOol000i[>P8000i[>P03001[>V/j00a[>P03001[>V/j00Q[>P800003JcX006/j
00a[>P03001[>V/j00Q[>P03001[>P0000e[>P03001[>V/j00=[>P8000EW>P03001W>fLk00aW>`03
001S>f<k00]S>`03001S>f<k009S>`03001S6f<K00IS6`8000IO6`03001O75lL00MO708000AO7003
001K75/L00YK70<002Aoo`03001oogoo01Aoo`03001oogoo00Eoo`005Woo00<007ooOol00P002f/j
00<006/jJcX03F/j00<006/j00002f/j0P003F/j00<006/jJcX02V/j0P0036/j00<006/jJcX02f/j
00<006/jJcX01V/j00D006LjIcYW>P0000eW>`800003Ic/006<k00US>`8000ES>`03001S6f<K00AS
6`8000QO6`03001O75lL00EO708000IO7003001K75/L00MK70<002Moo`03001oogoo01Aoo`03001o
ogoo00Eoo`005goo0P000Woo00<006/jJcX01f/j00<006/jJcX03F/j00@006/jJcY[>P8000Q[>P03
001[>V/j00a[>P04001[>P000009JcX00`00JcY[>P0;JcX00`00JcX0000:JcX2000:JcX00`00IcX0
000=Ic/00`00Ic]W>`02Ic/00`00Hc]S>`05Hc/20008Hc/00`00Ha]S6`02Ha/00`00Ga]O6`09Ga/0
0`00GaaO7002Ga`20009Ga`00`00FaaK7004Fa`2000[Ool00`00Oomoo`0COol00`00Oomoo`05Ool0
01Uoo`03001oogoo008000I[>P03001[>V/j00e[>P03001[>V/j00A[>P03001[>V/j00A[>P03001[
>V/j00a[>P03001[>V/j009[>P03001[>V/j00E[>P03001[>V/j00][>P05001[>V/jJcX00008JcX0
0`00JcY[>P0;JcX00`00Ic]W>`0:Ic/20006Ic/20004Hc/00`00Hc]S>`09Hc/01000Ha/000003ElK
00@005lL000000aO7003001K75/L00<002ioo`03001oogoo019oo`03001oogoo00Eoo`006Woo00D0
07ooOomoo`0000A[>P03001[>V/j00e[>P03001[>V/j00I[>P03001[>V/j009[>P03001[>V/j00a[
>P03001[>V/j00A[>P8000A[>P03001[>V/j00][>P03001[>V/j00=[>P8000E[>P03001[>V/j00][
>P03001W>P0000YW>`03001W>fLk00QW>`04001S>`00000=Hc/00`00Ga]O6`0>Ga/00`00GaaO700=
Ga`2000bOol00`00Oomoo`0AOol00`00Oomoo`05Ool001]oo`04001oogooOol200000f/j001[>P0>
JcX00`00JcY[>P08JcX200000f/j001[>P0=JcX00`00JcY[>P07JcX200000f/j001[>P0<JcX00`00
JcY[>P06JcX01@00JcY[>V/j00003F/j00D006LjIcYW>P0000MW>`8000aW>`03001S>f<k00]S>`80
0003Hc/005lK00aO6`800003Ga/005lL00]O708003Aoo`03001oogoo015oo`03001oogoo00Eoo`00
77oo0P000goo00<006/jJcX036/j0P003F/j00<006/jJcX036/j00<006/jJcX02V/j00<006/jJcX0
2f/j00<006/jJcX026/j00<006/j00003F/j00<006LjIcX00fLj0P0016Lk00<006LkIc/02fLk00<0
06Lk00002f<k00<006<kHc/00V<k00<005lKGa/025lK0P0015lK00<005lLGa`01elL0P00=goo00<0
07ooOol04Goo00<007ooOol017oo000NOol01000Oomoogoo0P002f/j00<006/jJcX036/j00<006/j
00002f/j0P0036/j00<006/j00002V/j0P0036/j00<006/jJcX02f/j00<006LjIcX01VLj00@006Lk
000000aW>`8000=W>`03001S>f<k00IS>`8000IS>`03001O6elK00EO6`8000IO6`03001O75lL00EO
708003Yoo`03001oogoo011oo`03001oogoo00Aoo`007goo00<007ooOol00Woo00<006/jJcX01f/j
00<006/jJcX036/j00@006/jJcY[>P8000Q[>P03001[>V/j00][>P05001[>V/jJcX00008JcX00`00
JcY[>P0;JcX00`00JcX0000:JcX2000:IcX00`00Ic]W>`0;Ic/00`00Ic]W>`03Ic/00`00Hc]S>`04
Hc/20009Hc/00`00Ga]O6`02Ga/20009Ga/00`00GaaO7002Ga`2000lOol00`00Oomoo`0@Ool00`00
Oomoo`04Ool0021oo`03001oogoo009oo`8000I[>P03001[>V/j00a[>P03001[>V/j00A[>P03001[
>V/j00A[>P03001[>V/j00][>P03001[>V/j00=[>P8000E[>P03001[>V/j00][>P05001[>V/jJcX0
0008JcX00`00IcYW>P09IcX00`00IcX0000;Ic/00`00Ic]W>`05Ic/00`00Hc]S>`02Hc/00`00Hc]S
>`0:Hc/01000Ga/0000035lK00@005lL000003moo`03001oogoo00moo`03001oogoo00Aoo`008Goo
0P0017oo00<006/jJcX00V/j00<006/jJcX036/j00<006/jJcX01V/j00<006/jJcX00V/j00<006/j
JcX02f/j00<006/jJcX01V/j00D006/jJcY[>P0000e[>P03001[>V/j00=[>P8000E[>P03001W>VLj
00UW>P05001W>VLjIcX00009Ic/00`00Ic]W>`07Ic/01000Hc/000003V<k00<005lKGa/03ElK00<0
07ooOol0@7oo00<007ooOol03Woo00<007ooOol017oo000SOol00`00Oomoo`02Ool200000f/j001[
>P0=JcX00`00JcY[>P08JcX200000f/j001[>P0<JcX00`00JcY[>P08JcX00`00JcX0000=JcX00`00
JcY[>P06JcX01@00JcY[>V/j00002fLj00<006LjIcX00fLj0P001VLk00<006LkIc/02FLk00<006<k
Hc/036<k0P0000=S6`00Ga/02elK0P00A7oo00<007ooOol03Goo00<007ooOol017oo000TOol00`00
Oomoo`03Ool00`00JcY[>P0<JcX00`00JcY[>P0;JcX00`00JcY[>P0;JcX00`00JcY[>P0:JcX00`00
JcY[>P0;JcX00`00JcY[>P08JcX00`00JcX0000;IcX00`00IcYW>P06IcX01000Ic]W>fLk0P002fLk
00<006Lk000036<k00<006<KHa/00V<K00<005lKGa/01elK0P00AWoo00<007ooOol03Goo00<007oo
Ool017oo000UOol00`00Oomoo`03Ool2000;JcX00`00JcY[>P0;JcX00`00JcX0000;JcX00`00JcY[
>P0:JcX00`00JcX0000:JcX2000<JcX00`00IcYW>P08IcX2000:IcX00`00Ic/0000<Ic/01@00Ic]W
>fLk00002F<k0P001F<K00<005lKGa/01ElK0P00BGoo00<007ooOol03Goo00<007ooOol00goo000V
Ool00`00Oomoo`04Ool00`00JcY[>P07JcX00`00JcY[>P0;JcX01000JcY[>V/j0P0026/j00<006/j
JcX02V/j00D006/jJcY[>P0000Q[>P03001[>V/j00][>P03001[>P0000QW>P03001W>VLj00]W>P03
001W>fLk00YW>`03001W>fLk00=W>`03001S>f<k00AS>`8000QS6`03001O6elK009O6`8004aoo`03
001oogoo00aoo`03001oogoo00=oo`009goo0P001Goo0P001V/j00<006/jJcX02f/j00<006/jJcX0
16/j00<006/jJcX016/j00<006/jJcX02V/j00<006/jJcX00f/j0P001F/j00<006/jJcX02f/j00D0
06/jJcY[>P0000IW>P03001W>VLj00]W>P03001W>P0000YW>`03001W>fLk00EW>`03001S>f<k009S
>`03001S6f<K00US6`04001O6`00001?Ool00`00Oomoo`0;Ool00`00Oomoo`03Ool002Uoo`03001o
ogoo00Aoo`03001[>V/j009[>P03001[>V/j00][>P03001[>V/j00I[>P03001[>V/j009[>P03001[
>V/j00Y[>P03001[>V/j00I[>P05001[>V/jJcX0000<JcX20005JcX20003IcX00`00IcYW>P0:IcX2
0003IcX00`00Ic]W>`05Ic/20009Ic/01000Hc/000003F<K00<007ooOol0Cgoo00<007ooOol02goo
00<007ooOol00goo000ZOol00`00Oomoo`04Ool200000f/j001[>P0<JcX00`00JcY[>P08JcX20000
0f/j001[>P0;JcX00`00JcY[>P08JcX00`00JcX0000<JcX00`00JcY[>P07JcX00`00IcX0000<IcX0
0`00IcYW>P04IcX00`00Ic]W>`03Ic/00`00Ic]W>`0:Ic/00`00Ha]S6`0;Ha/2001COol00`00Oomo
o`0:Ool00`00Oomoo`03Ool002]oo`03001oogoo00Eoo`03001[>V/j00Y[>P8000e[>P03001[>V/j
00Y[>P03001[>V/j00Y[>P03001[>V/j00Y[>P03001[>V/j00U[>P03001W>VLj00YW>P03001W>VLj
00IW>P05001W>fLkIc/0000;Ic/200000fLk001S6`09Ha/2001FOol00`00Oomoo`09Ool00`00Oomo
o`03Ool002aoo`8000Ioo`8000U[>P03001[>V/j00a[>P04001[>P000009JcX00`00JcY[>P0:JcX0
1000JcX000002F/j00<006/jJcX02F/j00@006/j000000UW>P03001W>VLj00QW>P03001W>`0000]W
>`03001W>fLk009W>`03001S6f<K00IS6`03001oogoo05Ioo`03001oogoo00Uoo`03001oogoo00=o
o`00;Woo00<007ooOol01Goo00<006/jJcX01F/j00<006/jJcX036/j00<006/jJcX00V/j00<006/j
JcX01F/j00<006/jJcX02V/j00<006/jJcX00V/j00<006/jJcX01F/j00<006/jJcX02F/j00<006/j
JcX00V/j00<006LjIcX01FLj00<006LjIcX02VLj00<006LkIc/02FLk00<006LkIc/016Lk0P0016<K
0P00FWoo00<007ooOol027oo00<007ooOol00goo000_Ool00`00Oomoo`05Ool20004JcX00`00JcY[
>P0<JcX00`00JcY[>P04JcX20004JcX00`00JcY[>P0:JcX00`00JcY[>P04JcX00`00JcY[>P02JcX2
000;JcX00`00JcY[>P04JcX00`00IcYW>P02IcX2000<IcX00`00IcX00008Ic/20009Ic/01000Ha/0
0000GGoo00<007ooOol027oo00<007ooOol00Woo000`Ool00`00Oomoo`06Ool200000f/j001[>P0=
JcX00`00JcY[>P07JcX200000f/j001[>P0;JcX00`00JcY[>P06JcX200000f/j001[>P0;JcX00`00
JcY[>P06JcX200000fLj001W>P0;IcX20003IcX00`00Ic]W>`04Ic/00`00Ic]W>`0:Ic/00`00Oomo
o`1NOol00`00Oomoo`07Ool00`00Oomoo`02Ool0035oo`8000Qoo`03001[>V/j00a[>P03001[>V/j
00Y[>P03001[>V/j00U[>P8000][>P03001[>V/j00U[>P8000][>P03001W>VLj00YW>P03001W>VLj
00AW>P03001W>fLk009W>`03001W>fLk00UW>`80065oo`03001oogoo00Moo`03001oogoo009oo`00
<goo00<007ooOol01Woo00<006/jJcX02F/j0P0036/j00@006/j000000Q[>P03001[>V/j00Y[>P04
001[>P000008JcX00`00JcY[>P0:JcX00`00JcX0000:IcX00`00IcYW>P06IcX01000Ic/000002fLk
00<007ooOol0HWoo00<007ooOol01Woo00<007ooOol00Woo000dOol00`00Oomoo`06Ool20008JcX0
0`00JcY[>P0;JcX00`00JcY[>P02JcX00`00JcY[>P04JcX00`00JcY[>P0:JcX00`00JcY[>P02JcX0
0`00JcY[>P04JcX00`00JcY[>P0:JcX01@00JcY[>V/j000026Lj00<006LjIcX026Lj00<006LkIc/0
2VLk00<007ooOol0I7oo00<007ooOol01Goo00<007ooOol00Woo000eOol00`00Oomoo`07Ool00`00
JcY[>P04JcX00`00JcY[>P0;JcX00`00JcY[>P04JcX00`00JcY[>P02JcX00`00JcY[>P0:JcX00`00
JcY[>P04JcX00`00JcY[>P02JcX00`00JcY[>P0:JcX00`00JcY[>P03JcX00`00IcYW>P04IcX00`00
IcYW>P08IcX00`00Ic/00009Ic/2001XOol00`00Oomoo`04Ool00`00Oomoo`02Ool003Ioo`03001o
ogoo00Moo`03001[>V/j009[>P03001[>V/j00][>P03001[>V/j00I[>P800003JcX006/j00][>P03
001[>V/j00I[>P800003JcX006/j00][>P03001[>V/j00E[>P04001W>VLjIcX2000:IcX01@00Ic]W
>fLk00001fLk00<007ooOol0J7oo00<007ooOol017oo00<007ooOol00Woo000gOol20008Ool20000
0f/j001[>P0<JcX00`00JcY[>P09JcX00`00JcY[>P09JcX2000;JcX00`00JcY[>P09JcX20009JcX0
0`00IcX0000;IcX00`00Ic]W>`03Ic/20004Ic/00`00Oomoo`1ZOol00`00Oomoo`04Ool00`00Oomo
o`01Ool003Uoo`03001oogoo00Moo`03001[>V/j00Y[>P8000][>P04001[>P000008JcX00`00JcY[
>P0:JcX01000JcX0000026/j00<006/jJcX02V/j00<006LjIcX02FLj00<006LkIc/01VLk00@006Lk
000006ioo`03001oogoo00=oo`03001oogoo005oo`00>Woo00<007ooOol01goo0P002F/j00<006/j
JcX02V/j00<006/jJcX00V/j00<006/jJcX016/j00<006/jJcX02V/j00<006/jJcX00V/j00<006/j
JcX016/j00<006/jJcX02V/j00<006/j000026Lj0P002VLk00<007ooOol0KWoo00<007ooOol00goo
00<007ooOol00Goo000kOol00`00Oomoo`08Ool00`00JcY[>P05JcX00`00JcY[>P0:JcX00`00JcY[
>P04JcX00`00JcY[>P02JcX00`00JcY[>P0:JcX00`00JcY[>P04JcX00`00JcY[>P02JcX00`00JcY[
>P09JcX20003JcX00`00IcYW>P04IcX00`00Ic]W>`09Ic/00`00Oomoo`1`Ool00`00Oomoo`02Ool0
0`00Oomoo`01Ool003aoo`8000Uoo`8000A[>P03001[>V/j00Y[>P03001[>V/j00I[>P800003JcX0
06/j00][>P03001[>V/j00I[>P800003JcX006/j00Y[>P03001[>V/j00A[>P8000=W>P03001W>fLk
00UW>`03001oogoo079oo`05001oogooOol00003Ool003ioo`03001oogoo00Qoo`800003JcX006/j
00][>P03001[>V/j00U[>P03001[>V/j00U[>P8000][>P03001[>V/j00U[>P03001[>V/j00M[>P03
001W>P0000]W>`03001oogoo07Aoo`04001oogoo0003Ool003moo`03001oogoo00Uoo`03001[>V/j
00U[>P8000][>P04001[>P000008JcX00`00JcY[>P0:JcX00`00JcX00009JcX00`00JcY[>P09JcX0
0`00Ic]W>`09Ic/00`00Oomoo`1eOol01000Oomoo`000goo0010Ool00`00Oomoo`09Ool20008JcX0
0`00JcY[>P0:JcX00`00JcY[>P02JcX00`00JcY[>P04JcX00`00JcY[>P0:JcX01@00JcY[>V/j0000
1f/j00<006/jJcX02F/j00<006Lj000026Lk0P00NGoo00<007oo00000goo0011Ool2000;Ool00`00
JcY[>P04JcX00`00JcY[>P0:JcX00`00JcY[>P04JcX00`00JcY[>P02JcX00`00JcY[>P0:JcX00`00
JcY[>P03JcX20003JcX2000:JcX20003IcX00`00Ic]W>`04Ic/00`00Oomoo`1jOol00`00Ool00002
Ool004=oo`03001oogoo00Uoo`03001[>V/j009[>P03001[>V/j00Y[>P03001[>V/j00I[>P800003
JcX006/j00][>P03001[>V/j00I[>P03001[>P0000][>P03001W>VLj00AW>P8000=W>`03001oogoo
07aoo`80009oo`00A7oo00<007ooOol02Goo0P0000=[>P00JcX02f/j00<006/jJcX02F/j00<006/j
JcX02F/j0P002V/j00<006/jJcX02F/j00<006LjIcX01fLj00<006Lk0000Ogoo0P000Woo0015Ool0
0`00Oomoo`0:Ool00`00JcY[>P0:JcX00`00JcY[>P09JcX01000JcX0000026/j00<006/jJcX02F/j
00<006/j00002F/j00<006LjIcX02FLj00<007ooOol0OWoo0P000Woo0016Ool00`00Oomoo`0:Ool2
0009JcX00`00JcY[>P09JcX00`00JcY[>P02JcX00`00JcY[>P04JcX00`00JcY[>P09JcX01@00JcY[
>V/j00001f/j00<006LjIcX02FLj00<007ooOol0Ngoo100017oo0017Ool2000<Ool00`00JcY[>P05
JcX00`00JcY[>P09JcX00`00JcY[>P04JcX00`00JcY[>P02JcX00`00JcY[>P09JcX00`00JcY[>P03
JcX20003JcX2000:IcX2001jOol40008Ool004Uoo`03001oogoo00Yoo`8000A[>P03001[>V/j00U[
>P03001[>V/j00I[>P800003JcX006/j00Y[>P03001[>V/j00I[>P03001[>P0000]W>P03001oogoo
07Ioo`@000aoo`00BWoo00<007ooOol02goo0P0000=[>P00JcX02V/j00<006/jJcX02F/j00<006/j
JcX026/j0P002V/j00<006LjIcX02FLj00<007ooOol0Lgoo100047oo001;Ool00`00Oomoo`0<Ool0
0`00JcY[>P09JcX00`00JcY[>P09JcX00`00JcX00008JcX00`00JcY[>P09JcX00`00JcX00009IcX0
0`00Oomoo`1aOol3000DOol004aoo`8000eoo`8000Q[>P03001[>V/j00U[>P05001[>V/jJcX00006
JcX00`00JcY[>P09JcX01@00JcY[>V/j00001fLj00<007ooOol0KWoo10005goo001>Ool00`00Oomo
o`0<Ool00`00JcY[>P04JcX00`00JcY[>P09JcX00`00JcY[>P03JcX20003JcX00`00JcY[>P09JcX0
0`00JcY[>P03JcX20003IcX2001]Ool4000KOol004moo`03001oogoo00aoo`03001[>V/j009[>P03
001[>V/j00U[>P03001[>V/j00I[>P03001[>P0000][>P03001[>V/j00I[>P03001W>P0006]oo`@0
01moo`00D7oo00<007ooOol037oo0P0000=[>P00JcX02F/j0P002V/j00<006/jJcX026/j0P002V/j
00<007ooOol0IWoo10008goo001AOol2000>Ool00`00JcY[>P08JcX00`00JcY[>P09JcX00`00JcX0
0008JcX00`00JcY[>P09JcX00`00Oomoo`1TOol3000WOol005=oo`03001oogoo00aoo`8000M[>P03
001[>V/j00U[>P05001[>V/jJcX00006JcX00`00JcY[>P08JcX2001SOol4000ZOol005Aoo`03001o
ogoo00eoo`8000A[>P03001[>V/j00U[>P03001[>V/j00=[>P8000=[>P03001[>V/j00Q[>P03001o
ogoo05moo`@002ioo`00EGoo00<007ooOol03Woo0P0000=[>P00JcX02V/j00<006/jJcX01V/j00<0
06/j00002V/j00<007ooOol0G7oo1000<Woo001FOol00`00Oomoo`0?Ool00`00JcY[>P08JcX2000:
JcX00`00JcY[>P08JcX00`00Oomoo`1JOol3000fOol005Moo`80011oo`03001[>V/j00I[>P03001[
>V/j00U[>P04001[>P000007JcX00`00Oomoo`1GOol4000iOol005Uoo`03001oogoo00ioo`03001[
>V/j00A[>P03001[>V/j00U[>P03001[>V/j009[>P03001[>V/j009[>P8005Ioo`@003eoo`00FWoo
00<007ooOol03Woo0P000f/j00<006/jJcX02F/j00<006/jJcX016/j0P0000=[>P00Ool0Dgoo1000
@Goo001KOol00`00Oomoo`0?Ool00`00JcX0000;JcX00`00JcY[>P07JcX00`00Oomoo`1?Ool40015
Ool005aoo`80011oo`03001[>V/j00Q[>P8000U[>P03001oogoo04eoo`<004Uoo`00GWoo00<007oo
Ool03Woo0P001f/j00<006/jJcX026/j00<007ooOol0BWoo1000C7oo001OOol00`00Oomoo`0?Ool2
0004JcX00`00JcY[>P08JcX00`00Oomoo`17Ool4001@Ool0061oo`03001oogoo011oo`800003JcX0
06/j00U[>P03001oogoo04Aoo`@005Aoo`00HGoo0P004Woo00<006/jJcX026/j00<007ooOol0@Goo
1000F7oo001SOol00`00Oomoo`0@Ool00`00JcY[>P06JcX00`00Oomoo`0oOol3001LOol006Aoo`03
001oogoo011oo`03001[>V/j00A[>P03001oogoo03aoo`@005moo`00IGoo00<007ooOol047oo0P00
0f/j00<007ooOol0>Goo1000Hgoo001VOol00`00Oomoo`0AOol00`00JcX0000hOol4001WOol006Mo
o`80019oo`03001oogoo03=oo`@006]oo`00JGoo00<007ooOol03Woo00<007ooOol0<Goo0`00Kgoo
001ZOol00`00Oomoo`0=Ool00`00Oomoo`0]Ool4001bOol006]oo`03001oogoo00aoo`03001oogoo
02Uoo`@007Ioo`00K7oo0P0037oo00<007ooOol09Goo1000NWoo001^Ool00`00Oomoo`09Ool00`00
Oomoo`0ROol3001nOol006moo`03001oogoo00Qoo`03001oogoo01ioo`@0085oo`00L7oo00<007oo
Ool01goo00<007ooOol06Woo1000QGoo001aOol20007Ool00`00Oomoo`0FOol40029Ool007=oo`03
001oogoo00Aoo`03001oogoo019oo`@008eoo`00M7oo00<007ooOol00goo00<007ooOol03goo0`00
TGoo001eOol00`00Oomoo`02Ool00`00Oomoo`0;Ool4002DOol007Ioo`05001oogooOol00009Ool4
002HOol007Moo`800003Ool007oo00Aoo`@009aoo`00NGoo0P0000=oo`0000000P00X7oo001jOol2
002TOol00001\
\>"],
  ImageRangeCache->{{{0, 287}, {232.563, 0}} -> {-0.0532144, 0.0209325, \
0.00351807, 0.00351807}}],

Cell[BoxData[
    TagBox[\(\[SkeletonIndicator]  SurfaceGraphics  \[SkeletonIndicator]\),
      False,
      Editable->False]], "Output"]
}, Open  ]],

Cell["Same thing with delta:", "Text"],

Cell[BoxData[
    \(optdelta[s_, t_] := 
      Delta /. \ Keyfigures[EuropeanOption[110, t, PutOption], 
          BlackScholesModel[s, 0.10, 0.01, 0.15]]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Plot3D[optdelta[s, t], {s, 70, 140}, {t, 0.01, 1}, 
      AxesLabel \[Rule] {"\<s\>", "\<t\>", "\<delta\>"}, 
      PlotPoints \[Rule] 30]\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .81114 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% SurfaceGraphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
1.11022e-016 1.04977 -0.0679587 1.04977 [
[.11158 .21312 -12 -8.95035 ]
[.11158 .21312 0 .04965 ]
[.28836 .14356 -17.0282 -9 ]
[.28836 .14356 .97183 0 ]
[.47788 .06904 -15.9559 -9 ]
[.47788 .06904 2.04408 0 ]
[.68156 -0.01098 -14.8837 -9 ]
[.68156 -0.01098 3.11632 0 ]
[.29165 .07573 -9.42756 -13.1875 ]
[.29165 .07573 .69744 0 ]
[.76488 .0931 0 -6.1657 ]
[.76488 .0931 18 2.8343 ]
[.82636 .18723 0 -6.07503 ]
[.82636 .18723 18 2.92497 ]
[.88261 .27338 0 -5.99372 ]
[.88261 .27338 18 3.00628 ]
[.93428 .35251 0 -5.9204 ]
[.93428 .35251 18 3.0796 ]
[.98191 .42546 0 -5.85393 ]
[.98191 .42546 6 3.14607 ]
[.91861 .21225 0 -8.83749 ]
[.91861 .21225 10.125 4.35001 ]
[.02412 .26509 -12 -2.7423 ]
[.02412 .26509 0 6.2577 ]
[.01568 .31793 -30 -2.78698 ]
[.01568 .31793 0 6.21302 ]
[.00696 .37263 -24 -2.83338 ]
[.00696 .37263 0 6.16662 ]
[-0.00208 .42926 -30 -2.8816 ]
[-0.00208 .42926 0 6.1184 ]
[-0.01144 .48795 -6 -2.93176 ]
[-0.01144 .48795 0 6.06824 ]
[-0.05457 .39542 -34.625 -4.15171 ]
[-0.05457 .39542 0 9.03579 ]
[ 0 0 0 0 ]
[ 1 .81114 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.03716 .25514 m
.68874 0 L
s
.12091 .22234 m
.12558 .22696 L
s
[(80)] .11158 .21312 1 .98897 Mshowa
.2971 .15335 m
.30147 .15825 L
s
[(100)] .28836 .14356 .89202 1 Mshowa
.4859 .07942 m
.48992 .08461 L
s
[(120)] .47788 .06904 .77288 1 Mshowa
.68874 0 m
.69233 .00549 L
s
[(140)] .68156 -0.01098 .65374 1 Mshowa
.125 Mabswid
.16384 .20553 m
.1666 .20834 L
s
.20751 .18843 m
.21022 .19129 L
s
.25192 .17104 m
.25459 .17394 L
s
.34306 .13535 m
.34564 .13834 L
s
.38984 .11704 m
.39236 .12006 L
s
.43745 .0984 m
.43991 .10147 L
s
.53524 .06011 m
.53759 .06327 L
s
.58547 .04044 m
.58776 .04364 L
s
.63663 .0204 m
.63885 .02365 L
s
.07869 .23888 m
.08153 .2416 L
s
gsave
.29165 .07573 -70.4276 -17.1875 Mabsadd m
1 1 Mabs scale
currentpoint translate
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding ISOLatin1Encoding def
        currentdict
    end

    newfontname exch definefont pop
} def

0 21.1875 translate 1 -1 scale
63.000 13.313 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.313 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(s) show
69.125 13.313 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.25 Mabswid
.68874 0 m
.96935 .42924 L
s
.75258 .09765 m
.74642 .09993 L
s
[(0.2)] .76488 .0931 -1 .37015 Mshowa
.81397 .19156 m
.80778 .19373 L
s
[(0.4)] .82636 .18723 -1 .35001 Mshowa
.87016 .27751 m
.86393 .27958 L
s
[(0.6)] .88261 .27338 -1 .33194 Mshowa
.92177 .35646 m
.91551 .35844 L
s
[(0.8)] .93428 .35251 -1 .31564 Mshowa
.96935 .42924 m
.96306 .43113 L
s
[(1)] .98191 .42546 -1 .30087 Mshowa
.125 Mabswid
.70266 .0213 m
.69899 .02272 L
s
.71969 .04735 m
.71602 .04876 L
s
.73633 .0728 m
.73264 .07418 L
s
.76845 .12194 m
.76475 .12329 L
s
.78397 .14567 m
.78026 .147 L
s
.79914 .16888 m
.79543 .17019 L
s
.82848 .21376 m
.82476 .21504 L
s
.84267 .23547 m
.83895 .23674 L
s
.85656 .25671 m
.85283 .25797 L
s
.88347 .29787 m
.87973 .2991 L
s
.8965 .31781 m
.89275 .31902 L
s
.90926 .33733 m
.90551 .33853 L
s
.93402 .37521 m
.93027 .37638 L
s
.94603 .39358 m
.94227 .39473 L
s
.95781 .41158 m
.95404 .41273 L
s
gsave
.91861 .21225 -61 -12.8375 Mabsadd m
1 1 Mabs scale
currentpoint translate
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding ISOLatin1Encoding def
        currentdict
    end

    newfontname exch definefont pop
} def

0 21.1875 translate 1 -1 scale
63.000 13.313 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.313 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(t) show
69.125 13.313 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.25 Mabswid
.03716 .25514 m
0 .48963 L
s
.03634 .26031 m
.04245 .25793 L
s
[(-1)] .02412 .26509 1 -0.3906 Mshowa
.02795 .31327 m
.03408 .31093 L
s
[(-0.75)] .01568 .31793 1 -0.38067 Mshowa
.01926 .36807 m
.02542 .36579 L
s
[(-0.5)] .00696 .37263 1 -0.37036 Mshowa
.01027 .42482 m
.01644 .4226 L
s
[(-0.25)] -0.00208 .42926 1 -0.35964 Mshowa
.00095 .48363 m
.00715 .48147 L
s
[(0)] -0.01144 .48795 1 -0.3485 Mshowa
.125 Mabswid
.03468 .27076 m
.03835 .26933 L
s
.03302 .28128 m
.03669 .27986 L
s
.03134 .29187 m
.03501 .29045 L
s
.02965 .30253 m
.03333 .30112 L
s
.02624 .32408 m
.02992 .32268 L
s
.02451 .33496 m
.02819 .33357 L
s
.02277 .34592 m
.02646 .34454 L
s
.02102 .35696 m
.02471 .35558 L
s
.01749 .37926 m
.02118 .3779 L
s
.0157 .39053 m
.0194 .38918 L
s
.01391 .40188 m
.0176 .40053 L
s
.01209 .41331 m
.0158 .41197 L
s
.00843 .43642 m
.01214 .43509 L
s
.00658 .44809 m
.01029 .44678 L
s
.00472 .45985 m
.00843 .45854 L
s
.00284 .4717 m
.00656 .4704 L
s
gsave
-0.05457 .39542 -95.625 -8.15171 Mabsadd m
1 1 Mabs scale
currentpoint translate
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding ISOLatin1Encoding def
        currentdict
    end

    newfontname exch definefont pop
} def

0 21.1875 translate 1 -1 scale
63.000 13.313 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.313 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(delta) show
93.625 13.313 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.25 Mabswid
.03716 .25514 m
0 .48963 L
s
0 .48963 m
.39787 .81114 L
s
.39787 .81114 m
.40529 .59895 L
s
.40529 .59895 m
.03716 .25514 L
s
.68874 0 m
.96935 .42924 L
s
.96935 .42924 m
1 .6535 L
s
1 .6535 m
.70298 .24544 L
s
.70298 .24544 m
.68874 0 L
s
.03716 .25514 m
0 .48963 L
s
0 .48963 m
.70298 .24544 L
s
.70298 .24544 m
.68874 0 L
s
.68874 0 m
.03716 .25514 L
s
.40529 .59895 m
.96935 .42924 L
s
.96935 .42924 m
1 .6535 L
s
1 .6535 m
.39787 .81114 L
s
.39787 .81114 m
.40529 .59895 L
s
0 0 m
1 0 L
1 .81114 L
0 .81114 L
closepath
clip
newpath
.5 Mabswid
.636 .727 .914 r
.39444 .59726 .40499 .60743 .42265 .60358 .41218 .59318 Metetra
.618 .719 .918 r
.41218 .59318 .42265 .60358 .44043 .60049 .43002 .58982 Metetra
.595 .707 .922 r
.43002 .58982 .44043 .60049 .45832 .5984 .44799 .58742 Metetra
.563 .69 .926 r
.44799 .58742 .45832 .5984 .47636 .59753 .46608 .5862 Metetra
.524 .669 .929 r
.46608 .5862 .47636 .59753 .49455 .59804 .48433 .58638 Metetra
.478 .643 .93 r
.48433 .58638 .49455 .59804 .51291 .60003 .50275 .58806 Metetra
.428 .615 .928 r
.50275 .58806 .51291 .60003 .53149 .6035 .52137 .59127 Metetra
.378 .585 .924 r
.52137 .59127 .53149 .6035 .5503 .60836 .54022 .59593 Metetra
.332 .558 .918 r
.54022 .59593 .5503 .60836 .56937 .61439 .55933 .60187 Metetra
.294 .536 .913 r
.55933 .60187 .56937 .61439 .58872 .62133 .57872 .60879 Metetra
.268 .522 .909 r
.57872 .60879 .58872 .62133 .60835 .62883 .5984 .61637 Metetra
.257 .517 .908 r
.5984 .61637 .60835 .62883 .62828 .63655 .61837 .62424 Metetra
.259 .523 .911 r
.61837 .62424 .62828 .63655 .64849 .64412 .63864 .63201 Metetra
.274 .537 .917 r
.63864 .63201 .64849 .64412 .66898 .65123 .65919 .63935 Metetra
.299 .557 .925 r
.65919 .63935 .66898 .65123 .68972 .6576 .68001 .64597 Metetra
.332 .582 .934 r
.68001 .64597 .68972 .6576 .71071 .66302 .70107 .65163 Metetra
.37 .608 .942 r
.70107 .65163 .71071 .66302 .73191 .66736 .72237 .65618 Metetra
.408 .635 .948 r
.72237 .65618 .73191 .66736 .75331 .67055 .74386 .65954 Metetra
.446 .659 .952 r
.74386 .65954 .75331 .67055 .77489 .67256 .76554 .66169 Metetra
.481 .68 .954 r
.76554 .66169 .77489 .67256 .79664 .67343 .7874 .66266 Metetra
.511 .699 .954 r
.7874 .66266 .79664 .67343 .81854 .67324 .80941 .66252 Metetra
.538 .713 .953 r
.80941 .66252 .81854 .67324 .84058 .67207 .83157 .66137 Metetra
.56 .725 .951 r
.83157 .66137 .84058 .67207 .86277 .67002 .85387 .65931 Metetra
.579 .735 .949 r
.85387 .65931 .86277 .67002 .88509 .66722 .87631 .65647 Metetra
.594 .742 .947 r
.87631 .65647 .88509 .66722 .90756 .66376 .8989 .65295 Metetra
.606 .747 .945 r
.8989 .65295 .90756 .66376 .93017 .65976 .92163 .64887 Metetra
.615 .752 .943 r
.92163 .64887 .93017 .65976 .95292 .65529 .94451 .64431 Metetra
.623 .755 .941 r
.94451 .64431 .95292 .65529 .97582 .65044 .96753 .63937 Metetra
.628 .757 .939 r
.96753 .63937 .97582 .65044 .99888 .64528 .99072 .63411 Metetra
.639 .729 .913 r
.38377 .587 .39444 .59726 .41218 .59318 .40158 .5827 Metetra
.623 .721 .917 r
.40158 .5827 .41218 .59318 .43002 .58982 .41949 .57906 Metetra
.601 .71 .921 r
.41949 .57906 .43002 .58982 .44799 .58742 .43753 .57634 Metetra
.571 .694 .925 r
.43753 .57634 .44799 .58742 .46608 .5862 .45568 .57478 Metetra
.533 .674 .928 r
.45568 .57478 .46608 .5862 .48433 .58638 .47399 .5746 Metetra
.487 .648 .929 r
.47399 .5746 .48433 .58638 .50275 .58806 .49246 .57595 Metetra
.436 .619 .928 r
.49246 .57595 .50275 .58806 .52137 .59127 .51113 .57888 Metetra
.384 .588 .924 r
.51113 .57888 .52137 .59127 .54022 .59593 .53003 .58333 Metetra
.335 .559 .918 r
.53003 .58333 .54022 .59593 .55933 .60187 .54918 .58915 Metetra
.294 .535 .912 r
.54918 .58915 .55933 .60187 .57872 .60879 .5686 .59605 Metetra
.265 .519 .908 r
.5686 .59605 .57872 .60879 .5984 .61637 .58832 .6037 Metetra
.25 .512 .906 r
.58832 .6037 .5984 .61637 .61837 .62424 .60834 .61171 Metetra
.25 .516 .909 r
.60834 .61171 .61837 .62424 .63864 .63201 .62867 .61969 Metetra
.264 .529 .915 r
.62867 .61969 .63864 .63201 .65919 .63935 .64928 .62728 Metetra
.289 .55 .923 r
.64928 .62728 .65919 .63935 .68001 .64597 .67017 .63415 Metetra
.322 .575 .932 r
.67017 .63415 .68001 .64597 .70107 .65163 .69131 .64006 Metetra
.36 .603 .94 r
.69131 .64006 .70107 .65163 .72237 .65618 .71269 .64484 Metetra
.4 .63 .947 r
.71269 .64484 .72237 .65618 .74386 .65954 .73429 .64839 Metetra
.439 .656 .952 r
.73429 .64839 .74386 .65954 .76554 .66169 .75607 .65069 Metetra
.476 .678 .954 r
.75607 .65069 .76554 .66169 .7874 .66266 .77803 .65175 Metetra
.508 .697 .954 r
.77803 .65175 .7874 .66266 .80941 .66252 .80015 .65167 Metetra
.536 .713 .953 r
.80015 .65167 .80941 .66252 .83157 .66137 .82242 .65053 Metetra
.559 .725 .952 r
.82242 .65053 .83157 .66137 .85387 .65931 .84484 .64846 Metetra
.578 .735 .949 r
.84484 .64846 .85387 .65931 .87631 .65647 .86741 .64558 Metetra
.593 .742 .947 r
.86741 .64558 .87631 .65647 .8989 .65295 .89011 .642 Metetra
.606 .748 .945 r
.89011 .642 .8989 .65295 .92163 .64887 .91297 .63783 Metetra
.615 .752 .943 r
.91297 .63783 .92163 .64887 .94451 .64431 .93597 .63319 Metetra
.623 .755 .941 r
.93597 .63319 .94451 .64431 .96753 .63937 .95913 .62815 Metetra
.629 .757 .939 r
.95913 .62815 .96753 .63937 .99072 .63411 .98245 .62279 Metetra
.643 .73 .912 r
.37296 .57664 .38377 .587 .40158 .5827 .39084 .57212 Metetra
.628 .723 .916 r
.39084 .57212 .40158 .5827 .41949 .57906 .40883 .56822 Metetra
.607 .713 .919 r
.40883 .56822 .41949 .57906 .43753 .57634 .42694 .56518 Metetra
.579 .698 .923 r
.42694 .56518 .43753 .57634 .45568 .57478 .44517 .56327 Metetra
.542 .678 .927 r
.44517 .56327 .45568 .57478 .47399 .5746 .46353 .56272 Metetra
.497 .653 .929 r
.46353 .56272 .47399 .5746 .49246 .57595 .48206 .56372 Metetra
.445 .623 .928 r
.48206 .56372 .49246 .57595 .51113 .57888 .50078 .56634 Metetra
.392 .592 .924 r
.50078 .56634 .51113 .57888 .53003 .58333 .51972 .57056 Metetra
.34 .561 .918 r
.51972 .57056 .53003 .58333 .54918 .58915 .5389 .57623 Metetra
.295 .535 .911 r
.5389 .57623 .54918 .58915 .5686 .59605 .55837 .58309 Metetra
.263 .516 .906 r
.55837 .58309 .5686 .59605 .58832 .6037 .57813 .5908 Metetra
.244 .507 .904 r
.57813 .5908 .58832 .6037 .60834 .61171 .59819 .59895 Metetra
.241 .509 .906 r
.59819 .59895 .60834 .61171 .62867 .61969 .61857 .60715 Metetra
.253 .522 .912 r
.61857 .60715 .62867 .61969 .64928 .62728 .63924 .61499 Metetra
.277 .542 .92 r
.63924 .61499 .64928 .62728 .67017 .63415 .6602 .62213 Metetra
.311 .568 .93 r
.6602 .62213 .67017 .63415 .69131 .64006 .68143 .62831 Metetra
.35 .596 .939 r
.68143 .62831 .69131 .64006 .71269 .64484 .70289 .63333 Metetra
.392 .625 .946 r
.70289 .63333 .71269 .64484 .73429 .64839 .72458 .63708 Metetra
.433 .652 .951 r
.72458 .63708 .73429 .64839 .75607 .65069 .74647 .63953 Metetra
.47 .676 .954 r
.74647 .63953 .75607 .65069 .77803 .65175 .76853 .64071 Metetra
.504 .696 .955 r
.76853 .64071 .77803 .65175 .80015 .65167 .79077 .64068 Metetra
.533 .712 .954 r
.79077 .64068 .80015 .65167 .82242 .65053 .81315 .63957 Metetra
.557 .725 .952 r
.81315 .63957 .82242 .65053 .84484 .64846 .83569 .63748 Metetra
.577 .735 .95 r
.83569 .63748 .84484 .64846 .86741 .64558 .85838 .63454 Metetra
.593 .742 .947 r
.85838 .63454 .86741 .64558 .89011 .642 .88121 .6309 Metetra
.606 .748 .945 r
.88121 .6309 .89011 .642 .91297 .63783 .90419 .62665 Metetra
.616 .752 .943 r
.90419 .62665 .91297 .63783 .93597 .63319 .92732 .62191 Metetra
.623 .755 .941 r
.92732 .62191 .93597 .63319 .95913 .62815 .9506 .61677 Metetra
.629 .757 .939 r
.9506 .61677 .95913 .62815 .98245 .62279 .97405 .6113 Metetra
.646 .732 .912 r
.36202 .56619 .37296 .57664 .39084 .57212 .37998 .56146 Metetra
.633 .725 .914 r
.37998 .56146 .39084 .57212 .40883 .56822 .39804 .5573 Metetra
.613 .716 .918 r
.39804 .5573 .40883 .56822 .42694 .56518 .41622 .55394 Metetra
.586 .702 .922 r
.41622 .55394 .42694 .56518 .44517 .56327 .43452 .55167 Metetra
.551 .682 .926 r
.43452 .55167 .44517 .56327 .46353 .56272 .45295 .55074 Metetra
.506 .658 .928 r
.45295 .55074 .46353 .56272 .48206 .56372 .47154 .55136 Metetra
.455 .628 .928 r
.47154 .55136 .48206 .56372 .50078 .56634 .4903 .55365 Metetra
.4 .596 .924 r
.4903 .55365 .50078 .56634 .51972 .57056 .50928 .55761 Metetra
.345 .563 .918 r
.50928 .55761 .51972 .57056 .5389 .57623 .52851 .56311 Metetra
.297 .534 .911 r
.52851 .56311 .5389 .57623 .55837 .58309 .54801 .56991 Metetra
.26 .513 .905 r
.54801 .56991 .55837 .58309 .57813 .5908 .56781 .57767 Metetra
.238 .502 .902 r
.56781 .57767 .57813 .5908 .59819 .59895 .58792 .58597 Metetra
.232 .502 .903 r
.58792 .58597 .59819 .59895 .61857 .60715 .60834 .59438 Metetra
.242 .514 .909 r
.60834 .59438 .61857 .60715 .63924 .61499 .62908 .60248 Metetra
.265 .534 .917 r
.62908 .60248 .63924 .61499 .6602 .62213 .6501 .60991 Metetra
.299 .56 .927 r
.6501 .60991 .6602 .62213 .68143 .62831 .67141 .61637 Metetra
.339 .59 .937 r
.67141 .61637 .68143 .62831 .70289 .63333 .69296 .62165 Metetra
.383 .62 .945 r
.69296 .62165 .70289 .63333 .72458 .63708 .71474 .62561 Metetra
.425 .648 .951 r
.71474 .62561 .72458 .63708 .74647 .63953 .73673 .62823 Metetra
.465 .673 .954 r
.73673 .62823 .74647 .63953 .76853 .64071 .75891 .62952 Metetra
.5 .694 .955 r
.75891 .62952 .76853 .64071 .79077 .64068 .78125 .62956 Metetra
.531 .711 .954 r
.78125 .62956 .79077 .64068 .81315 .63957 .80375 .62846 Metetra
.556 .724 .953 r
.80375 .62846 .81315 .63957 .83569 .63748 .82641 .62635 Metetra
.576 .735 .95 r
.82641 .62635 .83569 .63748 .85838 .63454 .84922 .62337 Metetra
.593 .743 .948 r
.84922 .62337 .85838 .63454 .88121 .6309 .87217 .61965 Metetra
.606 .748 .945 r
.87217 .61965 .88121 .6309 .90419 .62665 .89528 .61532 Metetra
.616 .753 .943 r
.89528 .61532 .90419 .62665 .92732 .62191 .91854 .61048 Metetra
.624 .756 .941 r
.91854 .61048 .92732 .62191 .9506 .61677 .94195 .60523 Metetra
.63 .758 .939 r
.94195 .60523 .9506 .61677 .97405 .6113 .96554 .59965 Metetra
.649 .733 .911 r
.35094 .55563 .36202 .56619 .37998 .56146 .36898 .55071 Metetra
.637 .727 .913 r
.36898 .55071 .37998 .56146 .39804 .5573 .38712 .54628 Metetra
.619 .718 .917 r
.38712 .54628 .39804 .5573 .41622 .55394 .40537 .54261 Metetra
.594 .705 .921 r
.40537 .54261 .41622 .55394 .43452 .55167 .42374 .53997 Metetra
.56 .687 .925 r
.42374 .53997 .43452 .55167 .45295 .55074 .44224 .53865 Metetra
.517 .663 .927 r
.44224 .53865 .45295 .55074 .47154 .55136 .46089 .53888 Metetra
.465 .634 .928 r
.46089 .53888 .47154 .55136 .4903 .55365 .47971 .54081 Metetra
.408 .6 .925 r
.47971 .54081 .4903 .55365 .50928 .55761 .49873 .54448 Metetra
.351 .566 .918 r
.49873 .54448 .50928 .55761 .52851 .56311 .518 .54979 Metetra
.3 .535 .911 r
.518 .54979 .52851 .56311 .54801 .56991 .53754 .55651 Metetra
.259 .511 .904 r
.53754 .55651 .54801 .56991 .56781 .57767 .55737 .5643 Metetra
.233 .497 .9 r
.55737 .5643 .56781 .57767 .58792 .58597 .57752 .57273 Metetra
.223 .495 .9 r
.57752 .57273 .58792 .58597 .60834 .59438 .59799 .58137 Metetra
.231 .505 .905 r
.59799 .58137 .60834 .59438 .62908 .60248 .61878 .58975 Metetra
.253 .525 .914 r
.61878 .58975 .62908 .60248 .6501 .60991 .63988 .59748 Metetra
.287 .552 .925 r
.63988 .59748 .6501 .60991 .67141 .61637 .66125 .60424 Metetra
.328 .583 .935 r
.66125 .60424 .67141 .61637 .69296 .62165 .6829 .60979 Metetra
.373 .614 .944 r
.6829 .60979 .69296 .62165 .71474 .62561 .70477 .61399 Metetra
.417 .644 .951 r
.70477 .61399 .71474 .62561 .73673 .62823 .72686 .61678 Metetra
.459 .67 .954 r
.72686 .61678 .73673 .62823 .75891 .62952 .74914 .61819 Metetra
.496 .692 .956 r
.74914 .61819 .75891 .62952 .78125 .62956 .7716 .6183 Metetra
.528 .71 .955 r
.7716 .6183 .78125 .62956 .80375 .62846 .79422 .61722 Metetra
.554 .724 .953 r
.79422 .61722 .80375 .62846 .82641 .62635 .817 .61509 Metetra
.576 .735 .951 r
.817 .61509 .82641 .62635 .84922 .62337 .83993 .61205 Metetra
.593 .743 .948 r
.83993 .61205 .84922 .62337 .87217 .61965 .863 .60826 Metetra
.606 .749 .945 r
.863 .60826 .87217 .61965 .89528 .61532 .88624 .60383 Metetra
.616 .753 .943 r
.88624 .60383 .89528 .61532 .91854 .61048 .90963 .59889 Metetra
.624 .756 .941 r
.90963 .59889 .91854 .61048 .94195 .60523 .93318 .59353 Metetra
.63 .758 .939 r
.93318 .59353 .94195 .60523 .96554 .59965 .95689 .58784 Metetra
.651 .734 .91 r
.33972 .54498 .35094 .55563 .36898 .55071 .35784 .53986 Metetra
.641 .729 .913 r
.35784 .53986 .36898 .55071 .38712 .54628 .37606 .53519 Metetra
.625 .721 .916 r
.37606 .53519 .38712 .54628 .40537 .54261 .39439 .5312 Metetra
.601 .709 .92 r
.39439 .5312 .40537 .54261 .42374 .53997 .41284 .5282 Metetra
.569 .692 .924 r
.41284 .5282 .42374 .53997 .44224 .53865 .4314 .52647 Metetra
.527 .668 .927 r
.4314 .52647 .44224 .53865 .46089 .53888 .45011 .52628 Metetra
.476 .639 .927 r
.45011 .52628 .46089 .53888 .47971 .54081 .46899 .52783 Metetra
.418 .605 .925 r
.46899 .52783 .47971 .54081 .49873 .54448 .48806 .53117 Metetra
.359 .569 .919 r
.48806 .53117 .49873 .54448 .518 .54979 .50737 .53626 Metetra
.303 .536 .91 r
.50737 .53626 .518 .54979 .53754 .55651 .52694 .54287 Metetra
.258 .509 .903 r
.52694 .54287 .53754 .55651 .55737 .5643 .54681 .55068 Metetra
.227 .492 .897 r
.54681 .55068 .55737 .5643 .57752 .57273 .567 .55925 Metetra
.214 .488 .897 r
.567 .55925 .57752 .57273 .59799 .58137 .58751 .56811 Metetra
.219 .496 .901 r
.58751 .56811 .59799 .58137 .61878 .58975 .60836 .57678 Metetra
.24 .515 .91 r
.60836 .57678 .61878 .58975 .63988 .59748 .62951 .58484 Metetra
.273 .543 .921 r
.62951 .58484 .63988 .59748 .66125 .60424 .65097 .59191 Metetra
.316 .575 .933 r
.65097 .59191 .66125 .60424 .6829 .60979 .67269 .59776 Metetra
.362 .608 .943 r
.67269 .59776 .6829 .60979 .70477 .61399 .69466 .6022 Metetra
.409 .64 .95 r
.69466 .6022 .70477 .61399 .72686 .61678 .71685 .60518 Metetra
.453 .667 .955 r
.71685 .60518 .72686 .61678 .74914 .61819 .73924 .60672 Metetra
.492 .691 .956 r
.73924 .60672 .74914 .61819 .7716 .6183 .76181 .6069 Metetra
.525 .709 .956 r
.76181 .6069 .7716 .6183 .79422 .61722 .78455 .60584 Metetra
.553 .724 .954 r
.78455 .60584 .79422 .61722 .817 .61509 .80745 .60368 Metetra
.575 .735 .951 r
.80745 .60368 .817 .61509 .83993 .61205 .8305 .60059 Metetra
.593 .743 .948 r
.8305 .60059 .83993 .61205 .863 .60826 .8537 .59671 Metetra
.606 .749 .946 r
.8537 .59671 .863 .60826 .88624 .60383 .87706 .59219 Metetra
.617 .753 .943 r
.87706 .59219 .88624 .60383 .90963 .59889 .90059 .58713 Metetra
.625 .756 .941 r
.90059 .58713 .90963 .59889 .93318 .59353 .92427 .58166 Metetra
.63 .758 .939 r
.92427 .58166 .93318 .59353 .95689 .58784 .94812 .57585 Metetra
.653 .735 .91 r
.32837 .53421 .33972 .54498 .35784 .53986 .34656 .52891 Metetra
.645 .731 .912 r
.34656 .52891 .35784 .53986 .37606 .53519 .36486 .524 Metetra
.63 .724 .915 r
.36486 .524 .37606 .53519 .39439 .5312 .38327 .51971 Metetra
.609 .712 .918 r
.38327 .51971 .39439 .5312 .41284 .5282 .40179 .51633 Metetra
.578 .696 .922 r
.40179 .51633 .41284 .5282 .4314 .52647 .42043 .51419 Metetra
.537 .674 .926 r
.42043 .51419 .4314 .52647 .45011 .52628 .43921 .51357 Metetra
.487 .645 .927 r
.43921 .51357 .45011 .52628 .46899 .52783 .45814 .5147 Metetra
.429 .611 .925 r
.45814 .5147 .46899 .52783 .48806 .53117 .47726 .51769 Metetra
.367 .573 .919 r
.47726 .51769 .48806 .53117 .50737 .53626 .49661 .52252 Metetra
.308 .537 .91 r
.49661 .52252 .50737 .53626 .52694 .54287 .51622 .529 Metetra
.258 .507 .901 r
.51622 .529 .52694 .54287 .54681 .55068 .53612 .53681 Metetra
.222 .487 .895 r
.53612 .53681 .54681 .55068 .567 .55925 .55635 .54551 Metetra
.205 .48 .893 r
.55635 .54551 .567 .55925 .58751 .56811 .5769 .55459 Metetra
.207 .487 .897 r
.5769 .55459 .58751 .56811 .60836 .57678 .5978 .56357 Metetra
.226 .505 .906 r
.5978 .56357 .60836 .57678 .62951 .58484 .61902 .57196 Metetra
.259 .533 .918 r
.61902 .57196 .62951 .58484 .65097 .59191 .64054 .57938 Metetra
.302 .566 .931 r
.64054 .57938 .65097 .59191 .67269 .59776 .66235 .58554 Metetra
.351 .601 .942 r
.66235 .58554 .67269 .59776 .69466 .6022 .68441 .59025 Metetra
.4 .635 .95 r
.68441 .59025 .69466 .6022 .71685 .60518 .70671 .59343 Metetra
.446 .664 .955 r
.70671 .59343 .71685 .60518 .73924 .60672 .7292 .59511 Metetra
.487 .689 .957 r
.7292 .59511 .73924 .60672 .76181 .6069 .75189 .59536 Metetra
.522 .709 .956 r
.75189 .59536 .76181 .6069 .78455 .60584 .77474 .59432 Metetra
.551 .724 .954 r
.77474 .59432 .78455 .60584 .80745 .60368 .79776 .59214 Metetra
.574 .735 .951 r
.79776 .59214 .80745 .60368 .8305 .60059 .82094 .58898 Metetra
.592 .744 .949 r
.82094 .58898 .8305 .60059 .8537 .59671 .84427 .58501 Metetra
.606 .75 .946 r
.84427 .58501 .8537 .59671 .87706 .59219 .86776 .58038 Metetra
.617 .754 .943 r
.86776 .58038 .87706 .59219 .90059 .58713 .89141 .57521 Metetra
.625 .757 .941 r
.89141 .57521 .90059 .58713 .92427 .58166 .91523 .56961 Metetra
.631 .759 .939 r
.91523 .56961 .92427 .58166 .94812 .57585 .93922 .56368 Metetra
.656 .736 .909 r
.31686 .52333 .32837 .53421 .34656 .52891 .33513 .51786 Metetra
.648 .733 .911 r
.33513 .51786 .34656 .52891 .36486 .524 .35352 .51272 Metetra
.635 .726 .913 r
.35352 .51272 .36486 .524 .38327 .51971 .37201 .50813 Metetra
.615 .716 .917 r
.37201 .50813 .38327 .51971 .40179 .51633 .39061 .50439 Metetra
.587 .7 .921 r
.39061 .50439 .40179 .51633 .42043 .51419 .40933 .50182 Metetra
.548 .679 .925 r
.40933 .50182 .42043 .51419 .43921 .51357 .42817 .50075 Metetra
.499 .651 .926 r
.42817 .50075 .43921 .51357 .45814 .5147 .44716 .50143 Metetra
.44 .616 .925 r
.44716 .50143 .45814 .5147 .47726 .51769 .46633 .50404 Metetra
.376 .578 .919 r
.46633 .50404 .47726 .51769 .49661 .52252 .48572 .50857 Metetra
.313 .539 .91 r
.48572 .50857 .49661 .52252 .51622 .529 .50537 .51489 Metetra
.258 .506 .901 r
.50537 .51489 .51622 .529 .53612 .53681 .52531 .52267 Metetra
.218 .483 .893 r
.52531 .52267 .53612 .53681 .55635 .54551 .54556 .53149 Metetra
.196 .473 .89 r
.54556 .53149 .55635 .54551 .5769 .55459 .56616 .54081 Metetra
.194 .477 .893 r
.56616 .54081 .5769 .55459 .5978 .56357 .5871 .5501 Metetra
.211 .495 .901 r
.5871 .5501 .5978 .56357 .61902 .57196 .60838 .55885 Metetra
.244 .523 .914 r
.60838 .55885 .61902 .57196 .64054 .57938 .62997 .56664 Metetra
.288 .557 .928 r
.62997 .56664 .64054 .57938 .66235 .58554 .65186 .57313 Metetra
.339 .594 .94 r
.65186 .57313 .66235 .58554 .68441 .59025 .67402 .57813 Metetra
.39 .63 .949 r
.67402 .57813 .68441 .59025 .70671 .59343 .69642 .58153 Metetra
.439 .661 .955 r
.69642 .58153 .70671 .59343 .7292 .59511 .71902 .58336 Metetra
.483 .687 .957 r
.71902 .58336 .7292 .59511 .75189 .59536 .74182 .58369 Metetra
.52 .708 .957 r
.74182 .58369 .75189 .59536 .77474 .59432 .76479 .58266 Metetra
.55 .724 .955 r
.76479 .58266 .77474 .59432 .79776 .59214 .78793 .58045 Metetra
.574 .736 .952 r
.78793 .58045 .79776 .59214 .82094 .58898 .81123 .57722 Metetra
.593 .744 .949 r
.81123 .57722 .82094 .58898 .84427 .58501 .83469 .57316 Metetra
.607 .75 .946 r
.83469 .57316 .84427 .58501 .86776 .58038 .85831 .56841 Metetra
.618 .754 .943 r
.85831 .56841 .86776 .58038 .89141 .57521 .8821 .56312 Metetra
.626 .757 .941 r
.8821 .56312 .89141 .57521 .91523 .56961 .90605 .55739 Metetra
.632 .759 .939 r
.90605 .55739 .91523 .56961 .93922 .56368 .93018 .55133 Metetra
.657 .737 .909 r
.30521 .51234 .31686 .52333 .33513 .51786 .32356 .50671 Metetra
.651 .734 .91 r
.32356 .50671 .33513 .51786 .35352 .51272 .34203 .50135 Metetra
.64 .728 .912 r
.34203 .50135 .35352 .51272 .37201 .50813 .36061 .49647 Metetra
.622 .719 .916 r
.36061 .49647 .37201 .50813 .39061 .50439 .37929 .49237 Metetra
.596 .705 .92 r
.37929 .49237 .39061 .50439 .40933 .50182 .39808 .48937 Metetra
.559 .685 .923 r
.39808 .48937 .40933 .50182 .42817 .50075 .417 .48782 Metetra
.511 .657 .926 r
.417 .48782 .42817 .50075 .44716 .50143 .43606 .48803 Metetra
.452 .623 .925 r
.43606 .48803 .44716 .50143 .46633 .50404 .45528 .49021 Metetra
.386 .583 .92 r
.45528 .49021 .46633 .50404 .48572 .50857 .47471 .49441 Metetra
.32 .542 .911 r
.47471 .49441 .48572 .50857 .50537 .51489 .49439 .50053 Metetra
.26 .506 .9 r
.49439 .50053 .50537 .51489 .52531 .52267 .51436 .50827 Metetra
.214 .478 .891 r
.51436 .50827 .52531 .52267 .54556 .53149 .53465 .51718 Metetra
.186 .465 .886 r
.53465 .51718 .54556 .53149 .56616 .54081 .55528 .52674 Metetra
.18 .467 .888 r
.55528 .52674 .56616 .54081 .5871 .5501 .57627 .53636 Metetra
.195 .483 .896 r
.57627 .53636 .5871 .5501 .60838 .55885 .5976 .5455 Metetra
.227 .511 .909 r
.5976 .5455 .60838 .55885 .62997 .56664 .61926 .55368 Metetra
.273 .547 .924 r
.61926 .55368 .62997 .56664 .65186 .57313 .64124 .56054 Metetra
.325 .586 .938 r
.64124 .56054 .65186 .57313 .67402 .57813 .66348 .56584 Metetra
.38 .624 .948 r
.66348 .56584 .67402 .57813 .69642 .58153 .68598 .56948 Metetra
.432 .657 .955 r
.68598 .56948 .69642 .58153 .71902 .58336 .70869 .57146 Metetra
.478 .685 .958 r
.70869 .57146 .71902 .58336 .74182 .58369 .7316 .57187 Metetra
.517 .707 .957 r
.7316 .57187 .74182 .58369 .76479 .58266 .7547 .57086 Metetra
.548 .724 .955 r
.7547 .57086 .76479 .58266 .78793 .58045 .77796 .56861 Metetra
.573 .736 .952 r
.77796 .56861 .78793 .58045 .81123 .57722 .80139 .56531 Metetra
.593 .745 .949 r
.80139 .56531 .81123 .57722 .83469 .57316 .82497 .56114 Metetra
.607 .751 .946 r
.82497 .56114 .83469 .57316 .85831 .56841 .84872 .55628 Metetra
.618 .755 .943 r
.84872 .55628 .85831 .56841 .8821 .56312 .87264 .55085 Metetra
.626 .758 .941 r
.87264 .55085 .8821 .56312 .90605 .55739 .89674 .54499 Metetra
.632 .76 .939 r
.89674 .54499 .90605 .55739 .93018 .55133 .92101 .53879 Metetra
.659 .738 .908 r
.29341 .50122 .30521 .51234 .32356 .50671 .31184 .49544 Metetra
.654 .735 .909 r
.31184 .49544 .32356 .50671 .34203 .50135 .33039 .48987 Metetra
.644 .73 .911 r
.33039 .48987 .34203 .50135 .36061 .49647 .34905 .48472 Metetra
.628 .722 .914 r
.34905 .48472 .36061 .49647 .37929 .49237 .36782 .48026 Metetra
.604 .709 .918 r
.36782 .48026 .37929 .49237 .39808 .48937 .3867 .47683 Metetra
.57 .69 .922 r
.3867 .47683 .39808 .48937 .417 .48782 .40569 .47479 Metetra
.523 .664 .925 r
.40569 .47479 .417 .48782 .43606 .48803 .42481 .4745 Metetra
.465 .63 .925 r
.42481 .4745 .43606 .48803 .45528 .49021 .4441 .47622 Metetra
.398 .589 .92 r
.4441 .47622 .45528 .49021 .47471 .49441 .46357 .48004 Metetra
.328 .546 .911 r
.46357 .48004 .47471 .49441 .49439 .50053 .48329 .48592 Metetra
.262 .506 .899 r
.48329 .48592 .49439 .50053 .51436 .50827 .50329 .49357 Metetra
.21 .474 .888 r
.50329 .49357 .51436 .50827 .53465 .51718 .52361 .50258 Metetra
.177 .457 .882 r
.52361 .50258 .53465 .51718 .55528 .52674 .54427 .51237 Metetra
.167 .456 .883 r
.54427 .51237 .55528 .52674 .57627 .53636 .5653 .52234 Metetra
.178 .471 .891 r
.5653 .52234 .57627 .53636 .5976 .5455 .58668 .53189 Metetra
.21 .499 .904 r
.58668 .53189 .5976 .5455 .61926 .55368 .60841 .54049 Metetra
.256 .536 .92 r
.60841 .54049 .61926 .55368 .64124 .56054 .63046 .54774 Metetra
.311 .578 .936 r
.63046 .54774 .64124 .56054 .66348 .56584 .6528 .55338 Metetra
.369 .618 .948 r
.6528 .55338 .66348 .56584 .68598 .56948 .67539 .55727 Metetra
.424 .654 .955 r
.67539 .55727 .68598 .56948 .70869 .57146 .69821 .55942 Metetra
.473 .683 .958 r
.69821 .55942 .70869 .57146 .7316 .57187 .72124 .55992 Metetra
.514 .707 .958 r
.72124 .55992 .7316 .57187 .7547 .57086 .74445 .55893 Metetra
.547 .724 .956 r
.74445 .55893 .7547 .57086 .77796 .56861 .76784 .55663 Metetra
.573 .736 .953 r
.76784 .55663 .77796 .56861 .80139 .56531 .79139 .55325 Metetra
.593 .745 .949 r
.79139 .55325 .80139 .56531 .82497 .56114 .81511 .54897 Metetra
.608 .751 .946 r
.81511 .54897 .82497 .56114 .84872 .55628 .83899 .54397 Metetra
.619 .755 .943 r
.83899 .54397 .84872 .55628 .87264 .55085 .86305 .5384 Metetra
.627 .758 .941 r
.86305 .5384 .87264 .55085 .89674 .54499 .88728 .5324 Metetra
.633 .76 .939 r
.88728 .5324 .89674 .54499 .92101 .53879 .91169 .52606 Metetra
.66 .739 .908 r
.28146 .48997 .29341 .50122 .31184 .49544 .29997 .48405 Metetra
.656 .736 .909 r
.29997 .48405 .31184 .49544 .33039 .48987 .3186 .4783 Metetra
.648 .732 .911 r
.3186 .4783 .33039 .48987 .34905 .48472 .33735 .47289 Metetra
.634 .725 .913 r
.33735 .47289 .34905 .48472 .36782 .48026 .3562 .46808 Metetra
.613 .713 .917 r
.3562 .46808 .36782 .48026 .3867 .47683 .37517 .46421 Metetra
.581 .696 .921 r
.37517 .46421 .3867 .47683 .40569 .47479 .39424 .46168 Metetra
.536 .67 .924 r
.39424 .46168 .40569 .47479 .42481 .4745 .41343 .46086 Metetra
.479 .637 .924 r
.41343 .46086 .42481 .4745 .4441 .47622 .43278 .46206 Metetra
.411 .596 .92 r
.43278 .46206 .4441 .47622 .46357 .48004 .4523 .46546 Metetra
.337 .55 .912 r
.4523 .46546 .46357 .48004 .48329 .48592 .47206 .47105 Metetra
.267 .507 .899 r
.47206 .47105 .48329 .48592 .50329 .49357 .49208 .47859 Metetra
.208 .471 .886 r
.49208 .47859 .50329 .49357 .52361 .50258 .51243 .48766 Metetra
.169 .449 .878 r
.51243 .48766 .52361 .50258 .54427 .51237 .53312 .49769 Metetra
.153 .444 .877 r
.53312 .49769 .54427 .51237 .5653 .52234 .55418 .50802 Metetra
.161 .457 .884 r
.55418 .50802 .5653 .52234 .58668 .53189 .57562 .51801 Metetra
.191 .486 .898 r
.57562 .51801 .58668 .53189 .60841 .54049 .59741 .52707 Metetra
.238 .524 .916 r
.59741 .52707 .60841 .54049 .63046 .54774 .61953 .53475 Metetra
.295 .568 .933 r
.61953 .53475 .63046 .54774 .6528 .55338 .64196 .54075 Metetra
.357 .611 .946 r
.64196 .54075 .6528 .55338 .67539 .55727 .66465 .54492 Metetra
.415 .65 .955 r
.66465 .54492 .67539 .55727 .69821 .55942 .68758 .54725 Metetra
.467 .681 .959 r
.68758 .54725 .69821 .55942 .72124 .55992 .71072 .54783 Metetra
.511 .706 .959 r
.71072 .54783 .72124 .55992 .74445 .55893 .73406 .54685 Metetra
.546 .724 .957 r
.73406 .54685 .74445 .55893 .76784 .55663 .75757 .54451 Metetra
.573 .737 .953 r
.75757 .54451 .76784 .55663 .79139 .55325 .78125 .54103 Metetra
.594 .746 .95 r
.78125 .54103 .79139 .55325 .81511 .54897 .80509 .53662 Metetra
.609 .752 .946 r
.80509 .53662 .81511 .54897 .83899 .54397 .82911 .53148 Metetra
.62 .756 .943 r
.82911 .53148 .83899 .54397 .86305 .5384 .8533 .52577 Metetra
.628 .759 .941 r
.8533 .52577 .86305 .5384 .88728 .5324 .87767 .51962 Metetra
.634 .76 .939 r
.87767 .51962 .88728 .5324 .91169 .52606 .90223 .51314 Metetra
.661 .739 .908 r
.26935 .47859 .28146 .48997 .29997 .48405 .28794 .47254 Metetra
.658 .737 .908 r
.28794 .47254 .29997 .48405 .3186 .4783 .30666 .46661 Metetra
.651 .734 .91 r
.30666 .46661 .3186 .4783 .33735 .47289 .32549 .46096 Metetra
.64 .728 .912 r
.32549 .46096 .33735 .47289 .3562 .46808 .34443 .45582 Metetra
.621 .717 .915 r
.34443 .45582 .3562 .46808 .37517 .46421 .36349 .45153 Metetra
.591 .701 .919 r
.36349 .45153 .37517 .46421 .39424 .46168 .38264 .44848 Metetra
.549 .677 .923 r
.38264 .44848 .39424 .46168 .41343 .46086 .40191 .4471 Metetra
.493 .644 .924 r
.40191 .4471 .41343 .46086 .43278 .46206 .42132 .44775 Metetra
.425 .603 .921 r
.42132 .44775 .43278 .46206 .4523 .46546 .4409 .45068 Metetra
.348 .556 .912 r
.4409 .45068 .4523 .46546 .47206 .47105 .46069 .45592 Metetra
.272 .508 .899 r
.46069 .45592 .47206 .47105 .49208 .47859 .48075 .4633 Metetra
.207 .468 .885 r
.48075 .4633 .49208 .47859 .51243 .48766 .50112 .47242 Metetra
.16 .441 .874 r
.50112 .47242 .51243 .48766 .53312 .49769 .52183 .48267 Metetra
.138 .432 .871 r
.52183 .48267 .53312 .49769 .55418 .50802 .54293 .49338 Metetra
.142 .443 .877 r
.54293 .49338 .55418 .50802 .57562 .51801 .5644 .50384 Metetra
.17 .471 .892 r
.5644 .50384 .57562 .51801 .59741 .52707 .58625 .5134 Metetra
.218 .511 .91 r
.58625 .5134 .59741 .52707 .61953 .53475 .60845 .52155 Metetra
.278 .557 .929 r
.60845 .52155 .61953 .53475 .64196 .54075 .63096 .52795 Metetra
.343 .604 .945 r
.63096 .52795 .64196 .54075 .66465 .54492 .65375 .53241 Metetra
.406 .645 .955 r
.65375 .53241 .66465 .54492 .68758 .54725 .67679 .53494 Metetra
.461 .679 .959 r
.67679 .53494 .68758 .54725 .71072 .54783 .70005 .53562 Metetra
.508 .706 .96 r
.70005 .53562 .71072 .54783 .73406 .54685 .7235 .53464 Metetra
.545 .725 .957 r
.7235 .53464 .73406 .54685 .75757 .54451 .74714 .53224 Metetra
.573 .738 .954 r
.74714 .53224 .75757 .54451 .78125 .54103 .77095 .52865 Metetra
.594 .747 .95 r
.77095 .52865 .78125 .54103 .80509 .53662 .79493 .52411 Metetra
.61 .753 .946 r
.79493 .52411 .80509 .53662 .82911 .53148 .81908 .51882 Metetra
.621 .757 .943 r
.81908 .51882 .82911 .53148 .8533 .52577 .84341 .51295 Metetra
.629 .759 .941 r
.84341 .51295 .8533 .52577 .87767 .51962 .86792 .50665 Metetra
.635 .761 .939 r
.86792 .50665 .87767 .51962 .90223 .51314 .89262 .50001 Metetra
.662 .739 .907 r
.25708 .46708 .26935 .47859 .28794 .47254 .27575 .46091 Metetra
.659 .738 .908 r
.27575 .46091 .28794 .47254 .30666 .46661 .29455 .45482 Metetra
.654 .735 .909 r
.29455 .45482 .30666 .46661 .32549 .46096 .31347 .44894 Metetra
.645 .73 .911 r
.31347 .44894 .32549 .46096 .34443 .45582 .33251 .44348 Metetra
.628 .721 .914 r
.33251 .44348 .34443 .45582 .36349 .45153 .35165 .43877 Metetra
.601 .706 .918 r
.35165 .43877 .36349 .45153 .38264 .44848 .3709 .43521 Metetra
.562 .684 .921 r
.3709 .43521 .38264 .44848 .40191 .4471 .39025 .43324 Metetra
.508 .652 .923 r
.39025 .43324 .40191 .4471 .42132 .44775 .40973 .4333 Metetra
.44 .611 .921 r
.40973 .4333 .42132 .44775 .4409 .45068 .42936 .43569 Metetra
.361 .562 .913 r
.42936 .43569 .4409 .45068 .46069 .45592 .44919 .44053 Metetra
.28 .511 .899 r
.44919 .44053 .46069 .45592 .48075 .4633 .46928 .4477 Metetra
.207 .466 .883 r
.46928 .4477 .48075 .4633 .50112 .47242 .48967 .45683 Metetra
.152 .433 .87 r
.48967 .45683 .50112 .47242 .52183 .48267 .51041 .46731 Metetra
.123 .42 .865 r
.51041 .46731 .52183 .48267 .54293 .49338 .53153 .47841 Metetra
.122 .427 .87 r
.53153 .47841 .54293 .49338 .5644 .50384 .55304 .48937 Metetra
.148 .454 .884 r
.55304 .48937 .5644 .50384 .58625 .5134 .57494 .49948 Metetra
.196 .496 .904 r
.57494 .49948 .58625 .5134 .60845 .52155 .59721 .50814 Metetra
.259 .546 .925 r
.59721 .50814 .60845 .52155 .63096 .52795 .61981 .51497 Metetra
.328 .596 .943 r
.61981 .51497 .63096 .52795 .65375 .53241 .6427 .51977 Metetra
.396 .641 .955 r
.6427 .51977 .65375 .53241 .67679 .53494 .66584 .52249 Metetra
.455 .677 .96 r
.66584 .52249 .67679 .53494 .70005 .53562 .68922 .52327 Metetra
.505 .705 .96 r
.68922 .52327 .70005 .53562 .7235 .53464 .71279 .52229 Metetra
.544 .725 .958 r
.71279 .52229 .7235 .53464 .74714 .53224 .73655 .51982 Metetra
.573 .739 .954 r
.73655 .51982 .74714 .53224 .77095 .52865 .76049 .51612 Metetra
.595 .748 .95 r
.76049 .51612 .77095 .52865 .79493 .52411 .7846 .51143 Metetra
.611 .754 .946 r
.7846 .51143 .79493 .52411 .81908 .51882 .80889 .50597 Metetra
.622 .758 .943 r
.80889 .50597 .81908 .51882 .84341 .51295 .83336 .49994 Metetra
.63 .76 .94 r
.83336 .49994 .84341 .51295 .86792 .50665 .85801 .49347 Metetra
.636 .761 .938 r
.85801 .49347 .86792 .50665 .89262 .50001 .88286 .48667 Metetra
.662 .74 .907 r
.24465 .45542 .25708 .46708 .27575 .46091 .2634 .44914 Metetra
.661 .739 .908 r
.2634 .44914 .27575 .46091 .29455 .45482 .28229 .4429 Metetra
.657 .737 .909 r
.28229 .4429 .29455 .45482 .31347 .44894 .30129 .43681 Metetra
.649 .732 .91 r
.30129 .43681 .31347 .44894 .33251 .44348 .32042 .43106 Metetra
.635 .725 .913 r
.32042 .43106 .33251 .44348 .35165 .43877 .33966 .42595 Metetra
.611 .711 .916 r
.33966 .42595 .35165 .43877 .3709 .43521 .35899 .42187 Metetra
.575 .691 .92 r
.35899 .42187 .3709 .43521 .39025 .43324 .37843 .41929 Metetra
.524 .661 .922 r
.37843 .41929 .39025 .43324 .40973 .4333 .39799 .41871 Metetra
.457 .62 .921 r
.39799 .41871 .40973 .4333 .42936 .43569 .41768 .4205 Metetra
.376 .57 .914 r
.41768 .4205 .42936 .43569 .44919 .44053 .43756 .42488 Metetra
.289 .515 .9 r
.43756 .42488 .44919 .44053 .46928 .4477 .45768 .43178 Metetra
.208 .464 .882 r
.45768 .43178 .46928 .4477 .48967 .45683 .47809 .44087 Metetra
.145 .426 .866 r
.47809 .44087 .48967 .45683 .51041 .46731 .49884 .45157 Metetra
.108 .407 .858 r
.49884 .45157 .51041 .46731 .53153 .47841 .51999 .46308 Metetra
.101 .411 .861 r
.51999 .46308 .53153 .47841 .55304 .48937 .54153 .47459 Metetra
.124 .437 .875 r
.54153 .47459 .55304 .48937 .57494 .49948 .56348 .48528 Metetra
.172 .48 .897 r
.56348 .48528 .57494 .49948 .59721 .50814 .58581 .49452 Metetra
.238 .533 .921 r
.58581 .49452 .59721 .50814 .61981 .51497 .60849 .50183 Metetra
.312 .587 .941 r
.60849 .50183 .61981 .51497 .6427 .51977 .63148 .50697 Metetra
.385 .636 .954 r
.63148 .50697 .6427 .51977 .66584 .52249 .65473 .50993 Metetra
.449 .675 .961 r
.65473 .50993 .66584 .52249 .68922 .52327 .67822 .5108 Metetra
.502 .705 .961 r
.67822 .5108 .68922 .52327 .71279 .52229 .70192 .50982 Metetra
.543 .726 .959 r
.70192 .50982 .71279 .52229 .73655 .51982 .7258 .50726 Metetra
.574 .74 .955 r
.7258 .50726 .73655 .51982 .76049 .51612 .74987 .50342 Metetra
.597 .749 .95 r
.74987 .50342 .76049 .51612 .7846 .51143 .77411 .49857 Metetra
.613 .755 .946 r
.77411 .49857 .7846 .51143 .80889 .50597 .79854 .49294 Metetra
.624 .758 .943 r
.79854 .49294 .80889 .50597 .83336 .49994 .82315 .48672 Metetra
.631 .76 .94 r
.82315 .48672 .83336 .49994 .85801 .49347 .84795 .48008 Metetra
.637 .762 .938 r
.84795 .48008 .85801 .49347 .88286 .48667 .87294 .47312 Metetra
.663 .74 .907 r
.23205 .44361 .24465 .45542 .2634 .44914 .25089 .43722 Metetra
.662 .739 .908 r
.25089 .43722 .2634 .44914 .28229 .4429 .26985 .43085 Metetra
.659 .738 .908 r
.26985 .43085 .28229 .4429 .30129 .43681 .28895 .42457 Metetra
.653 .734 .909 r
.28895 .42457 .30129 .43681 .32042 .43106 .30817 .41855 Metetra
.641 .728 .911 r
.30817 .41855 .32042 .43106 .33966 .42595 .3275 .41306 Metetra
.621 .716 .914 r
.3275 .41306 .33966 .42595 .35899 .42187 .34693 .40847 Metetra
.588 .698 .918 r
.34693 .40847 .35899 .42187 .37843 .41929 .36647 .40528 Metetra
.54 .669 .921 r
.36647 .40528 .37843 .41929 .39799 .41871 .3861 .40401 Metetra
.475 .63 .921 r
.3861 .40401 .39799 .41871 .41768 .4205 .40587 .40514 Metetra
.393 .579 .915 r
.40587 .40514 .41768 .4205 .43756 .42488 .42579 .40897 Metetra
.301 .521 .9 r
.42579 .40897 .43756 .42488 .45768 .43178 .44594 .41553 Metetra
.212 .464 .881 r
.44594 .41553 .45768 .43178 .47809 .44087 .46637 .42454 Metetra
.139 .419 .862 r
.46637 .42454 .47809 .44087 .49884 .45157 .48714 .43542 Metetra
.093 .393 .85 r
.48714 .43542 .49884 .45157 .51999 .46308 .5083 .44737 Metetra
.079 .393 .851 r
.5083 .44737 .51999 .46308 .54153 .47459 .52987 .45946 Metetra
.098 .417 .865 r
.52987 .45946 .54153 .47459 .56348 .48528 .55186 .47081 Metetra
.146 .461 .888 r
.55186 .47081 .56348 .48528 .58581 .49452 .57425 .48067 Metetra
.215 .518 .915 r
.57425 .48067 .58581 .49452 .60849 .50183 .59701 .48851 Metetra
.294 .577 .938 r
.59701 .48851 .60849 .50183 .63148 .50697 .62009 .49405 Metetra
.373 .63 .954 r
.62009 .49405 .63148 .50697 .65473 .50993 .64345 .49724 Metetra
.442 .673 .961 r
.64345 .49724 .65473 .50993 .67822 .5108 .66705 .49821 Metetra
.499 .705 .962 r
.66705 .49821 .67822 .5108 .70192 .50982 .69087 .49721 Metetra
.543 .727 .96 r
.69087 .49721 .70192 .50982 .7258 .50726 .71489 .49455 Metetra
.575 .741 .955 r
.71489 .49455 .7258 .50726 .74987 .50342 .73908 .49056 Metetra
.598 .751 .95 r
.73908 .49056 .74987 .50342 .77411 .49857 .76346 .48552 Metetra
.614 .756 .946 r
.76346 .48552 .77411 .49857 .79854 .49294 .78803 .4797 Metetra
.625 .759 .943 r
.78803 .4797 .79854 .49294 .82315 .48672 .81278 .4733 Metetra
.633 .761 .94 r
.81278 .4733 .82315 .48672 .84795 .48008 .83772 .46647 Metetra
.638 .762 .938 r
.83772 .46647 .84795 .48008 .87294 .47312 .86286 .45934 Metetra
.663 .74 .907 r
.21929 .43165 .23205 .44361 .25089 .43722 .2382 .42516 Metetra
.662 .74 .907 r
.2382 .42516 .25089 .43722 .26985 .43085 .25725 .41867 Metetra
.66 .738 .908 r
.25725 .41867 .26985 .43085 .28895 .42457 .27644 .41222 Metetra
.656 .736 .909 r
.27644 .41222 .28895 .42457 .30817 .41855 .29575 .40596 Metetra
.647 .731 .91 r
.29575 .40596 .30817 .41855 .3275 .41306 .31518 .4001 Metetra
.63 .721 .913 r
.31518 .4001 .3275 .41306 .34693 .40847 .33471 .39502 Metetra
.601 .704 .916 r
.33471 .39502 .34693 .40847 .36647 .40528 .35434 .3912 Metetra
.557 .678 .92 r
.35434 .3912 .36647 .40528 .3861 .40401 .37407 .38921 Metetra
.494 .64 .92 r
.37407 .38921 .3861 .40401 .40587 .40514 .39391 .38961 Metetra
.412 .589 .915 r
.39391 .38961 .40587 .40514 .42579 .40897 .41389 .3928 Metetra
.316 .528 .902 r
.41389 .3928 .42579 .40897 .44594 .41553 .43407 .39893 Metetra
.219 .465 .881 r
.43407 .39893 .44594 .41553 .46637 .42454 .45452 .4078 Metetra
.135 .412 .858 r
.45452 .4078 .46637 .42454 .48714 .43542 .47529 .41885 Metetra
.078 .379 .842 r
.47529 .41885 .48714 .43542 .5083 .44737 .49646 .43123 Metetra
.055 .373 .84 r
.49646 .43123 .5083 .44737 .52987 .45946 .51805 .44396 Metetra
.069 .395 .853 r
.51805 .44396 .52987 .45946 .55186 .47081 .54008 .45604 Metetra
.116 .44 .877 r
.54008 .45604 .55186 .47081 .57425 .48067 .56252 .4666 Metetra
.189 .501 .907 r
.56252 .4666 .57425 .48067 .59701 .48851 .58536 .47503 Metetra
.274 .565 .935 r
.58536 .47503 .59701 .48851 .62009 .49405 .60853 .481 Metetra
.36 .624 .953 r
.60853 .481 .62009 .49405 .64345 .49724 .632 .48445 Metetra
.435 .671 .962 r
.632 .48445 .64345 .49724 .66705 .49821 .65572 .48551 Metetra
.496 .706 .963 r
.65572 .48551 .66705 .49821 .69087 .49721 .67966 .48448 Metetra
.543 .728 .96 r
.67966 .48448 .69087 .49721 .71489 .49455 .7038 .4817 Metetra
.576 .743 .955 r
.7038 .4817 .71489 .49455 .73908 .49056 .72813 .47753 Metetra
.6 .752 .95 r
.72813 .47753 .73908 .49056 .76346 .48552 .75264 .47229 Metetra
.616 .757 .946 r
.75264 .47229 .76346 .48552 .78803 .4797 .77734 .46626 Metetra
.627 .76 .942 r
.77734 .46626 .78803 .4797 .81278 .4733 .80224 .45966 Metetra
.634 .762 .94 r
.80224 .45966 .81278 .4733 .83772 .46647 .82733 .45265 Metetra
.639 .762 .938 r
.82733 .45265 .83772 .46647 .86286 .45934 .85262 .44534 Metetra
.663 .74 .907 r
.20635 .41953 .21929 .43165 .2382 .42516 .22534 .41295 Metetra
.663 .74 .907 r
.22534 .41295 .2382 .42516 .25725 .41867 .24448 .40634 Metetra
.662 .739 .907 r
.24448 .40634 .25725 .41867 .27644 .41222 .26375 .39974 Metetra
.658 .737 .908 r
.26375 .39974 .27644 .41222 .29575 .40596 .28315 .39326 Metetra
.651 .733 .909 r
.28315 .39326 .29575 .40596 .31518 .4001 .30268 .38708 Metetra
.638 .725 .911 r
.30268 .38708 .31518 .4001 .33471 .39502 .32232 .38153 Metetra
.613 .711 .915 r
.32232 .38153 .33471 .39502 .35434 .3912 .34205 .37708 Metetra
.573 .687 .918 r
.34205 .37708 .35434 .3912 .37407 .38921 .36188 .37434 Metetra
.514 .651 .92 r
.36188 .37434 .37407 .38921 .39391 .38961 .3818 .37393 Metetra
.433 .601 .916 r
.3818 .37393 .39391 .38961 .41389 .3928 .40184 .37639 Metetra
.334 .537 .903 r
.40184 .37639 .41389 .3928 .43407 .39893 .42206 .38199 Metetra
.228 .469 .881 r
.42206 .38199 .43407 .39893 .45452 .4078 .44253 .39063 Metetra
.132 .407 .855 r
.44253 .39063 .45452 .4078 .47529 .41885 .46331 .4018 Metetra
.063 .365 .834 r
.46331 .4018 .47529 .41885 .49646 .43123 .48448 .41464 Metetra
.03 .352 .828 r
.48448 .41464 .49646 .43123 .51805 .44396 .50608 .42807 Metetra
.038 .37 .839 r
.50608 .42807 .51805 .44396 .54008 .45604 .52813 .44094 Metetra
.083 .416 .865 r
.52813 .44094 .54008 .45604 .56252 .4666 .55063 .45229 Metetra
.159 .481 .899 r
.55063 .45229 .56252 .4666 .58536 .47503 .57353 .46139 Metetra
.251 .553 .93 r
.57353 .46139 .58536 .47503 .60853 .481 .59679 .46784 Metetra
.345 .618 .952 r
.59679 .46784 .60853 .481 .632 .48445 .62037 .47156 Metetra
.428 .67 .963 r
.62037 .47156 .632 .48445 .65572 .48551 .6442 .47272 Metetra
.494 .706 .965 r
.6442 .47272 .65572 .48551 .67966 .48448 .66827 .47163 Metetra
.543 .73 .961 r
.66827 .47163 .67966 .48448 .7038 .4817 .69254 .46871 Metetra
.578 .745 .956 r
.69254 .46871 .7038 .4817 .72813 .47753 .717 .46433 Metetra
.602 .754 .95 r
.717 .46433 .72813 .47753 .75264 .47229 .74165 .45887 Metetra
.618 .758 .946 r
.74165 .45887 .75264 .47229 .77734 .46626 .76649 .45261 Metetra
.629 .761 .942 r
.76649 .45261 .77734 .46626 .80224 .45966 .79153 .4458 Metetra
.636 .762 .939 r
.79153 .4458 .80224 .45966 .82733 .45265 .81677 .43858 Metetra
.64 .763 .937 r
.81677 .43858 .82733 .45265 .85262 .44534 .84222 .43109 Metetra
.663 .74 .907 r
.19324 .40725 .20635 .41953 .22534 .41295 .21231 .40057 Metetra
.663 .74 .907 r
.21231 .40057 .22534 .41295 .24448 .40634 .23153 .39386 Metetra
.662 .74 .907 r
.23153 .39386 .24448 .40634 .26375 .39974 .25089 .38713 Metetra
.66 .738 .908 r
.25089 .38713 .26375 .39974 .28315 .39326 .27038 .38045 Metetra
.655 .735 .908 r
.27038 .38045 .28315 .39326 .30268 .38708 .29001 .37399 Metetra
.645 .729 .91 r
.29001 .37399 .30268 .38708 .32232 .38153 .30975 .368 Metetra
.624 .717 .913 r
.30975 .368 .32232 .38153 .34205 .37708 .32959 .36294 Metetra
.59 .696 .916 r
.32959 .36294 .34205 .37708 .36188 .37434 .34952 .35942 Metetra
.535 .663 .919 r
.34952 .35942 .36188 .37434 .3818 .37393 .36954 .35814 Metetra
.456 .613 .916 r
.36954 .35814 .3818 .37393 .40184 .37639 .38965 .35975 Metetra
.355 .549 .905 r
.38965 .35975 .40184 .37639 .42206 .38199 .40992 .3647 Metetra
.241 .474 .882 r
.40992 .3647 .42206 .38199 .44253 .39063 .4304 .37301 Metetra
.132 .403 .852 r
.4304 .37301 .44253 .39063 .46331 .4018 .45119 .38425 Metetra
.049 .351 .826 r
.45119 .38425 .46331 .4018 .48448 .41464 .47235 .39755 Metetra
.004 .329 .814 r
.47235 .39755 .48448 .41464 .50608 .42807 .49395 .41173 Metetra
.003 .343 .822 r
.49395 .41173 .50608 .42807 .52813 .44094 .51602 .42551 Metetra
.046 .389 .849 r
.51602 .42551 .52813 .44094 .55063 .45229 .53856 .43775 Metetra
.125 .459 .888 r
.53856 .43775 .55063 .45229 .57353 .46139 .56153 .4476 Metetra
.226 .538 .925 r
.56153 .4476 .57353 .46139 .59679 .46784 .58488 .45458 Metetra
.329 .611 .951 r
.58488 .45458 .59679 .46784 .62037 .47156 .60855 .4586 Metetra
.421 .668 .964 r
.60855 .4586 .62037 .47156 .6442 .47272 .63251 .45984 Metetra
.492 .708 .966 r
.63251 .45984 .6442 .47272 .66827 .47163 .65669 .45868 Metetra
.545 .733 .962 r
.65669 .45868 .66827 .47163 .69254 .46871 .68109 .45556 Metetra
.581 .747 .956 r
.68109 .45556 .69254 .46871 .717 .46433 .70569 .45095 Metetra
.605 .756 .95 r
.70569 .45095 .717 .46433 .74165 .45887 .73048 .44524 Metetra
.621 .76 .945 r
.73048 .44524 .74165 .45887 .76649 .45261 .75546 .43874 Metetra
.631 .762 .941 r
.75546 .43874 .76649 .45261 .79153 .4458 .78064 .4317 Metetra
.637 .763 .939 r
.78064 .4317 .79153 .4458 .81677 .43858 .80603 .42428 Metetra
.641 .763 .937 r
.80603 .42428 .81677 .43858 .84222 .43109 .83164 .41661 Metetra
.663 .74 .907 r
.17994 .39481 .19324 .40725 .21231 .40057 .1991 .38803 Metetra
.663 .74 .907 r
.1991 .38803 .21231 .40057 .23153 .39386 .2184 .38121 Metetra
.663 .74 .907 r
.2184 .38121 .23153 .39386 .25089 .38713 .23785 .37436 Metetra
.662 .739 .907 r
.23785 .37436 .25089 .38713 .27038 .38045 .25743 .36753 Metetra
.658 .737 .908 r
.25743 .36753 .27038 .38045 .29001 .37399 .27716 .36081 Metetra
.651 .732 .909 r
.27716 .36081 .29001 .37399 .30975 .368 .297 .35443 Metetra
.635 .723 .911 r
.297 .35443 .30975 .368 .32959 .36294 .31695 .34879 Metetra
.605 .705 .914 r
.31695 .34879 .32959 .36294 .34952 .35942 .337 .34448 Metetra
.556 .674 .917 r
.337 .34448 .34952 .35942 .36954 .35814 .35712 .34226 Metetra
.482 .628 .916 r
.35712 .34226 .36954 .35814 .38965 .35975 .37732 .34292 Metetra
.38 .562 .906 r
.37732 .34292 .38965 .35975 .40992 .3647 .39764 .34706 Metetra
.258 .482 .883 r
.39764 .34706 .40992 .3647 .4304 .37301 .41815 .35491 Metetra
.136 .401 .85 r
.41815 .35491 .4304 .37301 .45119 .38425 .43893 .36614 Metetra
.036 .337 .817 r
.43893 .36614 .45119 .38425 .47235 .39755 .46008 .3799 Metetra
0 .305 .799 r
.46008 .3799 .47235 .39755 .49395 .41173 .48167 .3949 Metetra
0 .312 .803 r
.48167 .3949 .49395 .41173 .51602 .42551 .50374 .40971 Metetra
.004 .357 .831 r
.50374 .40971 .51602 .42551 .53856 .43775 .52631 .42297 Metetra
.087 .433 .874 r
.52631 .42297 .53856 .43775 .56153 .4476 .54934 .43368 Metetra
.196 .521 .918 r
.54934 .43368 .56153 .4476 .58488 .45458 .57278 .44126 Metetra
.312 .603 .95 r
.57278 .44126 .58488 .45458 .60855 .4586 .59656 .44559 Metetra
.413 .667 .965 r
.59656 .44559 .60855 .4586 .63251 .45984 .62062 .44689 Metetra
.492 .71 .967 r
.62062 .44689 .63251 .45984 .65669 .45868 .64494 .44562 Metetra
.547 .736 .963 r
.64494 .44562 .65669 .45868 .68109 .45556 .66947 .44228 Metetra
.585 .75 .956 r
.66947 .44228 .68109 .45556 .70569 .45095 .69419 .43739 Metetra
.609 .758 .95 r
.69419 .43739 .70569 .45095 .73048 .44524 .71912 .43139 Metetra
.624 .761 .945 r
.71912 .43139 .73048 .44524 .75546 .43874 .74425 .42463 Metetra
.633 .763 .941 r
.74425 .42463 .75546 .43874 .78064 .4317 .76958 .41735 Metetra
.639 .763 .938 r
.76958 .41735 .78064 .4317 .80603 .42428 .79512 .40973 Metetra
.642 .763 .936 r
.79512 .40973 .80603 .42428 .83164 .41661 .82088 .40186 Metetra
.663 .74 .907 r
.16647 .38219 .17994 .39481 .1991 .38803 .18571 .37532 Metetra
.663 .74 .907 r
.18571 .37532 .1991 .38803 .2184 .38121 .2051 .3684 Metetra
.663 .74 .907 r
.2051 .3684 .2184 .38121 .23785 .37436 .22463 .36144 Metetra
.663 .74 .907 r
.22463 .36144 .23785 .37436 .25743 .36753 .2443 .35446 Metetra
.661 .738 .908 r
.2443 .35446 .25743 .36753 .27716 .36081 .26412 .34753 Metetra
.655 .735 .908 r
.26412 .34753 .27716 .36081 .297 .35443 .28407 .34081 Metetra
.644 .728 .91 r
.28407 .34081 .297 .35443 .31695 .34879 .30414 .33463 Metetra
.62 .713 .912 r
.30414 .33463 .31695 .34879 .337 .34448 .3243 .32955 Metetra
.577 .686 .915 r
.3243 .32955 .337 .34448 .35712 .34226 .34453 .32635 Metetra
.509 .643 .916 r
.34453 .32635 .35712 .34226 .37732 .34292 .36483 .32593 Metetra
.409 .578 .908 r
.36483 .32593 .37732 .34292 .39764 .34706 .38522 .3291 Metetra
.281 .494 .886 r
.38522 .3291 .39764 .34706 .41815 .35491 .40576 .3363 Metetra
.145 .402 .849 r
.40576 .3363 .41815 .35491 .43893 .36614 .42654 .34741 Metetra
.026 .324 .809 r
.42654 .34741 .43893 .36614 .46008 .3799 .44766 .36161 Metetra
0 .278 .782 r
.44766 .36161 .46008 .3799 .48167 .3949 .46923 .37752 Metetra
0 .277 .78 r
.46923 .37752 .48167 .3949 .50374 .40971 .4913 .3935 Metetra
0 .321 .807 r
.4913 .3935 .50374 .40971 .52631 .42297 .51389 .40795 Metetra
.042 .402 .856 r
.51389 .40795 .52631 .42297 .54934 .43368 .53697 .41965 Metetra
.162 .501 .909 r
.53697 .41965 .54934 .43368 .57278 .44126 .56048 .4279 Metetra
.292 .595 .948 r
.56048 .4279 .57278 .44126 .59656 .44559 .58437 .43256 Metetra
.406 .667 .967 r
.58437 .43256 .59656 .44559 .62062 .44689 .60855 .43391 Metetra
.492 .713 .969 r
.60855 .43391 .62062 .44689 .64494 .44562 .63299 .43246 Metetra
.551 .74 .964 r
.63299 .43246 .64494 .44562 .66947 .44228 .65765 .42884 Metetra
.589 .753 .956 r
.65765 .42884 .66947 .44228 .69419 .43739 .68251 .42363 Metetra
.613 .76 .949 r
.68251 .42363 .69419 .43739 .71912 .43139 .70758 .41732 Metetra
.627 .763 .944 r
.70758 .41732 .71912 .43139 .74425 .42463 .73285 .41028 Metetra
.635 .763 .94 r
.73285 .41028 .74425 .42463 .76958 .41735 .75833 .40275 Metetra
.64 .763 .938 r
.75833 .40275 .76958 .41735 .79512 .40973 .78402 .39491 Metetra
.643 .763 .936 r
.78402 .39491 .79512 .40973 .82088 .40186 .80994 .38686 Metetra
.663 .74 .907 r
.15281 .3694 .16647 .38219 .18571 .37532 .17213 .36243 Metetra
.663 .74 .907 r
.17213 .36243 .18571 .37532 .2051 .3684 .1916 .35542 Metetra
.663 .74 .907 r
.1916 .35542 .2051 .3684 .22463 .36144 .21122 .34835 Metetra
.663 .74 .907 r
.21122 .34835 .22463 .36144 .2443 .35446 .23098 .34124 Metetra
.662 .739 .907 r
.23098 .34124 .2443 .35446 .26412 .34753 .25089 .33414 Metetra
.659 .737 .908 r
.25089 .33414 .26412 .34753 .28407 .34081 .27095 .32713 Metetra
.651 .732 .909 r
.27095 .32713 .28407 .34081 .30414 .33463 .29113 .32049 Metetra
.633 .721 .91 r
.29113 .32049 .30414 .33463 .3243 .32955 .31141 .31467 Metetra
.598 .698 .913 r
.31141 .31467 .3243 .32955 .34453 .32635 .33177 .31046 Metetra
.537 .659 .915 r
.33177 .31046 .34453 .32635 .36483 .32593 .35218 .30883 Metetra
.442 .597 .909 r
.35218 .30883 .36483 .32593 .38522 .3291 .37266 .31083 Metetra
.311 .509 .889 r
.37266 .31083 .38522 .3291 .40576 .3363 .39324 .31718 Metetra
.159 .407 .85 r
.39324 .31718 .40576 .3363 .42654 .34741 .41402 .328 Metetra
.019 .312 .801 r
.41402 .328 .42654 .34741 .44766 .36161 .43512 .34259 Metetra
0 .25 .763 r
.43512 .34259 .44766 .36161 .46923 .37752 .45664 .3595 Metetra
0 .237 .753 r
.45664 .3595 .46923 .37752 .4913 .3935 .47868 .37684 Metetra
0 .278 .779 r
.47868 .37684 .4913 .3935 .51389 .40795 .50128 .39269 Metetra
0 .365 .833 r
.50128 .39269 .51389 .40795 .53697 .41965 .5244 .40555 Metetra
.123 .478 .897 r
.5244 .40555 .53697 .41965 .56048 .4279 .54799 .41457 Metetra
.27 .587 .946 r
.54799 .41457 .56048 .4279 .58437 .43256 .57198 .41957 Metetra
.399 .668 .968 r
.57198 .41957 .58437 .43256 .60855 .43391 .59628 .4209 Metetra
.494 .718 .971 r
.59628 .4209 .60855 .43391 .63299 .43246 .62084 .41921 Metetra
.556 .744 .964 r
.62084 .41921 .63299 .43246 .65765 .42884 .64563 .41524 Metetra
.594 .757 .956 r
.64563 .41524 .65765 .42884 .68251 .42363 .67063 .40966 Metetra
.617 .762 .948 r
.67063 .40966 .68251 .42363 .70758 .41732 .69584 .403 Metetra
.63 .764 .943 r
.69584 .403 .70758 .41732 .73285 .41028 .72125 .39566 Metetra
.638 .764 .939 r
.72125 .39566 .73285 .41028 .75833 .40275 .74689 .38788 Metetra
.642 .764 .937 r
.74689 .38788 .75833 .40275 .78402 .39491 .77274 .37982 Metetra
.644 .763 .935 r
.77274 .37982 .78402 .39491 .80994 .38686 .79882 .37159 Metetra
.663 .74 .907 r
.13896 .35643 .15281 .3694 .17213 .36243 .15836 .34937 Metetra
.663 .74 .907 r
.15836 .34937 .17213 .36243 .1916 .35542 .17792 .34225 Metetra
.663 .74 .907 r
.17792 .34225 .1916 .35542 .21122 .34835 .19762 .33508 Metetra
.663 .74 .907 r
.19762 .33508 .21122 .34835 .23098 .34124 .21747 .32786 Metetra
.663 .74 .907 r
.21747 .32786 .23098 .34124 .25089 .33414 .23748 .32061 Metetra
.661 .739 .907 r
.23748 .32061 .25089 .33414 .27095 .32713 .25763 .31338 Metetra
.657 .735 .908 r
.25763 .31338 .27095 .32713 .29113 .32049 .27792 .30634 Metetra
.644 .727 .909 r
.27792 .30634 .29113 .32049 .31141 .31467 .29833 .29987 Metetra
.618 .71 .911 r
.29833 .29987 .31141 .31467 .33177 .31046 .31882 .29465 Metetra
.566 .676 .913 r
.31882 .29465 .33177 .31046 .35218 .30883 .33937 .29173 Metetra
.479 .618 .91 r
.33937 .29173 .35218 .30883 .37266 .31083 .35995 .29233 Metetra
.348 .53 .893 r
.35995 .29233 .37266 .31083 .39324 .31718 .3806 .29753 Metetra
.183 .417 .853 r
.3806 .29753 .39324 .31718 .41402 .328 .40138 .30783 Metetra
.017 .303 .795 r
.40138 .30783 .41402 .328 .43512 .34259 .42244 .32272 Metetra
0 .221 .743 r
.42244 .32272 .43512 .34259 .45664 .3595 .44391 .34072 Metetra
0 .192 .721 r
.44391 .34072 .45664 .3595 .47868 .37684 .4659 .35965 Metetra
0 .227 .742 r
.4659 .35965 .47868 .37684 .50128 .39269 .48848 .37718 Metetra
0 .32 .803 r
.48848 .37718 .50128 .39269 .5244 .40555 .51163 .39143 Metetra
.076 .45 .881 r
.51163 .39143 .5244 .40555 .54799 .41457 .5353 .40133 Metetra
.246 .578 .943 r
.5353 .40133 .54799 .41457 .57198 .41957 .55938 .40666 Metetra
.394 .671 .971 r
.55938 .40666 .57198 .41957 .59628 .4209 .5838 .40791 Metetra
.498 .724 .973 r
.5838 .40791 .59628 .4209 .62084 .41921 .60849 .40588 Metetra
.563 .75 .965 r
.60849 .40588 .62084 .41921 .64563 .41524 .63341 .40147 Metetra
.601 .761 .955 r
.63341 .40147 .64563 .41524 .67063 .40966 .65855 .39545 Metetra
.623 .764 .947 r
.65855 .39545 .67063 .40966 .69584 .403 .6839 .38841 Metetra
.634 .765 .942 r
.6839 .38841 .69584 .403 .72125 .39566 .70946 .38075 Metetra
.64 .764 .938 r
.70946 .38075 .72125 .39566 .74689 .38788 .73525 .37272 Metetra
.643 .764 .936 r
.73525 .37272 .74689 .38788 .77274 .37982 .76126 .36445 Metetra
.645 .763 .935 r
.76126 .36445 .77274 .37982 .79882 .37159 .78751 .35604 Metetra
.663 .74 .907 r
.12492 .34328 .13896 .35643 .15836 .34937 .1444 .33612 Metetra
.663 .74 .907 r
.1444 .33612 .15836 .34937 .17792 .34225 .16404 .3289 Metetra
.663 .74 .907 r
.16404 .3289 .17792 .34225 .19762 .33508 .18383 .32162 Metetra
.663 .74 .907 r
.18383 .32162 .19762 .33508 .21747 .32786 .20377 .3143 Metetra
.663 .74 .907 r
.20377 .3143 .21747 .32786 .23748 .32061 .22387 .30692 Metetra
.663 .74 .907 r
.22387 .30692 .23748 .32061 .25763 .31338 .24412 .29952 Metetra
.66 .738 .907 r
.24412 .29952 .25763 .31338 .27792 .30634 .26451 .29218 Metetra
.653 .733 .908 r
.26451 .29218 .27792 .30634 .29833 .29987 .28505 .28515 Metetra
.635 .72 .909 r
.28505 .28515 .29833 .29987 .31882 .29465 .30568 .279 Metetra
.595 .693 .911 r
.30568 .279 .31882 .29465 .33937 .29173 .32638 .27471 Metetra
.519 .642 .91 r
.32638 .27471 .33937 .29173 .35995 .29233 .34709 .27368 Metetra
.393 .556 .897 r
.34709 .27368 .35995 .29233 .3806 .29753 .36782 .27738 Metetra
.217 .434 .858 r
.36782 .27738 .3806 .29753 .40138 .30783 .38863 .28679 Metetra
.024 .299 .791 r
.38863 .28679 .40138 .30783 .42244 .32272 .40965 .30182 Metetra
0 .19 .722 r
.40965 .30182 .42244 .32272 .44391 .34072 .43103 .32102 Metetra
0 .142 .683 r
.43103 .32102 .44391 .34072 .4659 .35965 .45295 .34186 Metetra
0 .166 .696 r
.45295 .34186 .4659 .35965 .48848 .37718 .47549 .36146 Metetra
0 .264 .763 r
.47549 .36146 .48848 .37718 .51163 .39143 .49866 .37739 Metetra
.019 .416 .86 r
.49866 .37739 .51163 .39143 .5353 .40133 .52239 .38829 Metetra
.219 .569 .939 r
.52239 .38829 .5353 .40133 .55938 .40666 .54657 .39393 Metetra
.391 .676 .974 r
.54657 .39393 .55938 .40666 .5838 .40791 .57111 .39497 Metetra
.506 .733 .975 r
.57111 .39497 .5838 .40791 .60849 .40588 .59593 .39246 Metetra
.573 .756 .965 r
.59593 .39246 .60849 .40588 .63341 .40147 .62099 .3875 Metetra
.609 .764 .954 r
.62099 .3875 .63341 .40147 .65855 .39545 .64626 .38098 Metetra
.628 .766 .945 r
.64626 .38098 .65855 .39545 .6839 .38841 .67176 .37353 Metetra
.638 .765 .94 r
.67176 .37353 .6839 .38841 .70946 .38075 .69747 .36555 Metetra
.642 .764 .937 r
.69747 .36555 .70946 .38075 .73525 .37272 .72341 .35726 Metetra
.644 .764 .935 r
.72341 .35726 .73525 .37272 .76126 .36445 .74959 .3488 Metetra
.645 .763 .935 r
.74959 .3488 .76126 .36445 .78751 .35604 .776 .34021 Metetra
.663 .74 .907 r
.11067 .32994 .12492 .34328 .1444 .33612 .13024 .32268 Metetra
.663 .74 .907 r
.13024 .32268 .1444 .33612 .16404 .3289 .14997 .31536 Metetra
.663 .74 .907 r
.14997 .31536 .16404 .3289 .18383 .32162 .16984 .30798 Metetra
.663 .74 .907 r
.16984 .30798 .18383 .32162 .20377 .3143 .18987 .30054 Metetra
.663 .74 .907 r
.18987 .30054 .20377 .3143 .22387 .30692 .21006 .29305 Metetra
.663 .74 .907 r
.21006 .29305 .22387 .30692 .24412 .29952 .2304 .28551 Metetra
.662 .739 .907 r
.2304 .28551 .24412 .29952 .26451 .29218 .2509 .27795 Metetra
.659 .737 .907 r
.2509 .27795 .26451 .29218 .28505 .28515 .27155 .27051 Metetra
.648 .729 .908 r
.27155 .27051 .28505 .28515 .30568 .279 .29233 .26356 Metetra
.621 .709 .909 r
.29233 .26356 .30568 .279 .32638 .27471 .31318 .25793 Metetra
.56 .667 .909 r
.31318 .25793 .32638 .27471 .34709 .27368 .33406 .25506 Metetra
.447 .588 .9 r
.33406 .25506 .34709 .27368 .36782 .27738 .35492 .2568 Metetra
.267 .462 .865 r
.35492 .2568 .36782 .27738 .38863 .28679 .37578 .26482 Metetra
.045 .303 .792 r
.37578 .26482 .38863 .28679 .40965 .30182 .39675 .27967 Metetra
0 .161 .701 r
.39675 .27967 .40965 .30182 .43103 .32102 .41804 .30013 Metetra
0 .084 .639 r
.41804 .30013 .43103 .32102 .45295 .34186 .43984 .32331 Metetra
0 .094 .637 r
.43984 .32331 .45295 .34186 .47549 .36146 .46231 .34554 Metetra
0 .195 .708 r
.46231 .34554 .47549 .36146 .49866 .37739 .48547 .36358 Metetra
0 .374 .83 r
.48547 .36358 .49866 .37739 .52239 .38829 .50925 .37563 Metetra
.191 .562 .935 r
.50925 .37563 .52239 .38829 .54657 .39393 .53354 .38149 Metetra
.393 .687 .978 r
.53354 .38149 .54657 .39393 .57111 .39497 .5582 .38211 Metetra
.519 .744 .977 r
.5582 .38211 .57111 .39497 .59593 .39246 .58315 .37892 Metetra
.586 .764 .964 r
.58315 .37892 .59593 .39246 .62099 .3875 .60834 .37327 Metetra
.619 .768 .951 r
.60834 .37327 .62099 .3875 .64626 .38098 .63376 .36618 Metetra
.634 .767 .943 r
.63376 .36618 .64626 .38098 .67176 .37353 .6594 .35831 Metetra
.641 .766 .938 r
.6594 .35831 .67176 .37353 .69747 .36555 .68527 .35002 Metetra
.644 .764 .936 r
.68527 .35002 .69747 .36555 .72341 .35726 .71137 .3415 Metetra
.645 .763 .935 r
.71137 .3415 .72341 .35726 .74959 .3488 .73771 .33285 Metetra
.646 .763 .934 r
.73771 .33285 .74959 .3488 .776 .34021 .7643 .32409 Metetra
.663 .74 .907 r
.09623 .31642 .11067 .32994 .13024 .32268 .11588 .30905 Metetra
.663 .74 .907 r
.11588 .30905 .13024 .32268 .14997 .31536 .13569 .30162 Metetra
.663 .74 .907 r
.13569 .30162 .14997 .31536 .16984 .30798 .15565 .29414 Metetra
.663 .74 .907 r
.15565 .29414 .16984 .30798 .18987 .30054 .17577 .28659 Metetra
.663 .74 .907 r
.17577 .28659 .18987 .30054 .21006 .29305 .19605 .27899 Metetra
.663 .74 .907 r
.19605 .27899 .21006 .29305 .2304 .28551 .21649 .27133 Metetra
.663 .74 .907 r
.21649 .27133 .2304 .28551 .2509 .27795 .23709 .26362 Metetra
.662 .739 .907 r
.23709 .26362 .2509 .27795 .27155 .27051 .25784 .2559 Metetra
.658 .735 .907 r
.25784 .2559 .27155 .27051 .29233 .26356 .27875 .24835 Metetra
.642 .723 .907 r
.27875 .24835 .29233 .26356 .31318 .25793 .29977 .24154 Metetra
.601 .692 .907 r
.29977 .24154 .31318 .25793 .33406 .25506 .32084 .23671 Metetra
.508 .626 .902 r
.32084 .23671 .33406 .25506 .35492 .2568 .34187 .23598 Metetra
.336 .502 .874 r
.34187 .23598 .35492 .2568 .37578 .26482 .36283 .24183 Metetra
.087 .321 .798 r
.36283 .24183 .37578 .26482 .39675 .27967 .38379 .25591 Metetra
0 .136 .683 r
.38379 .25591 .39675 .27967 .41804 .30013 .40494 .27764 Metetra
0 .02 .587 r
.40494 .27764 .41804 .30013 .43984 .32331 .42658 .30378 Metetra
0 .006 .562 r
.42658 .30378 .43984 .32331 .46231 .34554 .44892 .32949 Metetra
0 .108 .632 r
.44892 .32949 .46231 .34554 .48547 .36358 .47204 .35026 Metetra
0 .321 .786 r
.47204 .35026 .48547 .36358 .50925 .37563 .49588 .36362 Metetra
.162 .559 .931 r
.49588 .36362 .50925 .37563 .53354 .38149 .52027 .36948 Metetra
.406 .705 .983 r
.52027 .36948 .53354 .38149 .5582 .38211 .54506 .36934 Metetra
.54 .759 .978 r
.54506 .36934 .5582 .38211 .58315 .37892 .57014 .36519 Metetra
.602 .771 .961 r
.57014 .36519 .58315 .37892 .60834 .37327 .59548 .3587 Metetra
.629 .771 .948 r
.59548 .3587 .60834 .37327 .63376 .36618 .62104 .35101 Metetra
.639 .767 .941 r
.62104 .35101 .63376 .36618 .6594 .35831 .64683 .34273 Metetra
.644 .765 .937 r
.64683 .34273 .6594 .35831 .68527 .35002 .67286 .33416 Metetra
.645 .764 .935 r
.67286 .33416 .68527 .35002 .71137 .3415 .69912 .32544 Metetra
.646 .763 .934 r
.69912 .32544 .71137 .3415 .73771 .33285 .72563 .31661 Metetra
.646 .763 .934 r
.72563 .31661 .73771 .33285 .7643 .32409 .75239 .30769 Metetra
.663 .74 .907 r
.08158 .3027 .09623 .31642 .11588 .30905 .10132 .29522 Metetra
.663 .74 .907 r
.10132 .29522 .11588 .30905 .13569 .30162 .12121 .28769 Metetra
.663 .74 .907 r
.12121 .28769 .13569 .30162 .15565 .29414 .14126 .28009 Metetra
.663 .74 .907 r
.14126 .28009 .15565 .29414 .17577 .28659 .16147 .27244 Metetra
.663 .74 .907 r
.16147 .27244 .17577 .28659 .19605 .27899 .18184 .26473 Metetra
.663 .74 .907 r
.18184 .26473 .19605 .27899 .21649 .27133 .20237 .25695 Metetra
.663 .74 .907 r
.20237 .25695 .21649 .27133 .23709 .26362 .22306 .24911 Metetra
.663 .74 .907 r
.22306 .24911 .23709 .26362 .25784 .2559 .24392 .24122 Metetra
.662 .739 .907 r
.24392 .24122 .25784 .2559 .27875 .24835 .26495 .23332 Metetra
.657 .734 .906 r
.26495 .23332 .27875 .24835 .29977 .24154 .28612 .22565 Metetra
.635 .716 .905 r
.28612 .22565 .29977 .24154 .32084 .23671 .30739 .21901 Metetra
.573 .667 .902 r
.30739 .21901 .32084 .23671 .34187 .23598 .32866 .21535 Metetra
.428 .558 .884 r
.32866 .21535 .34187 .23598 .36283 .24183 .3498 .21793 Metetra
.165 .363 .814 r
.3498 .21793 .36283 .24183 .38379 .25591 .37079 .23008 Metetra
0 .124 .673 r
.37079 .23008 .38379 .25591 .40494 .27764 .3918 .25279 Metetra
0 0 .529 r
.3918 .25279 .40494 .27764 .42658 .30378 .41319 .28282 Metetra
0 0 .464 r
.41319 .28282 .42658 .30378 .44892 .32949 .43531 .31347 Metetra
0 0 .524 r
.43531 .31347 .44892 .32949 .47204 .35026 .45835 .33797 Metetra
0 .254 .719 r
.45835 .33797 .47204 .35026 .49588 .36362 .48223 .35273 Metetra
.142 .569 .927 r
.48223 .35273 .49588 .36362 .52027 .36948 .50674 .35807 Metetra
.436 .736 .99 r
.50674 .35807 .52027 .36948 .54506 .36934 .53167 .35658 Metetra
.57 .776 .976 r
.53167 .35658 .54506 .36934 .57014 .36519 .5569 .35111 Metetra
.62 .777 .956 r
.5569 .35111 .57014 .36519 .59548 .3587 .58238 .34368 Metetra
.638 .771 .944 r
.58238 .34368 .59548 .3587 .62104 .35101 .60809 .3354 Metetra
.643 .766 .938 r
.60809 .3354 .62104 .35101 .64683 .34273 .63404 .32676 Metetra
.645 .764 .935 r
.63404 .32676 .64683 .34273 .67286 .33416 .66023 .31797 Metetra
.646 .763 .934 r
.66023 .31797 .67286 .33416 .69912 .32544 .68666 .30906 Metetra
.646 .763 .934 r
.68666 .30906 .69912 .32544 .72563 .31661 .71334 .30007 Metetra
.646 .762 .934 r
.71334 .30007 .72563 .31661 .75239 .30769 .74027 .29099 Metetra
.663 .74 .907 r
.06671 .28878 .08158 .3027 .10132 .29522 .08654 .28119 Metetra
.663 .74 .907 r
.08654 .28119 .10132 .29522 .12121 .28769 .10652 .27355 Metetra
.663 .74 .907 r
.10652 .27355 .12121 .28769 .14126 .28009 .12666 .26585 Metetra
.663 .74 .907 r
.12666 .26585 .14126 .28009 .16147 .27244 .14695 .25808 Metetra
.663 .74 .907 r
.14695 .25808 .16147 .27244 .18184 .26473 .16741 .25025 Metetra
.663 .74 .907 r
.16741 .25025 .18184 .26473 .20237 .25695 .18804 .24236 Metetra
.663 .74 .907 r
.18804 .24236 .20237 .25695 .22306 .24911 .20883 .23441 Metetra
.663 .74 .907 r
.20883 .23441 .22306 .24911 .24392 .24122 .22979 .22639 Metetra
.663 .74 .907 r
.22979 .22639 .24392 .24122 .26495 .23332 .25092 .21831 Metetra
.663 .739 .907 r
.25092 .21831 .26495 .23332 .28612 .22565 .27221 .2102 Metetra
.657 .733 .905 r
.27221 .2102 .28612 .22565 .30739 .21901 .29367 .20231 Metetra
.63 .707 .901 r
.29367 .20231 .30739 .21901 .32866 .21535 .3152 .19574 Metetra
.537 .63 .891 r
.3152 .19574 .32866 .21535 .3498 .21793 .33665 .19366 Metetra
.301 .446 .84 r
.33665 .19366 .3498 .21793 .37079 .23008 .35782 .20157 Metetra
0 .143 .682 r
.35782 .20157 .37079 .23008 .3918 .25279 .37871 .22409 Metetra
0 0 .468 r
.37871 .22409 .3918 .25279 .41319 .28282 .39972 .25945 Metetra
.625 .223 0 r
.39972 .25945 .41319 .28282 .43531 .31347 .42147 .29791 Metetra
.62 .14 0 r
.42147 .29791 .43531 .31347 .45835 .33797 .44435 .32789 Metetra
0 .176 .616 r
.44435 .32789 .45835 .33797 .48223 .35273 .46829 .34375 Metetra
.155 .613 .931 r
.46829 .34375 .48223 .35273 .50674 .35807 .49295 .34731 Metetra
.498 .781 .995 r
.49295 .34731 .50674 .35807 .53167 .35658 .51804 .34351 Metetra
.607 .79 .969 r
.51804 .34351 .53167 .35658 .5569 .35111 .54342 .33638 Metetra
.637 .777 .948 r
.54342 .33638 .5569 .35111 .58238 .34368 .56905 .32805 Metetra
.644 .768 .939 r
.56905 .32805 .58238 .34368 .60809 .3354 .59491 .31932 Metetra
.645 .764 .935 r
.59491 .31932 .60809 .3354 .63404 .32676 .62102 .31044 Metetra
.646 .763 .934 r
.62102 .31044 .63404 .32676 .66023 .31797 .64737 .30146 Metetra
.646 .763 .934 r
.64737 .30146 .66023 .31797 .68666 .30906 .67398 .29239 Metetra
.646 .762 .934 r
.67398 .29239 .68666 .30906 .71334 .30007 .70083 .28324 Metetra
.646 .762 .934 r
.70083 .28324 .71334 .30007 .74027 .29099 .72794 .274 Metetra
.663 .74 .907 r
.05164 .27466 .06671 .28878 .08654 .28119 .07154 .26696 Metetra
.663 .74 .907 r
.07154 .26696 .08654 .28119 .10652 .27355 .09161 .25921 Metetra
.663 .74 .907 r
.09161 .25921 .10652 .27355 .12666 .26585 .11184 .25139 Metetra
.663 .74 .907 r
.11184 .25139 .12666 .26585 .14695 .25808 .13222 .24351 Metetra
.663 .74 .907 r
.13222 .24351 .14695 .25808 .16741 .25025 .15278 .23556 Metetra
.663 .74 .907 r
.15278 .23556 .16741 .25025 .18804 .24236 .1735 .22755 Metetra
.663 .74 .907 r
.1735 .22755 .18804 .24236 .20883 .23441 .19438 .21948 Metetra
.663 .74 .907 r
.19438 .21948 .20883 .23441 .22979 .22639 .21544 .21134 Metetra
.663 .74 .907 r
.21544 .21134 .22979 .22639 .25092 .21831 .23667 .20313 Metetra
.663 .74 .907 r
.23667 .20313 .25092 .21831 .27221 .2102 .25807 .19486 Metetra
.663 .74 .907 r
.25807 .19486 .27221 .2102 .29367 .20231 .27965 .18652 Metetra
.661 .734 .904 r
.27965 .18652 .29367 .20231 .3152 .19574 .3014 .17825 Metetra
.636 .703 .895 r
.3014 .17825 .3152 .19574 .33665 .19366 .32325 .17115 Metetra
.508 .584 .867 r
.32325 .17115 .33665 .19366 .35782 .20157 .34491 .17051 Metetra
.106 .257 .737 r
.34491 .17051 .35782 .20157 .37871 .22409 .36591 .18797 Metetra
0 0 .424 r
.36591 .18797 .37871 .22409 .39972 .25945 .38634 .23068 Metetra
.741 .371 0 r
.38634 .23068 .39972 .25945 .42147 .29791 .40731 .28446 Metetra
.776 .302 0 r
.40731 .28446 .42147 .29791 .44435 .32789 .42989 .32336 Metetra
.467 0 0 r
.42989 .32336 .44435 .32789 .46829 .34375 .45397 .33779 Metetra
.287 .744 .953 r
.45397 .33779 .46829 .34375 .49295 .34731 .47887 .33642 Metetra
.593 .823 .987 r
.47887 .33642 .49295 .34731 .51804 .34351 .50415 .32925 Metetra
.639 .787 .953 r
.50415 .32925 .51804 .34351 .54342 .33638 .52969 .32068 Metetra
.645 .769 .939 r
.52969 .32068 .54342 .33638 .56905 .32805 .55548 .31181 Metetra
.646 .764 .935 r
.55548 .31181 .56905 .32805 .59491 .31932 .5815 .30285 Metetra
.646 .763 .934 r
.5815 .30285 .59491 .31932 .62102 .31044 .60777 .29379 Metetra
.646 .762 .934 r
.60777 .29379 .62102 .31044 .64737 .30146 .6343 .28465 Metetra
.646 .762 .934 r
.6343 .28465 .64737 .30146 .67398 .29239 .66107 .27543 Metetra
.646 .762 .934 r
.66107 .27543 .67398 .29239 .70083 .28324 .6881 .26611 Metetra
.646 .762 .934 r
.6881 .26611 .70083 .28324 .72794 .274 .71539 .25671 Metetra
.663 .74 .907 r
.03634 .26033 .05164 .27466 .07154 .26696 .05633 .25252 Metetra
.663 .74 .907 r
.05633 .25252 .07154 .26696 .09161 .25921 .07648 .24465 Metetra
.663 .74 .907 r
.07648 .24465 .09161 .25921 .11184 .25139 .0968 .23672 Metetra
.663 .74 .907 r
.0968 .23672 .11184 .25139 .13222 .24351 .11728 .22872 Metetra
.663 .74 .907 r
.11728 .22872 .13222 .24351 .15278 .23556 .13792 .22065 Metetra
.663 .74 .907 r
.13792 .22065 .15278 .23556 .1735 .22755 .15874 .21252 Metetra
.663 .74 .907 r
.15874 .21252 .1735 .22755 .19438 .21948 .17972 .20433 Metetra
.663 .74 .907 r
.17972 .20433 .19438 .21948 .21544 .21134 .20088 .19607 Metetra
.663 .74 .907 r
.20088 .19607 .21544 .21134 .23667 .20313 .22221 .18773 Metetra
.663 .74 .907 r
.22221 .18773 .23667 .20313 .25807 .19486 .24371 .17933 Metetra
.663 .74 .907 r
.24371 .17933 .25807 .19486 .27965 .18652 .2654 .17086 Metetra
.663 .74 .907 r
.2654 .17086 .27965 .18652 .3014 .17825 .28727 .16232 Metetra
.663 .738 .906 r
.28727 .16232 .3014 .17825 .32325 .17115 .30932 .15371 Metetra
.661 .718 .891 r
.30932 .15371 .32325 .17115 .34491 .17051 .33155 .14504 Metetra
.608 .608 .825 r
.33155 .14504 .34491 .17051 .36591 .18797 .35383 .13872 Metetra
0 .032 .566 r
.35383 .13872 .36591 .18797 .38634 .23068 .3742 .1749 Metetra
.836 .554 .009 r
.3742 .1749 .38634 .23068 .40731 .28446 .39213 .28876 Metetra
.837 .393 .168 r
.39213 .28876 .40731 .28446 .42989 .32336 .4145 .3354 Metetra
.143 0 0 r
.4145 .3354 .42989 .32336 .45397 .33779 .43935 .3308 Metetra
.593 .884 .991 r
.43935 .3308 .45397 .33779 .47887 .33642 .46457 .32206 Metetra
.644 .789 .951 r
.46457 .32206 .47887 .33642 .50415 .32925 .49003 .31319 Metetra
.646 .766 .936 r
.49003 .31319 .50415 .32925 .52969 .32068 .51573 .30423 Metetra
.646 .763 .934 r
.51573 .30423 .52969 .32068 .55548 .31181 .54167 .29519 Metetra
.646 .762 .934 r
.54167 .29519 .55548 .31181 .5815 .30285 .56786 .28607 Metetra
.646 .762 .934 r
.56786 .28607 .5815 .30285 .60777 .29379 .59429 .27685 Metetra
.646 .762 .934 r
.59429 .27685 .60777 .29379 .6343 .28465 .62099 .26755 Metetra
.646 .762 .934 r
.62099 .26755 .6343 .28465 .66107 .27543 .64793 .25816 Metetra
.646 .762 .934 r
.64793 .25816 .66107 .27543 .6881 .26611 .67514 .24868 Metetra
.646 .762 .934 r
.67514 .24868 .6881 .26611 .71539 .25671 .70262 .2391 Metetra
0 g
.25 Mabswid
.68874 0 m
.96935 .42924 L
s
.96935 .42924 m
1 .6535 L
s
1 .6535 m
.70298 .24544 L
s
.70298 .24544 m
.68874 0 L
s
.03716 .25514 m
0 .48963 L
s
0 .48963 m
.70298 .24544 L
s
.70298 .24544 m
.68874 0 L
s
.68874 0 m
.03716 .25514 L
s
.03716 .25514 m
.68874 0 L
s
.12091 .22234 m
.12558 .22696 L
s
[(80)] .11158 .21312 1 .98897 Mshowa
.2971 .15335 m
.30147 .15825 L
s
[(100)] .28836 .14356 .89202 1 Mshowa
.4859 .07942 m
.48992 .08461 L
s
[(120)] .47788 .06904 .77288 1 Mshowa
.68874 0 m
.69233 .00549 L
s
[(140)] .68156 -0.01098 .65374 1 Mshowa
.125 Mabswid
.16384 .20553 m
.1666 .20834 L
s
.20751 .18843 m
.21022 .19129 L
s
.25192 .17104 m
.25459 .17394 L
s
.34306 .13535 m
.34564 .13834 L
s
.38984 .11704 m
.39236 .12006 L
s
.43745 .0984 m
.43991 .10147 L
s
.53524 .06011 m
.53759 .06327 L
s
.58547 .04044 m
.58776 .04364 L
s
.63663 .0204 m
.63885 .02365 L
s
.07869 .23888 m
.08153 .2416 L
s
gsave
.29165 .07573 -70.4276 -17.1875 Mabsadd m
1 1 Mabs scale
currentpoint translate
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding ISOLatin1Encoding def
        currentdict
    end

    newfontname exch definefont pop
} def

0 21.1875 translate 1 -1 scale
63.000 13.313 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.313 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(s) show
69.125 13.313 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 233.563},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000jA000`40O003h00Oogoo8Goo003oOolQ
Ool00?moob5oo`00ogoo8Goo003oOolQOol00?moob5oo`00ogoo8Goo003oOolQOol00?moob5oo`00
ogoo8Goo003oOolQOol00?moob5oo`00ogoo8Goo003oOolQOol00?moob5oo`00`goo1@000goo0`00
0Woo0P00CWoo0035Ool00`00Oomoo`04Ool01000Oomoo`000Woo00<007ooOol0Bgoo0035Ool01000
Oomoogoo1@0000Eoo`00Oomoo`0004eoo`00aGoo00D007ooOomoo`00009oo`04001oogoo0002Ool0
0`00Oomoo`1;Ool00<Eoo`03001oogoo009oo`03001oo`00009oo`04001oogoo001=Ool00<Aoo`80
00Eoo`8000=oo`8004ioo`00ogoo8Goo003oOolQOol00?moob5oo`00ogoo8Goo003oOolQOol00=5o
o`8004eoo`00cgoo0P0000=oo`000000C7oo003<Ool30003Ool2001<Ool00<Yoo`8000Ioo`03001o
o`0004]oo`00agoo0`0027oo00@007ooOol004Yoo`00a7oo0`002goo00<007ooOol00P00BGoo0032
Ool2000>Ool01@00Oomoogoo0000BGoo002oOol3000@Ool00`00Oomoo`02Ool00`00Oomoo`16Ool0
095oo`D00003Ool00000008000=oo`8001eoo`8001Aoo`05001oogooOol00018Ool009=oo`05001o
ogooOol00005Ool01000Oomoo`006Goo0`005Woo00<007ooOol00Woo00<007ooOol0AGoo001^Ool3
000ROol00`00Oomoo`02Ool00`00Oomoo`02Ool01000Oomoo`005goo0`0067oo00<007ooOol00goo
00<007ooOol0A7oo001aOol00`00Oomoo`0OOol00`00Oomoo`03Ool01@00Oomoogoo00000Woo00<0
07ooOol04Woo0`006goo00<007ooOol00goo0P00AGoo001^Ool3000ROol01@00Oomoogoo00000Woo
00@007ooOol0009oo`03001oogoo00moo`<001ioo`03001oogoo00Aoo`03001oogoo04=oo`00KWoo
100087oo0P0017oo0P0017oo0P0047oo0P008Goo00<007ooOol01Goo00<007ooOol0@Woo002]Ool3
000SOol00`00Oomoo`05Ool00`00Oomoo`12Ool00:]oo`800003Ool007oo02=oo`03001oogoo00Io
o`03001oogoo045oo`00Z7oo0`00:7oo00<007ooOol01goo00<007ooOol0@7oo002VOol2000[Ool0
0`00Oomoo`07Ool20011Ool00:=oo`<002eoo`03001oogoo00Qoo`03001oogoo00Eoo`8000Aoo`80
00=oo`@002]oo`00X7oo0`00<7oo00<007ooOol027oo00<007ooOol017oo00@007ooOol000Qoo`03
001oogoo02aoo`00WWoo0P000goo00<007ooOol0;Goo00<007ooOol02Goo00<007ooOol00goo00@0
07ooOol000Uoo`03001oogoo02]oo`00Vgoo0`00=Goo00<007ooOol02Woo00<007ooOol00Woo00@0
07ooOol000Yoo`03001oogoo02Yoo`00VGoo0P00>Goo00<007ooOol02Goo00<007ooOol00Woo00@0
07ooOol000Qoo`04001oogoo000[Ool009Ioo`<003]oo`03001oogoo00Yoo`03001oogoo009oo`80
00Yoo`8002aoo`00U7oo0P0000=oo`00Ool0>goo00<007ooOol02Goo00<007oo0000?Goo002AOol3
0010Ool00`00Oomoo`0;Ool00`00Oomoo`0kOol006=oo`D0009oo`8000Aoo`8001aoo`<004=oo`03
001oogoo00aoo`03001oogoo03Yoo`00IGoo00D007ooOomoo`00009oo`04001oogoo0002Ool00`00
Oomoo`0GOol20016Ool00`00Oomoo`0=Ool00`00Oomoo`0iOol006Eoo`05001oogooOol00002Ool0
1000Oomoo`000Woo00<007ooOol057oo1000Agoo00<007ooOol03Goo00<007ooOol0>Goo001UOol0
1@00Oomoogoo00000Woo00@007ooOol0009oo`03001oogoo019oo`8004]oo`03001oogoo00eoo`80
03Yoo`00IGoo00D007ooOomoo`00009oo`04001oogoo0002Ool00`00Oomoo`0?Ool3001=Ool00`00
Oomoo`0?Ool00`00Oomoo`0gOol006Aoo`8000Aoo`8000Aoo`80011oo`80051oo`03001oogoo00mo
o`03001oogoo03Moo`00Ogoo0`00DWoo00<007ooOol047oo00<007ooOol0=Woo001mOol20002Ool3
001@Ool00`00Oomoo`0AOol00`00Oomoo`0eOol007Yoo`L0009>NP8004moo`03001oogoo015oo`80
03Ioo`00Mgoo1P0000=Fg000CWX014ij00<007ooOol0CGoo00<007ooOol04Goo0P00=Woo001eOol5
0005E]`00`00CWY>NP02CWX2001>Ool00`00Oomoo`0COol00`00Oomoo`0cOol0079oo`H00003E_`0
05KL00AFg003001>NTij009>NP8004moo`03001oogoo019oo`03001oogoo03=oo`00L7oo1@001EKl
00<005KLE]`00eKL00L004ijCWY>NP000380001>Ool00`00Oomoo`0COol00`00Oomoo`0bOol006eo
o`D0009Fo@03001Fo5Kl00AFo003001Fg5KL009Fg003001>NTij009>NP030000<P0004eoo`03001o
ogoo01Aoo`03001oogoo035oo`00Jgoo1@0015Km00<005KlE_`015Kl1`000dij00<0000b0000CGoo
00<007ooOol057oo00<007ooOol01Woo0P0017oo0P001Goo0`006goo001XOol50002E_d00`00E_eF
o@03E_d00`00E_aFo00300000e;L0012F`03@U/20002CWX010000380<P00C7oo00<007ooOol05Goo
00<007ooOol017oo00@007ooOol000]oo`03001oogoo01Yoo`00IGoo1P0015Km00<005KmE_d00eKm
0`001E;L00<0049K@U/00T9K10000P0b00<007ooOol0BWoo00<007ooOol05Woo00<007ooOol00goo
00@007ooOol000Qoo`D001]oo`00@7oo0P0017oo0P006goo0P0000=oo`0000000UKm00<005KmE_d0
0eKm00<005KmE_d00P000UKl00<005;LD]`00e;L00<0049K@U/00T9K00<000dG3AL00P0000<0<P00
Ool0Bgoo00<007ooOol05Woo00<007ooOol00goo00@007ooOol000Qoo`04001oogoo000LOol003mo
o`04001oogoo0002Ool01000Oomoo`005goo1P0015Km00<005KmE_d00eKm0`001EKl00<005;LD]`0
0e;L00D0049K@U]2F`0000<=5`040000<P00001;Ool00`00Oomoo`0GOol00`00Oomoo`02Ool01000
Oomoo`002Goo00<007oo000077oo000oOol01000Oomoo`000Woo00@007ooOol001Eoo`D0009Fo@03
001FoEKm00=Fo@03001FoEKm0080009Fo@03001Fo5Kl00AFo003001Bg5;L009Bg0030012Fd9K0092
F`04000=5`dG000203800`00Oomoo`19Ool00`00Oomoo`0HOol00`00Oomoo`02Ool2000;Ool2000L
Ool0041oo`8000=oo`04001oogoo000BOol60004E_d00`00E_eFo@03E_d30005E_d00`00E_aFo003
E_`80002@U/01`003AL=5`dG0000<P0004]oo`03001oogoo01Moo`8001aoo`80015oo`00?goo00@0
07ooOol0009oo`04001oogoo000@Ool200000goo00000002E_d00`00E_eFo@03E_d00`00E_eFo@02
0002E_d00`00E_eFo@04E_d00`00E_aFo0020002D]`01P00AYa6W4JL000UfP800007@U/000dG3AL0
000b0000Bgoo00<007ooOol06Goo00<007ooOol067oo00@007ooOol0011oo`00@7oo0P0017oo0P00
3Woo1P0015Km00<005KmE_d00eKm0`001EKm00<005KmE_d00eKm0`0015;L00<004JLAY`00TJL00<0
02GJ9MX00P000PdG0`00Bgoo00<007ooOol06Woo00<007ooOol05goo00<007ooOol04Goo001COol6
0002E_d00`00E_eFo@03E_d00`00E_eFo@020002E_d00`00E_eFo@04E_d00`00E_eFo@020002E_`0
0`00D]aBg003D]`01@00AYa6W4JL00000bGJ0P000PdG0`00BWoo00<007ooOol06Woo00<007ooOol0
5Woo1@0047oo001AOol2000017oo0000000015Km00<005KmE_d00eKm0`001EKm00<005KmE_d00eKm
0`0015Kl00<005;LD]`00e;L00<004JLAY`00TJL00@002GJ9MXUfP8000043AL00000001:Ool00`00
Oomoo`0KOol00`00Oomoo`0FOol00`00Oomoo`0AOol004ioo`<00003Ool00000009Fo@03001FoEKm
00=Fo@03001FoEKm0080009Fo@03001FoEKm00AFo@03001FoEKm0080009Fo@03001Fo5Kl00=Fo004
001Bg5;LD]`500000dJL000UfP029MX00`0009D00003001;Ool00`00Oomoo`0JOol00`00Oomoo`0Z
Ool004aoo`800004Ool000000004E_d00`00E_eFo@03E_d30005E_d00`00E_eFo@03E_d30005E_d0
1000E_aFo5Kl1@000dZl00<003Hl00000`000RGJ00<0002E00000P00Bgoo00<007ooOol06goo00<0
07ooOol0:Goo0019Ool60002E_d00`00E_eFo@03E_d00`00E_eFo@020002E_d00`00E_eFo@04E_d0
0`00E_eFo@020002E_d00`00E_eFo@03E_d40003D]`00`00B[a:_002B[`00`00=S`f?002=S`40002
09D3001:Ool00`00Oomoo`0LOol00`00Oomoo`0XOol004Moo`800003Ool00000008000=Fo@03001F
oEKm00=Fo@<000EFo@8000=Fo@@000EFo@03001FoEKm0080009Fo003001Bg5;L009Bg003001:_4Zl
00=:_005000f?3Hl=S`000025GX20000102E00000000BWoo00<007ooOol077oo00<007ooOol0:7oo
0014Ool300000goo00000002E_d00`00E_eFo@03E_d00`00E_eFo@020002E_d00`00E_eFo@04E_d4
0003E_d00`00E_eFo@03E_d30004E_`00`00D]aBg003D]`00`00B[a:_002B[`00`00=S`f?002=S`0
0`005GXENP0200000`2E0000001:Ool00`00Oomoo`0MOol00`00Oomoo`0WOol0045oo`<00004Ool0
00000004E_d00`00E_eFo@03E_d30005E_d20002E_d20002E_d00`00E_eFo@03E_d00`00E_eFo@02
0002E_d00`00E_aFo003E_`01000D]aBg5;L2P0000<f?0005GX00QEj1000BWoo00<007ooOol07Goo
0P00:7oo000oOol2000017oo000000000UKm00<005KmE_d00eKm00<005KmE_d00P000UKm00<005Km
E_d015Km0`0015Km00<005KmE_d00eKm0`0015Km00<005KlE_`00eKl10000dkM00<004:L@Y`00T:L
00<002XK00000P000QEj0`0000=Z8000Ool0Agoo0P0087oo00<007ooOol09Woo000lOol3000017oo
000000001EKm0P000eKm0`001EKm0P000UKm0P000UKm00<005KmE_d00eKm00<005KmE_d00P000UKm
00<005KmE_d00eKm00<005KlE_`00P000U;m00<004kMC]d00TkM00<004:L@Y`00T:L00@002XK:Q/Z
6`80008ENP@004Aoo`L0021oo`03001oogoo02Eoo`00>Woo0P000Woo0P000UKm00<005KmE_d015Km
10000UKm00<005KmE_d015Km0`0015Km00<005KmE_d00eKm0`0015Km00<005KmE_d00eKm0`0015;m
00<004kMC]d00dkM00<004:L@Y`00T:L00@002XK:Q/Z6`@0000407D006XP0010Ool3000015<M0000
00000U<M00<007ooOol07goo00<007ooOol097oo000hOol20002Ool20005E_d20002E_d20002E_d0
0`00E_eFo@02E_d20002E_d20002E_d00`00E_eFo@03E_d00`00E_eFo@020002E_d00`00E_eFo@03
E_d00`00E_eFo@020002E_d00`00D_eBo@03D_d01000C]e>gDkM1@0000=2W000:Q/00RXK00<000UI
2ET00`0000=Z8000Ool0?7oo0`0000AC7@00000000EC7@80021oo`03001oogoo02Aoo`00>7oo1000
00=Fo@00E_d01EKm0`0015Km00<005KmE_d00eKm0`0015Km00<005KmE_d00eKm0`001EKm00@005Km
E_eFo@@000EFo@04001BoE;mD_d50003A[d00`00>U`000030002:Q/00`002ET9F@0200000fXP0000
000iOol300001U<M00000000Dad000MC7@8001moo`8000Moo`8000Aoo`8000=oo`<0011oo`00>7oo
0P0015Km0P000UKm0P000UKm00<005KmE_d00eKm00<005KmE_d00P000UKm00<005KmE_d00eKm00<0
05KmE_d00P000UKm00<005KmE_d00eKm10000eKm00<005KmE_d00eKm10000dkM00<004JmA[d00TJm
00<003YL>U`00SYL10000`UI00@006XPJR0003Eoo`@0009C7@8000EC7@03001C7E<M00@00003Dad0
0000021oo`03001oogoo00=oo`04001oogoo0008Ool01000Oomoo`003goo000`Ool50003Ool20005
E_d30004E_d00`00E_eFo@03E_d30005E_d01000E_eFoEKm10001EKm00<005KmE_d00P000UKm00<0
05KmE_d00eKm00<005KmE_d00P000U;m00<004kMC]d00TkM00<004JmA[d00dJm00D003YL>U`jG000
008Qf`800089F@03001Z86XP0080039oo`8000=C7@@000IC7@<000IC7@03001oogoo01ioo`03001o
ogoo00=oo`04001oogoo0008Ool01000Oomoo`003goo000bOol00`00Oomoo`02Ool2000017oo001F
oEKm0P000UKm00<005KmE_d00eKm00<005KmE_d00P000UKm00<005KmE_d015Km0`000eKm00<005Km
E_d015Km0P001EKm00<005KmE_d00eKm0P001E;m00<004kMC]d00TkM00<004JmA[d00TJm00<003YL
>U`00SYL00@0027K8M/Qf`<000=Z8003001oogoo02ioo`800003Dad00000008000AC7@03001C7E<M
00<0009C7@03001C7E<M00AC7@80021oo`03001oogoo009oo`04001oogoo0008Ool3000@Ool0039o
o`03001oogoo009oo`04001oogooOol20005E_d00`00E_eFo@03E_d20005E_d20002E_d200000eKm
001Fo@04E_d20002E_d200000eKm001Fo@05E_d00`00E_eFo@0200000eKm001Bo@05D_d01000C]e>
gDkM2@0000<jG0008M/00R7K00@000Di0000009Z808002eoo`800004Dad000000007Dad30005Dad0
0`00DaeC7@05Dad2000POol01@00Oomoogoo00000Woo00<007ooOol01goo00<007ooOol03goo000Z
Ool40004Ool00`00Oomoo`02Ool20003Ool00`00E_eFo@04E_d00`00E_eFo@0200000eKm001Fo@05
E_d30004E_d00`00E_eFo@03E_d30004E_d00`00E_eFo@03E_d30004E_d20003D_d50002C]d00`00
@Ye2W@02@Yd00`00<S`0000300028M/00`001CT00003JR000`00Oomoo`0XOol60003Dad00`00DaeC
7@040002Dad00`00DaeC7@04Dad00`00DaeC7@0400000e<M0000000NOol20004Ool2000:Ool3000?
Ool0039oo`03001oogoo009oo`03001oogoo00=oo`8000=Fo@@000AFo@80009Fo@80009Fo@03001F
oEKm00=Fo@03001FoEKm0080009Fo@03001FoEKm00=Fo@03001FoEKm0080009Fo@03001FoEKm00=F
o@@000=Bo@05001>gDkMC]d00004@Yd00`00<S`b?002<S`4000010Di001Z86XP0`009Woo1P001e<M
0P001U<M00<005<MDad015<M0`001E<M0P0087oo00<007ooOol07goo000aOol20004Ool00`00Oomo
o`05Ool30003E_d00`00E_eFo@03E_d30004E_d00`00E_eFo@03E_d30005E_d01000E_eFoEKm1000
1EKm00<005KmE_d00P000U;m00<005;mD_d00U;m00<004kMC]d00TkM00<004:M@Yd00T:M00D0038l
<S`b?000008I^`D00004JHD00000000ROol60002Dad00`00DaeC7@03Dad200000e<M001C7@06Dad0
0`00DaeC7@030002Dad00`00DaeC7@04Dad2000POol00`00Oomoo`0NOol003Ioo`8000Qoo`03001F
oEKm00=Fo@03001FoEKm0080009Fo@03001FoEKm00=Fo@03001FoEKm0080009Fo@03001FoEKm00=F
o@@000=Fo@03001FoEKm00=Fo@<000ABo@03001BoE;m00=Bo@04001>gDkMC]d900000c8l000I^`02
6K/00`00JHEYQ@02JHD2000NOol70006Dad00`00DaeC7@020003Dad00`00DaeC7@05Dad30005Dad0
0`00DaeC7@05Dad2000OOol00`00Oomoo`0NOol003Ioo`03001oogoo00Moo`03001FoEKm00=Fo@<0
00EFo@04001FoEKmE_d40005E_d00`00E_eFo@020002E_d00`00E_eFo@03E_d00`00E_eFo@020002
E_d00`00D_eBo@03D_d01000D_eBoE;m1@000TZm00<003im?Wd00Sim00<002hL00000P000QVk0P00
0fV50`006Woo0`000U<M0P000U<M00<005<MDad015<M0`001U<M00<005<MDad010000U<M00<005<M
Dad015<M00<005<MDad0100000=C7@00Ool07goo00<007ooOol07Goo000fOol20009Ool00`00E_eF
o@020002E_d00`00E_eFo@04E_d30003E_d00`00E_eFo@03E_d30004E_d00`00E_eFo@03E_d30005
E_d01000D_eBoE;m1@000dkM00<004ZmB[d00TZm00D003im?WdnO@0000@^708000036K/006V500=Y
Q@<001Ioo`<0009C7@<000EC7@03001C7E<M00<0009C7@03001C7E<M00AC7@<000IC7@03001C7E<M
00AC7@<000EC7@8001moo`8001ioo`00=Woo00<007ooOol02Goo0P001EKm0P000UKm0P0000=Fo@00
E_d015Km00<005KmE_d00P000UKm00<005KmE_d00eKm00<005KmE_d00P000UKm00<005KmE_d00eKm
10000e;m00<004kMC]d00dkM00<004ZmB[d00TZm00D003im?WdnO@0000<^7003000E^`00008000=Y
Q@04001Q8000000BOol30002Dad300000e<M001C7@05Dad30006Dad00`00DaeC7@030002Dad00`00
DaeC7@05Dad00`00DaeC7@030002Dad00`00DaeC7@04Dad00`00Oomoo`0NOol00`00Oomoo`0LOol0
03Ioo`03001oogoo00Yoo`03001FoEKm00AFo@<000AFo@03001FoEKm00=Fo@8000EFo@04001FoEKm
E_d40005E_d00`00E_eFo@020002E_d00`00D_eBo@03D_d01000C]e>gDkM2@0000<nO@00;Q`00bhL
00@001Fk5K/000AYQ@<000moo`<0009C7@<000EC7@03001C7E<M00<0009C7@03001C7E<M00EC7@80
00IC7@03001C7E<M00AC7@<000EC7@03001C7E<M00AC7@80021oo`03001oogoo01]oo`00=Woo0P00
37oo0P000UKm0P000UKm00<005KmE_d00eKm00<005KmE_d00P0000=Fo@00E_d01EKm0`000eKm00<0
05KmE_d00eKm0`0015Km00<005;mD_d00e;m10000dkM00<004JmA[d00TJm1P0000<^70005K/00QFk
00H006V5JHEYQ@00HB02000;Ool30002Dad300000e<M001C7@05Dad30005Dad00`00DaeC7@03Dad2
00000e<M001C7@06Dad00`00DaeC7@030002Dad00`00DaeC7@04Dad00`00DaeC7@030002Dad00`00
Oomoo`0NOol00`00Oomoo`0JOol003Ioo`03001oogoo00eoo`8000EFo@04001FoEKmE_d40004E_d2
0002E_d200000eKm001Fo@04E_d00`00E_eFo@020002E_d00`00E_eFo@03E_d00`00D_eBo@020002
D_d00`00C]e>g@02C]d00`00A[e6_@02A[d01P00>WdjOCYm000V70<0008E^`03001YQFV5009YQ@04
001Q80000007Ool30002Dch30005Dad00`00DaeC7@030002Dah00`00DaeC7@04Dad00`00DaeC7@02
0003Dad00`00DaeC7@05Dad30005Dad00`00DaeC7@04Dad30005Dad00`00Oomoo`0NOol2000KOol0
01eoo`8000Aoo`8000Eoo`03001oogoo009oo`8000Aoo`80011oo`03001FoEKm00=Fo@@000=Fo@03
001FoEKm00=Fo@<000AFo@03001FoEKm00=Fo@8000EFo@04001FoEKmE_d40004D_d00`00C]e>g@03
C]d00`00A[e6_@02A[d01000>WdjO@0012HL0P0000<E^`00JHD00VV500<0064PHB000P000goo0`00
0U<n0`0000=C?P00Dad01E<M0`001E<N00<005<MDad015<M0`001U<M00<005<MDad010000U<M00<0
05<MDad015<M00<005<MDad00`000U<M00<005<MDad015<M00<007ooOol07Goo0P006goo000LOol0
1000Oomoo`002Woo00D007ooOomoo`00009oo`05001oogooOol0000BOol00`00E_eFo@020002E_d0
0`00E_eFo@03E_d00`00E_eFo@020002E_d00`00E_eFo@03E_d00`00E_eFo@0200000eKm001Fo@05
E_d30003D_d00`00D_eBo@03D_d01000C]e>gDkM2@0000<jO@009Q`00RHL00@001Fk000000AYQ@<0
0004C80000000002Bil30005Dch00`00DaeC7@030002Dch00`00DaiC7P04Dah00`00DaeC7@030002
Dah00`00DaeC7@04Dad30006Dad00`00DaeC7@04Dad30005Dad00`00DaeC7@04Dad2000POol00`00
Oomoo`0HOol001aoo`04001oogoo000:Ool00`00Oomoo`04Ool01@00Oomoogoo00004goo0P0015Km
00<005KmE_d00eKm0`001EKm00@005KmE_eFo@@000AFo@80009Fo@800003E_d005;m00ABo@04001B
oE;mD_d50003B[d01P00@Ye2WD:M000fG@@0008V7005000E^aFk001YQ@060002Bil300000d^O001C
?P05Dch30005Dch00`00DaiC7P04Dah30006Dah00`00DaeC7@030002Dah00`00DaeC7@05Dad00`00
DaeC7@030002Dad00`00DaeC7@04Dad00`00DaeC7@0300000e<M0000000OOol00`00Oomoo`0HOol0
01Ioo`@0009oo`04001oogoo000;Ool01000Oomoogoo0P0017oo0P004goo00<005KmE_d00eKm00<0
05KmE_d00P000UKm00<005KmE_d00eKm10000eKm00<005KmE_d00eKm0`000eKm00<005;mD_d00e;m
10000dkM00<004ZmB[d00TZm00<004:M@Yd00T:M00@003IM=UdfG@@0008E^`X000E;W`03001C?U<n
00<0009;G`03001C?U<n00AC?P03001C7U<N00<0009C7P03001C7U<N00AC7P<000EC7P03001C7E<M
00EC7@<000EC7@03001C7E<M00AC7@<000EC7@03001oogoo01ioo`03001oogoo01Moo`0077oo00@0
07ooOol000Qoo`04001oogoo0003Ool00`00Oomoo`03Ool00`00Oomoo`0COol20003E_d30005E_d0
0`00E_eFo@020002E_d00`00E_eFo@03E_d00`00E_eFo@020002E_d00`00E_eFo@02E_d00`00D_eB
o@020002D_d00`00C]e>g@02C]d00`00B[e:_@03B[d01@00@Ye2WD:M00000cIM00<0027l8O`00`00
00<E^`000CT03`001T]O00<005<nDch00U<n10001U<N00<005<NDah00`000U<N00<005<NDah015<N
00<005<MDad0100000=C7@00Dad01U<M00<005<MDad00`000U<M00<005<MDad015<M00<007ooOol0
7Goo0P001goo0P0017oo0P0017oo0P000goo000MOol20009Ool40003Ool30003Ool00`00Oomoo`0E
Ool30002E_d00`00E_eFo@04E_d20005E_d01000E_eFoEKm10001EKm00@005KmE_eFo@<000ABo@03
001>gDkM00=>g@04001:_DZmB[d900000cIM000Qo0028O`700001@3g0000U@0007@01@0000<VoP00
000014]O1P000dlo00<005<NDah015<N0`001E<N00<005<NDah015<N0`001U<M00<005<MDad015<M
0`001E<M00<005<MDad015<M0P007goo00<007ooOol017oo00@007ooOol000Qoo`04001oogoo0002
Ool003Aoo`8001Qoo`03001FoEKm009Fo@80009Fo@800003E_d005Km00AFo@@000=Fo@03001FoEKm
00=Fo@@0009Bo@03001BoE;m00=Bo@04001>gDkMC]d50003A[d01P00?YdnWCjM000b?0@000038O`0
000000<00081>@060000m`0009D0001d2`000clO00<004loCcl014lo00<005<NDah00`000TlN00<0
05<NDah015<N00<005<NDah00`000U<N00<005<MDad015<M00<005<MDad010000U<M00<005<MDad0
15<M00<005<MDad00`0000=C7@0000007goo00<007ooOol00goo00@007ooOol000Qoo`04001oogoo
0002Ool003Aoo`03001oogoo01Qoo`04001FoEKmE_d40004E_d00`00E_eFo@020002E_d00`00E_eF
o@03E_d00`00E_eFo@020002E_d00`00D_eBo@02D_d01000D_eBoE;m1@000dkM00<004JmA[d00TJm
00D003jM?YdnW@0000<b?0D000<AV`030001>@0000D0000407@000000AL200024Wd00`00?alo7`04
?al00`00Ccm??`02Ccl40005Cah00`00DaiC7P04Dah30005Dah00`00DaeC7@04Dad30006Dad00`00
DaeC7@04Dad30005Dad2000OOol00`00Oomoo`02Ool01000Oomoo`002Goo0P000goo000dOol2000J
Ool30003E_d00`00E_eFo@03E_d20004E_d00`00E_eFo@03E_d30005E_d01000D_eBoE;m10000e;m
00<004kMC]d00dkM00<004JmA[d00TJm00@003jM?YdnW@<00004<S`0027l8O`200024I/400000`3h
0000e`0;0004?al60003Bal00`00Cai?7P04Cah00`00DaiC7P030002Dah00`00DaiC7P04Dah00`00
DaeC7@030002Dah00`00DaeC7@05Dad00`00DaeC7@030002Dad00`00DaeC7@04Dad2000NOol20003
Ool01000Oomoo`0027oo00@007ooOol0009oo`00=7oo00<007ooOol06Woo00<005KmE_d00eKm00<0
05KmE_d00P0000=Fo@00E_d015Km00<005KmE_d00P000UKm00<005KmE_d00eKm10000U;m00<005;m
D_d00e;m00@004kMC]e>g@h00003<S`0027l008Qo08000044I/0005I00020?P01P000=L0002f0000
eP@000034Ud0000000<000<fo`03001;7d/O00A;7`03001?7TlN009?7P@000EC7P03001C7U<N00AC
7P<000EC7P03001C7E<M00EC7@<000EC7@03001C7E<M00H001eoo`<000Aoo`8000Yoo`8000=oo`00
=7oo00<007ooOol06goo00@005KmE_eFo@@000AFo@8000=Fo@<000EFo@03001FoEKm0080009Fo@03
001BoE;m009Bo@04001BoE;mD_d50003B]d400000dJm000jO@02>Wd00`00;S`0000300000b7l000A
V`0200020ET00`000?P0n00500000`3F0001F@060005=_l00`00Bam;7`02Bal60003Cah00`00DaiC
7P04Dah00`00DaiC7P030002Dah00`00DaiC7P04Dah00`00DaeC7@040002Dad00`00DaeC7@03Dad3
0005Dad2000NOol00`00Oomoo`0COol003Aoo`8001eoo`<000=Fo@03001FoEKm00=Fo@@0009Fo@03
001FoEKm00=Fo@<000AFo@03001BoE;m00=Bo@<000=>g@04001:gD[MB]d20004A[d01@00>WdjOCYm
000012hl10000Q6K00<0005I0ET00`0000<0n00000001@0000<1F@005Sd030000dLO00<004lNCah0
14lN00<005<NDah00U<N10001U<N00<005<NDah00U<N10001U<M00<005<MDad00`000U<M00<005<M
Dad015<M00<007ooOol07Goo00<007ooOol04Woo000QOol20004Ool20004Ool20004Ool00`00Oomo
o`0NOol00`00E_eFo@03E_d00`00E_eFo@020002E_d00`00E_eFo@02E_d00`00E_eFo@020002E_d0
0`00E_eFo@03E_d00`00D_eBo@0200000e;m001>g@04C]d30002B]d00`00A[e6_@02A[d01@00>Wdj
OCYm00000bhl00<001gl7O`00P000Q6K0`0000@16@000000n0T000<F?@03000bgc;O00<bg`030017
7dLO00=77`03001?7TlN00A?7PD000=?7P03001C7U<N00AC7PD000=C7P03001C7E<M00AC7@<000EC
7@03001C7E<M00AC7@8001moo`03001oogoo015oo`0087oo00@007ooOol000Qoo`04001oogoo0003
Ool2000POol20003E_d30005E_d01000E_eFoEKm0`001EKm00@005KmE_eFo@@000ABo@03001>g@00
01@00003;S`001gl008Mo0<000031EX0000000816@090000n03h0000m`000AP0005j00@000035Sd0
000000@bg`0300177dLO00=77`03001?7TlN00<0009;7P03001?7TlN00A?7P03001C7U<N00<0009C
7P03001C7U<N00AC7P03001C7E<M00<0009C7P03001C7E<M00AC7@03001C7E<M00<00003Dad00000
01ioo`03001oogoo015oo`0087oo00@007ooOol000]oo`05001oogooOol0000SOol30002E_d00`00
E_eFo@03E_d40002E_d00`00E_eFo@04E_d30003D_d01000D_eBoE;m1@000dkM00<004ZmB[d00TZm
00<004:M@Yd00T:M00H003Im=WdfO@00:S`400000agl000AV`0200021EX00`000AT16@0200001@3h
0000m`000AP00P0000<1NP006Sd040001D/N00<004lNCah014lN0`001E<N00<005<NDah015<N0`00
1E<N00<005<MDad015<M0`001E<M00<007ooOol07Goo00<007ooOol047oo000JOol40002Ool01000
Oomoo`002Goo0P0017oo00<007ooOol08Woo00<005KmE_d00UKm00<005KmE_d00P000UKm00<005Km
E_d00UKm0P000UKm0P0000=Fo@00D_d015;m1@0000=Bo@00C]d00dkM00<004ZmB[d00dZm00D004:M
@Ye2W@0000<fO@03000Z?2Xl008Z?0@0008AV`030005FPEJ00<000030AP0000000@000040K/00000
00046Sd00`00<[lb_`02<[l00`00@_m2o`03@_l00`00Bai;7P04Bah00`00Cai?7P030002Cah00`00
DaiC7P04Dah00`00DaiC7P030002Dah00`00DaiC7P04Dah00`00DaeC7@030002Dad00`00DaeC7@04
Dad00`00Oomoo`0MOol00`00Oomoo`0?Ool0021oo`04001oogoo0009Ool00`00Oomoo`03Ool2000T
Ool01000E_eFoEKm0`001EKm00@005KmE_eFo@@000AFo@03001Bo@0000D000=Bo@03001>gDkM00=>
g@04001:_DZmB[d900000cIm000Z?002:S`00`007O`Mo00200024I/30000104i0000004H0P0000@1
60000ET00081^`@0008J?@05000b_c:o<[l00006@_l00`00Bai;7P02Bah40005Cah00`00DaiC7P04
Dah30005Dah00`00DaiC7P04Dah30006Dad00`00DaeC7@02Dad3000NOol00`00Oomoo`0?Ool0025o
o`8000Yoo`<000=oo`03001oogoo02Aoo`<0009Fo@03001FoEKm00=Fo@@000=Fo@04001FoEKmE_d5
0003D_d00`00D_eBo@02D_d01000C]e>gDkM1@000dJm00H003jM?YdnW@00=Ud500000bXl000Mo002
7O`300000`Ej000000020CT01`000AP160000AP0005I00800081^`03000NGP0000D00003<[l004;o
00=2o`H000=;7P03001?7TlN00A?7P03001C7U<N00<0009?7P03001C7U<N00AC7P03001C7U<N00<0
009C7P03001C7E<M00AC7@D0009C7@8001eoo`80011oo`0017oo10000Woo0`000Woo0`0017oo0P00
0goo1@004goo00<007ooOol09Goo00<005KmE_d00UKm00<005KmE_d00P000UKm00<005KmE_d00eKm
1@0000=Fo@00D_d015;m00@005;mD_eBo@@000=>g@030016_DJm0096_@05000nWCjM?Yd00004=Ud0
0`00:Q`0000200027O`010004K/000000PEj00D0004i0CT0004H008000030AP0000000D000@NGP03
000b_`0000D000=2o`03001;7T/N00A;7P03001?7TlN009?7P@000I?7P03001C7U<N009C7P@000IC
7P03001C7E<M00<0009C7P03001C7E<M00=C7@8001ioo`03001oogoo00eoo`000goo00@007ooOol0
009oo`03001oogoo00Aoo`03001oogoo009oo`04001oogoo0002Ool4000COol2000XOol01000E_eF
oEKm0`001EKm00<005KmE_d01@0015Km00@005;mD_eBo@D0009Bo@03001>gDkM00=>g@030016_DJm
0096_@05000nWCjM?Yd00003=Ud00`00:Q`Z7002:Q`200001Agl000A^a6k00000PEj200000<1NP00
00000PWL10000QiN00D003:o<[lb_`0000I2o`03001;7T/N009;7PH000=?7P03001?7TlN00A?7PD0
00=C7P03001C7U<N00AC7P<000EC7P03001C7E<M00AC7@8001eoo`03001oogoo00eoo`000goo00@0
07ooOol0009oo`@000=oo`03001oogoo009oo`03001oogoo00Ioo`03001oogoo015oo`03001oogoo
02Qoo`<0009Fo@03001FoEKm00=Fo@@000=Fo@03001FoEKm00=Fo@<000=Bo@03001BoE;m009Bo@04
001>gDkMC]d>00000cIM000Z7002:Q`010007M`000000a6k0`0000D1FP000001>@000081F@030001
NP5j00800089g003000NGP0000D00003<[l004;o00=2o`H000=:oP03001?7TlN00A?7P03001?7TlN
00<0009?7P03001C7U<N00AC7P03001C7U<N00<00003Dah005<N00IC7P03001C7E<M00D001ioo`03
001oogoo00aoo`0017oo0`000goo0P0017oo00@007ooOomoo`D000=oo`<001=oo`03001oogoo02Uo
o`03001FoEKm009Fo@03001Fo@000080009Fo@03001FoEKm00=Fo@03001FoEKm00800003D_d005;m
00ABo@04001BoE;mD_d40003B]d00`00A[e6_@02A[d01P00?YdnWCjM000bG@D00004:Q`001gL7M`2
00000a6k0009NP0200020EX00`000CT000020ET200000`5j0000000300047Uh00`00<[l000050003
?_l00`00B_i:oP04B_h00`00Cai?7P02Cah40005Cah00`00DaiC7P04Dah30005Dah00`00DaiC7P04
Dah30004Dad2000NOol00`00Oomoo`04Ool50002Ool000Ioo`03001oogoo00Moo`03001oogoo009o
o`03001oogoo01Yoo`8002]oo`03001FoEKm00@000EFo@04001FoEKmE_d40004D_d01000D_eBoE;m
1@000TkM00<004[MB]d00d[M00D004JmA[e6_@0000<nW@03000bGC9M008bG@03000V70000080008M
g0@00089NP050001FP5J0001>@0200000`5I0000000200023O`400027Uh01@00<[lb_c:o00001Sko
00<004[nB_h00T[n1P000dlN00<004lNCah014lN00<005<NDah00`0000=C7P00Dah01E<N00<005<N
Dah010000U<N00<005<MDad00e<M00<007ooOol06goo0P0027oo00<007ooOol00Woo0005Ool20008
Ool2000QOol00`00Oomoo`0ZOol40002E_d00`00E_eFo@04E_d30003E_d00`00D_eBo@03D_d30003
D_d00`00C]e>g@02C]d01000B]e:gD[M2@0000<nW@00<Ud00S9M00<002HL9Q`00RHL0`000Q6k0P00
0PUj1@0000D1NP00000000FK00<0008=o003000RGP0000D00003<[l003ko00<no`H000=6oP03001?
7TlN00A?7P03001?7TlN009?7P@000EC7P03001C7U<N00=C7P<000IC7P03001C7E<M009C7@<001io
o`03001oogoo00Eoo`03001oogoo009oo`00<Woo0P00:7oo0`000Woo00<005KmE_d00UKm0P000UKm
0P0000=Fo@00E_d015Km00<005;mD_d00P0000=Bo@00D_d015;m00@004kMC]e>g@@000=:g@060012
_D:m@[d003Ym1@0000<bG@009Q`00RHL00@001WL000000<A^`<000050GX000000EX000020GX20002
1I/500048Uh00`00<[h000050003?]l00`00A_i6oP03A_h00`00Cai?7P04Cah50003Cah00`00DaiC
7P03Dah00`00DaiC7P030002Dah00`00DaiC7P04Dah50002Dad2000NOol00`00Oomoo`04Ool00`00
Oomoo`02Ool0035oo`8002Ioo`<000Ioo`04001FoEKmE_d40004E_d01000E_eFoEKm100015;m00@0
05;mD_eBo@D0009>g@03001:gD[M009:g@050012_D:m@[d00004>Wd00`00<Ud00004000012HL000I
g1WL0P0000<A^`002IX00P000P5j00<0005J00000`5j0`000aHM10000R9N00D003:n<[hb_P0000Hn
g`030016oTKn00=6oP03001?7TlN00<0009;7P03001?7TlN00A?7P04001C7U<NDah40006Dah00`00
DaiC7P030002Dah00`00DaeC7@03Dad2000MOol00`00Oomoo`04Ool00`00Oomoo`02Ool001Uoo`80
00Aoo`8000=oo`@000=oo`8000Aoo`03001oogoo029oo`<000Yoo`<000=Fo@03001FoEKm00=Fo@<0
00=Fo@04001BoE;mD_d50003D_d00`00C]e>g@02C]d00`00B]e:g@02B]d01@00@[e2_D:m000013Ym
00H0039M<UdbG@009Q`300026M`400022IX00`000GX1NP0400000`5j0009^`0200045Qd00`009Wh0
000500000c:n000ng`03?]l:0005Bah00`00Cai?7P04Cah40003Cah00`00DaiC7P04Dah30005Dah0
0`00DaeC7@04Dad2000KOol30004Ool20004Ool001Qoo`04001oogoo0008Ool00`00Oomoo`03Ool0
1000Oomoo`000goo0P0087oo0`003Woo00<005KmE_d00eKm00<005KmE_d00P0000=Fo@00E_d015Km
0`000e;m00<005;mD_d00U;m00@004kMC]e>gA<00004<Ud002HL9Q`400024K/200022IX300000`Ej
0000000300022K/300025Qd01@009WhVORIn00000S:n1P000ckO00<004KnA_h00dKn00<004/NBah0
14/N00<004lNCah00`000TlN00<004lNCah00dlN00<005<NDah00`000U<N00<005<NDah015<N00<0
05<MDad0100000=C7@00Ool05Woo100000=oo`00Ool02Goo000HOol01000Oomoo`002Goo00<007oo
Ool01Goo00D007ooOomoo`0001moo`80019oo`04001FoEKmE_d40004E_d00`00E_eFo@0200000eKm
001Bo@04D_d01000D_eBoE;m10000dkM00<004JmA[d00TJm00D004:M@Ye2W@0000@jO@03000^?@00
0080008V7004000Ig00000024K/400000`Ej000000021GX00`001I/0000200000`Vk000J7@050002
9Wh00`00<[hb_P02<[h00`00?]lng`03?]l00`00A_i6oP04A_h00`00Bai;7P02Bah40005Cah00`00
Cai?7P02Cah40005Dah00`00DaiC7P04Dah30004Dad2000DOol30005Ool00`00Oomoo`08Ool0019o
o`@0009oo`04001oogoo000:Ool00`00Oomoo`02Ool20004Ool00`00Oomoo`0JOol3000EOol30003
E_d01000E_eFoEKm100015Km00@005;mD_eBo@D0009Bo@03001>gDkM00=>g@050016_DJmA[d00004
@Yd01@00>WdjOCYm00000bhm10000aWL0P0000<A^`002I/00P000PEj00<000Ej1GX00P0000<5V`00
00000P0011XM10001C:n00<003kO?]l00ckO00<004KnA_h00TKn1P000d[n00<004lNCah00dlN1@00
0dlN00<005<NDah015<N00<005<NDah00`000U<N00<005<MDad00e<M0P0047oo0`0027oo00<007oo
Ool027oo000HOol01000Oomoo`0027oo00@007ooOol000=oo`03001oogoo00=oo`8001Qoo`<001Uo
o`03001FoEKm00=Fo@<000=Fo@03001FoEKm00=Fo@<000=Bo@03001BoE;m009Bo@04001>gDkMC]d>
00000cYm000^?@02;Sd00`008Q`0000200000aWL000A^`0200032I/00`001GX5NP0300001@Vk0000
00003M`00`000QXM00<002In9Wh04@000dKn00<004[nB_h00d[n00<004lNCah00`000TlN00<004lN
Cah014lN00<005<NDah00U<N10001E<N00<005<MDad00`0000=C7P0000002goo100037oo00<007oo
Ool01goo000IOol2000:Ool20004Ool30002Ool00`00Oomoo`0EOol3000MOol00`00E_eFo@020000
0eKm001Fo@04E_d00`00E_eFo@0200000e;m001Bo@04D_d01000D_eBoE;m10000d[M00D004JmA[e6
_@0000@nW@03000fO@0000<00003;Sd0028L008R70@0008A^`030009V`VK00@000032I/000000089
^`<000033M`0000000<000<VOP03000f_SJn00<f_P03000ngSkN00<ngP030016oTKn00=6oP03001:
oT[n009:oP@000E?7P03001?7TlN00A?7PD000=C7P03001C7U<N00=C7P<000AC7P8000Qoo`<0011o
o`03001oogoo00Moo`00<7oo00<007ooOol04Woo0`008Goo0P0015Km00@005KmE_eFo@@000ABo@04
001BoE;mD_d50002C]d00`00B]e:g@03B]d01@00A[e6_DJm000013jM00@003Im=WdfO@D000048Q`0
01WL6M`200024K/200000`VK0009V`0200022I/200022K/400027Sd400000bIn000f_P04=[h00`00
?]hngP03?]h01000A_i6oTKn1P000d[n00<004lNCah014lN00<004lNCah00`000TlN00<005<NDah0
0e<N00<005<NDah00`000U<N00<005<NDah00e<N0P0017oo0`004goo00<007ooOol01goo000`Ool2
000@Ool3000UOol00`00E_eFo@03E_d30003E_d20003D_d40003D_d00`00C]e>g@02C]d01000B]e:
gD[M3@0000<fO@00;Sd00Rhm10000aWL0`0000<=^`0000000`VK0P0000<9V`0000000P000QGl0`00
0Qhm00<002Yn000040000dKn00<004[nB_h014[n00<004lNCah00TlN10001DlN00<005<NDah00U<N
10001U<N00<005<NDah00U<N1P005Woo00<007ooOol01goo000`Ool00`00Oomoo`0<Ool3000YOol0
0`00E_eFo@0200000eKm001Fo@04E_d40002D_d00`00D_eBo@02D_d01000C]e>gDkM1@000d[M00D0
04JmA[e6_@0000<nW@03000fO@0000<00004;Sd0028L8Q`300000aWL000A^`0200023K/200022I/4
000010gL000000000QGl1@000bYn00<003Jn=[h00cJn00<003kN?]h00ckN00<004KnA_h00dKn00<0
04[nB_h00T[n1P000dlN00<004lNCah00dlN1@000e<N00<005<NDah015<N1@000U<N0P0067oo00<0
07ooOol01goo000`Ool2000:Ool3000]Ool20004E_d00`00E_eFo@020002E_d00`00D_eBo@02D_d0
1000D_eBoE;m10000dkM00<004[MB]d00T[M00<004JmA[d00TJm00D003jM?YdnW@0000<fO@@000<R
70@0008A^`8000033K/000000080008=^`80008=g0D0008R?@@00003:Wh003Jn00@f_P03000ngSkN
008ngP030016oTKn0096oPH000=:oP03001?7TlN00=?7P03001?7TlN00<0009?7P03001C7U<N00AC
7P03001C7U<N00<00003Dah005<N00AC7P8001Qoo`03001oogoo00Moo`00<7oo00<007ooOol01goo
0P00<Goo0P000eKm0`001EKm00@005;mD_eBo@@0009Bo@03001>gDkM00=>g@04001:gD[MB]d80000
0cjM000fO@02=Wd00`00:SdZ?@0400000b8L000Ig00300000a6k0000000200000`fk000000033K/3
00036Qd300028Sd00`00;Wh0000@0003A_h00`00B_i:oP04B_h01000Cai?7TlN10001DlN00<005<N
Dah015<N0`001E<N00<005<NDah00e<N0P005goo00<007ooOol01goo000_Ool00`00Oomoo`05Ool3
000eOol30002E_d00`00E_eFo@03E_d40002D_d00`00D_eBo@02D_d01000C]e>gDkM1@000T[M00H0
04:m@[e2_@00>Yd500000cIm000Z?@02:Sd00`008Q`0000300000aWL0000000200000`fk00000002
3K/200000`fk000Ag00200046Qd40004;Wh00`00=[hf_P03=[h00`00?]hngP02?]h00`00A_i6oP04
A_h00`00B_i:oP02B_h50003Cah00`00Cai?7P04Cah00`00DaiC7P030002Cah00`00DaiC7P03Dah0
0`00DaiC7P0300000e<N0000000FOol00`00Oomoo`07Ool002moo`03001oo`00009oo`<003Uoo`03
001FoEKm009Fo@03001FoEKm0080009Bo@03001BoE;m009Bo@04001BoE;mD_d40003C]d01@00B]e:
gD[M000014:m00H003ZM>YdjW@00<Ud500001RXm000R728L000Io0<000035M/00000008=^`<00004
3K/0000000024M`200026Qd00`008UdRG@040002;Wh00`00=[hf_P02=[h00`00?]hngP03?]h00`00
A_i6oP02A_h60003B_h00`00Cai?7P03Cah00`00Cai?7P02Cah40005Cah00`00DaiC7P02Dah40004
Dah2000GOol00`00Oomoo`06Ool002moo`D003eoo`04001FoEKmE_d30004D_d01000D_eBoE;m1@00
0U;m00<004kMC]d00TkM00<004[MB]d00T[M00D004:m@[e2_@0000@jW@06000bGC9M<Ud002Xm0`00
0R8L00<001Wl6O`00P0000<Ef`0000000P0000<=^`0000000Pfk0`0000<Ag00000000`000b9M00<0
02in00004@000dKn00<004[nB_h00d[n00<004lNCah00TlN1P000dlN00<004lNCah00dlN1@000e<N
00<005<NDah00e<N0P005Woo00<007ooOol01Woo000XOol20005Ool20011Ool30002E_d20003D_d4
0003D_d00`00D_eBo@02D_d01000C]e>gDkM4P0000@bG@00:SdZ?@@000<Io08000035M/0016k0080
008=^`80008=^`@0008J7@@000038Ud002in00<^OP03000f_SJn008f_P03000ngSkN00<ngP030016
oTKn00A6oP04001:oT[nB_h60003Cah00`00Cai?7P03Cah00`00Cai?7P030002Cah00`00DaiC7P03
Dah00`00DaiC7P0300000e<N0000000EOol00`00Oomoo`06Ool002Moo`04001oogoo0005Ool00`00
Oomoo`10Ool00`00E_eFo@03E_d30002D_d00`00D_eBo@03D_d01000D_eBoE;m10000TkM00<004Jm
A[d00TJm00D004:m@[e2_@0000@jO@03000bG@000080008Z?@03000R70000080008Io0<000<A^`80
00033K/000000080008Eo0<0008J7@03000VGP0000@0008^OP03000f_SJn008f_P03000ngSkN00<n
gP030016oTKn0096oPD000=:oP03001?7TlN00A?7P04001?7TlNCah40006Cah01000DaiC7U<N1000
1E<N0P0057oo00<007ooOol01Woo000WOol01000Oomoo`001Woo0P00@Goo00<005KmE_d00P0000=F
o@00D_d00e;m00@005;mD_eBo@D0009>o@8000=>g@030016_DJm0096_@050012_D:m@[d00004>Wd0
1000<UdbGC9M0`000b8L10000QGL0P000Q6k100000@Ag0000000008Eo0@000@VGP03000^WP000140
00=6oP03001:oT[n00=:oP03001?7TlN009?7PD000=?7P03001?7TlN00A?7P@000=C7P03001C7U<N
00=C7P<001=oo`03001oogoo00Ioo`009goo00@007ooOol000Qoo`03001oogoo03moo`8000AFo@04
001BoE;mD_d40003D_d01000C_e>oDkm5P0000<bG@00:Sd0100000<R70006O`00`0000<Eg0000000
0P000Q7L0P000Q7L1@0000<N?@0000000P000RIN00<002jN;Yh00RjN00<003Zn>[h00SZn00<003kN
?]h00ckN00<004KnA_h00dKn00<004[nB_h00T[n1P000dkn00<004lNCah00dlN00<004lNCah00`00
0TlN00<005<NDah00e<N00<005<NDah01@004goo00<007ooOol01Woo000WOol01000Oomoo`002Goo
00<007ooOol0?goo00@005KmE_eFo@D0009Bo@8000=Bo@@000=>g@05001:gD[MB]d00004A[d01@00
?YdnWCjM00000cYm00<0039M00000`0000<Z?@008Q`0100000<Io00000000P0000<Ag00000000Q7L
10000aWl0`0000<N?@009Uh01P0000<^WP00>[h013Zn00<003kN?]h00SkN00<004KnA_h00TKn1P00
0d[n00<004knC_h00dkn00<004lNCah00TlN10001DlN00<005<NDah00U<N100015<N0P004Woo00<0
07ooOol01Woo000XOol2000;Ool00`00Oomoo`0oOol30003E_d00`00D_eBo@02D_d40002D_d00`00
C]e>g@02C]d00`00B]e:g@02B]d01@00A[e6_DJm000013jM00D003Ym>WdjO@0000<bG@@000<R70@0
00035M`00000008Ag08000034M`001Gl008000<Io003000R?@00008000@VGP03000bWP00010000=6
oP03001:oT[n00=:oP03001>oTkn009>oPH000=?7P03001?7TlN00=?7PD000=C7P03001C7U<N009C
7P8001=oo`03001oogoo00Eoo`00=Woo0P00@7oo00<005KmE_d00UKm00<005;mD_d00P000U;m00<0
05;mD_d00U;m00@004kMC]e>gA400003>Wd0039M008bG@03000Z?BXm00<000048Q`001Wl6O`20002
5M`300000aGL000000035O`200026O`00`008SdR?@0400000bIN000bWP03<Yh00`00>[hj_P03>[h0
0`00@]i2gP02@]h00`00A_i6oP04A_h01000B_i:oT[n1P000dkn00<004lNCah00dlN00<004lNCah0
0`0000=?7P00Dah01E<N00<005<NDah00U<N0`004Woo00<007ooOol01Goo000hOol00`00Oomoo`0n
Ool01000E_eFoEKm0`001E;m00@005;mD_eBo@@0009>g@03001:gD[M009:g@050016_DJmA[d00004
?Yd600000c9M000Z?@02:Sd400026O`400000aGL000000025M`200025O`400048Sd00`00:Wh00004
0002<Yh00`00>[hj_P02>[h00`00@]i2gP03@]h00`00A_i6oP02A_h50003B_h00`00C_i>oP03C_h0
0`00Cai?7P02Cah40005Cah00`00DaiC7P03Dah50002Dah2000AOol00`00Oomoo`05Ool003Uoo`03
001oogoo03ioo`<0009Bo@8000=Bo@D0009Bo@03001>gDkM009>g@05001:gD[MB]d00004A[d01@00
?YdnWCjM00000cIm00<002iM00000`0000@Z?@008Q`R7080008Io004000Eg00000025M`200000aGL
0000000300000aXM0000000200028Sd01P00:WhZORYn000bWQ8000=6oP03001:oT[n00=:oP04001>
oTknC_h60003Cah00`00Cai?7P03Cah00`00DaiC7P030002Dah00`00DaiC7P03Dah00`00Oomoo`0?
Ool00`00Oomoo`05Ool003Yoo`03001oogoo03ioo`03001BoE;m00=Bo@<000=Bo@03001BoE;m009B
o@04001>gDkMC]dA00000cIm000^G@02;Ud500028Q`300035M`200000aGL0000000200026O`30000
0aXM000R?@0600000bYn000bWP03<Yh00`00>[hj_P02>[h00`00@]i2gP03@]h00`00A_i6oP03A_h0
0`00B_i:oP02B_h50003C_h00`00Cai?7P04Cah01000Cai?7TlN10001E<N00<005<NDah00U<N0`00
47oo00<007ooOol01Goo000kOol00`00Oomoo`0nOol00`00D_eBo@0200000e;m001Bo@03D_d01000
D_eBoE;m1@000TkM00D004[MB]e:g@0000A2_@06000nWCjM?Yd003Im100000<^G@009Sd00RHm1000
00@Io0000000008Eg0@000045O`0000000026O`400048Sd00`00:Wh0000400000c:N000j_P04>[h0
0`00@]i2gP03@]h01000A_i6oTKn1P000d[n00<004knC_h00dkn00<004lNCah00TlN1@000dlN00<0
05<NDah00e<N1@000U<N0P003goo00<007ooOol01Goo000lOol2000oOol20004D_d01000D_eBoE;m
10000dkm00D004kMC]e>g@0000A:g@050012_D:m@[d00004?Yd01000=WdfOCIm1@0000@V?@008Q`R
70<000036O`000000080008Eg0<000035O`000000080008N7@@000038Sd002Yn00<ZOP03000bWP00
010000=6oP03001:oT[n00=:oP03001>oTkn009>oPH000=?7P03001?7TlN00=?7P03001C7U<N0080
009C7P03001C7U<N00=C7P03001oogoo00eoo`03001oogoo00Eoo`00?Woo00<007ooOol0?Goo0P00
0e;m10000U;m0P000dkm5@0000<fO@00;Ud00RiM10000b8L100000@Io0000000008Eg0@000046O`0
000000027Qd00`009Ud000040002:Wh01@00<YhbWS:N00001CZn00<004;N@]h00d;N00<004KnA_h0
0dKn00@004[nB_i:oPH000=>oP03001?7TlN00=?7P03001?7TlN009?7P@000EC7P04001C7U<NDah3
000>Ool00`00Oomoo`05Ool003moo`03001oogoo03ioo`<0009Bo@8000=Bo@@0009>g@05001>gDkM
C]d00004A[d01@00@[e2_D:m000013ZM1P0000<^G@009Sd0100000<R70007O`00`0000<Io0000000
0P000QWl0`0000<Io00000000P0012IM00<002in00001@0000<bWP00>[h013Zn00<004;N@]h00T;N
00<004KnA_h00TKn1@000d[n00<004knC_h00dkn00<004lNCah00TlN1P000dlN00<005<NDah00U<N
1@000U<N0P003Woo00<007ooOol017oo0010Ool00`00Oomoo`0nOol00`00D_eBo@02D_d40002D_d0
0`00C]e>g@02C]d01@00C]e>gDkM00000dJm100000=2_@00>Yd00SZM00H003Im=WdfO@00;Ud30002
9Sd00`008Q`0000200027O`4000011Wl000000000QWl100000<R?@0000000P000RIM00<002in;Wh0
0Rin00<003JN000040000dKn00<004[nB_h00d[n00@004knC_i>oPH000=?7P03001?7TlN00=?7PD0
00=C7P03001C7U<N009C7P8000eoo`03001oogoo00Aoo`00@Goo0P00?Woo00<005;mD_d00P000U;m
00D005;mD_eBo@0000=>g@T00096_@030012_@0000P00003=Wd002iM008^G@<000<R70@0008Io0<0
00036O`000000080008N7@<000038Sd002IM00H00003;Wh003JN00<fWP8000@j_P030012gT;N0092
gP8000A6oP03001:oT[n009:oPD000=:oP03001?7TlN00=?7P03001?7TlN00<0009?7P03001C7U<N
00=C7P03001C7U<N009C7P<000aoo`03001oogoo00Aoo`00@goo00<007ooOol0?7oo0P0015;m00@0
05;mD_eBo@@000=>g@04001:gD[MB]d20003A[d01@00@[e2_D:m00000cZM00<0039m00000`0000<^
G@009Sd00`000R8L00@001gl0000008Io003000Io000008000046O`0000000027Qd400049Ud00`00
;Wh0000:00000cZn0012gP03@]h;0003B_h00`00B_i:oP03B_h01000Cai?7TlN10001DlN00<005<N
Dah00U<N1P0000AC7P00Dah000]oo`03001oogoo00Aoo`00A7oo00<007ooOol0?7oo0P000e;m1000
0U;m00<004kMC]d00TkM00@004[MB]e:g@P00003@[d003ZM00<jW@04000bOC9m<Wd500000bHm0000
000200037O`2000011Wl000Io1Wl0`0000<Io00000000P000R8m100000<VG@00;Wh00bin00<003JN
=Yh00cJN00<003jn00001@000d;N00<004KnA_h00TKn00<004[nB_h00d[n00<004[nB_h00T[n1@00
0dlN00<004lNCah00dlN1@000e<N00<005<NDah00U<N0`002Woo00<007ooOol017oo0015Ool00`00
Oomoo`0mOol30002D_d20003D_d01000C]e>gDkM10000d[M00@004JmA[e6_@X00003<Wd002iM008^
G@@000078Q`00000000Mo0006O`00P000aWl00<001hL00000P0000@N7@000000008R?@03000ZGP00
00@0008^OP03000fWSJN008fWP03000n_Sjn008n_P030012gT;N00=2gP030016oTKn0096oP03001:
oT[n009:oPH000=:oP03001?7TlN009?7P03001?7TlN00<0009?7P03001C7U<N00=C7P03001C7U<N
00@00003Dah007oo00Qoo`03001oogoo00Aoo`00AWoo0P00?Woo00<005;mD_d00U;m20000TkM0P00
0d[M00D004JmA[e6_@0000<nW@8000<jO@H00003;Ud002Hm00<000038Q`000000080008Io080008I
o003000N71hL00<000037Qd0000000<000<ZGP03000bOP0001/000=:oP03001:oT[n00=:oP04001?
7TlNCah40005Cah00`00DaiC7P02Dah40004Dah20009Ool00`00Oomoo`04Ool004Qoo`03001oogoo
03]oo`03001BoE;m0080009Bo@05001BoE;mD_d00003C]dA00000cYm000bG@02<Ud00`00:Sd00002
00029Sd400001Agl000000006O`010000ahL00<0028m00000P000RHm100000<ZGP00<Wh00c9n00<0
03Jn=[h00SJn00<003jn?[h00Sjn00<004;N@]h00d;N00<004KnA_h00TKn0P0014[n00<004[nB_h0
0T[n1@000dkn0P000dlN1P000dlN00<005<NDah00U<N0`0027oo00<007ooOol017oo0019Ool00`00
Oomoo`0kOol20004D_d01000D_eBoE;m10000dkM00D004[MB]e:g@0000=6_@03000nWCjM008nW@H0
0004<Ud002Xm:Sd400028Q`200027O`4000011gl000000000QhL00<0028m8Sd00`000RHm00<002Yn
0000100000<bOP00=[h00cJn00<003jn?[h00cjn00<004;N@]h00T;N00<004KnA_h00TKn2`000d[n
00<004knC_h00dkn10000dlN00<004lNCah00dlN00<005<NDah0100000=C7P00Ool01goo00<007oo
Ool00goo001:Ool00`00Oomoo`0kOol20003D_d40002C_d00`00C]e>g@02C]d00`00B]e:g@02B]d0
1@00A[e6_DJm00000cjM00H003Im=WdfO@00<Ud30002:Sd00`009Sd0000200028Q`300027O`30000
0agl0000000200048Sd40004:Wh00`00<Yh0000E0003A_h00`00B_i:oP02B_h20004B_h00`00C_i>
oP030002C_h00`00Cai?7P03Cah01000Cai?7TlN100015<N0P0027oo00<007ooOol00goo001;Ool2
000mOol30002D_d20003C_d01000C]e>gDkM4@0000<fO@00<Ud00S9M0`000bHm100000@N70000000
008Mo0@000037Qd00000008000048Sd002IM9Ud400000bYn000bWP03<Yh00`00>[hj_P02>[h00`00
?[hn_P03?[h00`00@]i2gP02@]h00`00A_i6oP03A_h00`00B_i:oP02B_h90005C_h00`00Cai?7P02
Cah50003Cah00`00DaiC7P02Dah30007Ool00`00Oomoo`03Ool004eoo`03001oogoo03]oo`8000=B
o@P0009>g@03001:gD[M009:g@050012_D:m@[d00003?Yd00`00=Wd0000300000c9M000Z?@040000
0bHm000R700300000ahL0000000200027Q`300000ahM000R?@0400029Ud00`00;Wh000040002<Yh0
0`00>[hj_P02>[h00`00?[hn_P02?[h00`00@]i2gP03@]h01000A_i6oTKn1P000d[n00<004[nB_h0
0T[n00<004knC_h00Tkn1P000dlN00<004lNCah00TlN00<005<NDah0100000=C7P00Ool01Goo00<0
07ooOol00goo001>Ool00`00Oomoo`0lOol30002D_d01@00C_e>oDkm000014kM00D004[MB]e:g@00
00=2_@03000nWCjM008nW@04000fOCIm=Wd500000bXm000V?@0400000b8L00000002000011hL0000
00000QhL10000b8m100012in00<003:N00005@000dKn00<004[nB_h00d[n00@004[nB_i:oPH000=>
oP03001?7TlN00=?7P04001?7TlNCah40004Dah20006Ool00`00Oomoo`03Ool004moo`03001oogoo
03aoo`05001BoE;mD_d00003C_dE00000cIm000bG@02<Ud400039Sd400001AhL000000007Q`01000
0R8M100000<R?@00:Ud010000Rin00<003:N<Yh00S:N00<003Zn>[h00SZn00<003kN?]h00SkN00<0
04;N@]h00d;N00<004KnA_h00TKn00<004[nB_h00T[n1@000d[n00<004knC_h00dkn00<004lNCah0
0TlN1@000dlN00<005<NDah00U<N0`001Goo00<007ooOol00goo001@Ool2000lOol01000D_eBoE;m
10005Goo100000@bG@00:SdZ?@<000049Sd0028M8Qd200027Q`400000ahM0000000200000b8M000V
?@0B00000cZn000ngP04?]h01000@]i2gT;N2`000d[n00<004[nB_h00d[n00@004knC_i>oPH000=?
7P03001?7TlN009?7PD00004Dah005<N0004Ool00`00Oomoo`03Ool0059oo`03001oogoo03Yoo`<0
01eoo`@0008Z?@D0008R7@<0008N70<000037Qd0028m00@0008V?@03000ZG@0000@00003;Wh003JN
00<fWP03000j_P0000X000=2gP030016oTKn0096oP03001:oT[n00=:oP03001:oT[n009:oPD000=>
oP03001?7TlN00=?7P03001?7TlN0080009?7P03001C7U<N009C7P8000Eoo`03001oogoo009oo`00
Dgoo00<007ooOol0>Goo00<007ooOol08Goo0`000bHm100000@N70000000008N70@000<R?@@000@Z
G@03000^OP0000@00003=Yh003Zn00<j_P8000@ngP030012gT;N0092gP030016oTKn00=6oP04001:
oT[nB_h60003B_h00`00C_i>oP02C_h00`00Cai?7P02Cah40004Cah00`00DaiC7P02Dah30004Ool0
0`00Oomoo`02Ool005Aoo`03001oogoo03Qoo`03001oogoo02=oo`@000039Sd0028M00<000037Q`0
00000080008R7@@000038Sd002Hm00@00003:Ud002in00<^OP03000fWP0001X000=:oP03001:oT[n
00=:oP04001>oTknC_h60003Cah01000Cai?7TlN1P000U<N0P000goo00<007ooOol00Woo001EOol0
0`00Oomoo`0gOol00`00Oomoo`0WOol300028Qd4000011hM000000000R8M00<0028m00000`0000<V
?@00:Ud01@0000<^OP00=Yh00cJN00<003Zn>[h00cZn00<003kN?]h00SkN00<004;N@]h00T;N00<0
04KnA_h00dKn00<004[nB_h00T[n00<004[nB_h00T[n1@000dkn00<004lNCah00TlN1@000dlN00<0
05<NDah00U<N0P000goo00<007ooOol00Woo001FOol2000gOol00`00Oomoo`0ZOol300028Qd20002
7Qd400048Sd40003:Ud00`00<Wh000040002=Yh00`00>[hj_P02>[h00`00?]hngP02?]h00`00@]i2
gP03@]h01000A_i6oTKn2`000d[n00@004knC_i>oPH000=?7P03001?7TlN009?7P03001C7U<N009C
7P<0009oo`03001oogoo009oo`00F7oo00<007ooOol0=7oo00<007ooOol0;7oo0`000R8M0`000R8M
100000@R?@009UdVG@@00003:Ud0039n008bOP03000fWP0001D000=6oP03001:oT[n009:oP8000=:
oPD000=>oP03001?7TlN00=?7P04001?7TlNCah60002Dah200000goo001oo`03Ool005Uoo`03001o
ogoo03=oo`03001oogoo02moo`<000038Qd00000008000038Qd002Hm00@0008VG@03000^OP0000@0
0003<Wh003JN00<fWP03000j_SZn008j_P03000ngSkN008ngP030012gT;N00=2gP030016oTKn0096
oP03001:oT[n009:oPD000=:oP03001>oTkn009>oP03001?7TlN009?7PD000=?7P03001C7U<N009C
7P@000Aoo`00FWoo00<007ooOol0<Woo00<007ooOol0<Goo0`0000<R7@008Sd010000RHm100012in
00<0039n00002@0000<j_P00?]h00ckN00<004;N@]h00T;N2`000d[n00<004[nB_h00T[n00@004kn
C_i>oPH000=?7P03001?7TlN009?7PD00004Dah005<N0004Ool005]oo`80039oo`03001oogoo03Ao
o`<000<R?@@0008ZG@@00003;Wh0039n00<bOP03000fWSJN008fWP03000n_P0000X000=6gP030016
oTKn0096oP03001:oT[n00=:oP04001:oT[nB_h50003C_h00`00Cai?7P03Cah00`00Cai?7P020002
Cah00`00DaiC7P02Dah30003Ool005eoo`03001oogoo02moo`03001oogoo03Ioo`@000038Sd002Hm
00@00003:Ud002in00D00003<Wh003JN00<fWP03000n_Sjn008n_P030012gT;N0092gP8000A6gP03
0016oTKn0096oP03001:oT[n009:oPD000=:oP03001>oTkn00=>oP04001?7TlNCah40004Cah00`00
DaiC7P02Dah30003Ool005ioo`03001oogoo02ioo`03001oogoo03Yoo`@000039Sd002YM00D00003
;Wh003:N020000=:oP03001:oT[n009:oP03001>oTkn009>oPD000=?7P04001?7TlNCah600000goo
00000003Ool005moo`03001oogoo02aoo`03001oogoo03moo`<000<ZG@03000^OP0000@00003<Yh0
03ZN00<jWP03000n_Sjn008n_P030012gT;N00=2gP030016gTKN0096gP030016oTKn0096oP03001:
oT[n009:oP/000=>oP04001?7TlNCah50007Ool20003Ool0061oo`8002aoo`03001oogoo045oo`@0
0003:Ud002in008^OP03000fWP0000@0008jWP03000n_Sjn008n_P030012gT;N0092gP030016gTKN
0096gP040016oTKnA_h60003B_h00`00C_i>oP02C_h00`00C_i>oP02C_h50008Ool40005Ool0069o
o`03001oogoo02Uoo`03001oogoo04Eoo`D00003;Wh003JN008fWP03000j_P0001@000=6oP03001:
oT[n00=:oP04001>oTknC_h60009Ool40009Ool006=oo`03001oogoo02Qoo`03001oogoo04Yoo`D0
0003=Yh003Zn00<j_P03000n_Sjn008n_P030012gT;N0092gP030016gTKN0096gP030016oTKn00=6
oP04001:oT[nB_h5000;Ool4000=Ool006Aoo`03001oogoo02Moo`03001oogoo04moo`X00003?[h0
04;N0092gQ0000aoo`@0015oo`00IGoo0P009goo00<007ooOol0FGoo1@0067oo10005Goo001WOol0
0`00Oomoo`0TOol00`00Oomoo`1bOol4000IOol006Qoo`03001oogoo02=oo`03001oogoo06moo`<0
01eoo`00JGoo00<007ooOol08Woo00<007ooOol0Jgoo100087oo001ZOol00`00Oomoo`0QOol00`00
Oomoo`1WOol4000TOol006]oo`80025oo`03001oogoo06=oo`@002Qoo`00KGoo00<007ooOol07Woo
00<007ooOol0Ggoo1000;7oo001^Ool00`00Oomoo`0MOol00`00Oomoo`1KOol4000`Ool006moo`03
001oogoo01aoo`03001oogoo05Moo`@003Aoo`00L7oo0P0077oo00<007ooOol0Dgoo1000>7oo001b
Ool00`00Oomoo`0IOol00`00Oomoo`1@Ool3000lOol007=oo`03001oogoo01Qoo`03001oogoo04ao
o`@003moo`00M7oo00<007ooOol05goo00<007ooOol0B7oo1000@goo001eOol2000GOol00`00Oomo
o`14Ool40017Ool007Moo`03001oogoo01Aoo`03001oogoo041oo`@004]oo`00N7oo00<007ooOol0
4goo00<007ooOol0?7oo1000Cgoo001iOol00`00Oomoo`0BOol00`00Oomoo`0hOol4001COol007Yo
o`80019oo`03001oogoo03Eoo`<005Moo`00O7oo00<007ooOol03goo00<007ooOol0<Goo1000FWoo
001mOol00`00Oomoo`0>Ool00`00Oomoo`0]Ool4001NOol007ioo`03001oogoo00aoo`03001oogoo
02Yoo`@0069oo`00Ogoo0P0037oo00<007ooOol09Woo1000IWoo0021Ool00`00Oomoo`09Ool00`00
Oomoo`0ROol4001ZOol0089oo`03001oogoo00Qoo`03001oogoo01ioo`@006ioo`00Pgoo00<007oo
Ool01goo00<007ooOol06Woo1000LWoo0024Ool00`00Oomoo`06Ool00`00Oomoo`0GOol3001fOol0
08Eoo`8000Ioo`03001oogoo01=oo`@007Uoo`00Qgoo00<007ooOol00goo00<007ooOol03goo1000
OGoo0028Ool00`00Oomoo`02Ool00`00Oomoo`0;Ool40021Ool008Uoo`05001oogooOol00009Ool4
0025Ool008Yoo`800003Ool007oo00Aoo`@008Uoo`00S7oo0P0000=oo`0000000P00SGoo002=Ool2
002AOol00?moob5oo`00ogoo8Goo003oOolQOol00?moob5oo`00ogoo8Goo003oOolQOol00?moob5o
o`00ogoo8Goo003oOolQOol00?moob5oo`00ogoo8Goo003oOolQOol00?moob5oo`00ogoo8Goo0000
\
\>"],
  ImageRangeCache->{{{0, 287}, {232.563, 0}} -> {-0.191149, -0.0395837, \
0.0040191, 0.0040191}}],

Cell[BoxData[
    TagBox[\(\[SkeletonIndicator]  SurfaceGraphics  \[SkeletonIndicator]\),
      False,
      Editable->False]], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Monte Carlo simulation", "Section"],

Cell["Remember how we defined the model:", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(model\)], "Input"],

Cell[BoxData[
    \(BlackScholesModel[100, 0.1`, 0.01`, 0.15`]\)], "Output"]
}, Open  ]],

Cell["\<\
We can simulate a random path under this model with Monte Carlo simulation\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?MonteCarloPath\)\)], "Input"],

Cell[BoxData[
    \("MonteCarloPath[model, length, timesteps] gives a simulated path for \
the given model under the risk neutral measure."\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(ListPlot[MonteCarloPath[model, 15, 1000], 
      PlotJoined \[Rule] True]\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.000952381 -0.379533 0.00596161 [
[.21429 .0849 -9 -9 ]
[.21429 .0849 9 0 ]
[.40476 .0849 -9 -9 ]
[.40476 .0849 9 0 ]
[.59524 .0849 -9 -9 ]
[.59524 .0849 9 0 ]
[.78571 .0849 -9 -9 ]
[.78571 .0849 9 0 ]
[.97619 .0849 -12 -9 ]
[.97619 .0849 12 0 ]
[.01131 .21663 -18 -4.5 ]
[.01131 .21663 0 4.5 ]
[.01131 .33586 -18 -4.5 ]
[.01131 .33586 0 4.5 ]
[.01131 .45509 -18 -4.5 ]
[.01131 .45509 0 4.5 ]
[.01131 .57433 -18 -4.5 ]
[.01131 .57433 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.21429 .0974 m
.21429 .10365 L
s
[(200)] .21429 .0849 0 1 Mshowa
.40476 .0974 m
.40476 .10365 L
s
[(400)] .40476 .0849 0 1 Mshowa
.59524 .0974 m
.59524 .10365 L
s
[(600)] .59524 .0849 0 1 Mshowa
.78571 .0974 m
.78571 .10365 L
s
[(800)] .78571 .0849 0 1 Mshowa
.97619 .0974 m
.97619 .10365 L
s
[(1000)] .97619 .0849 0 1 Mshowa
.125 Mabswid
.07143 .0974 m
.07143 .10115 L
s
.11905 .0974 m
.11905 .10115 L
s
.16667 .0974 m
.16667 .10115 L
s
.2619 .0974 m
.2619 .10115 L
s
.30952 .0974 m
.30952 .10115 L
s
.35714 .0974 m
.35714 .10115 L
s
.45238 .0974 m
.45238 .10115 L
s
.5 .0974 m
.5 .10115 L
s
.54762 .0974 m
.54762 .10115 L
s
.64286 .0974 m
.64286 .10115 L
s
.69048 .0974 m
.69048 .10115 L
s
.7381 .0974 m
.7381 .10115 L
s
.83333 .0974 m
.83333 .10115 L
s
.88095 .0974 m
.88095 .10115 L
s
.92857 .0974 m
.92857 .10115 L
s
.25 Mabswid
0 .0974 m
1 .0974 L
s
.02381 .21663 m
.03006 .21663 L
s
[(100)] .01131 .21663 1 0 Mshowa
.02381 .33586 m
.03006 .33586 L
s
[(120)] .01131 .33586 1 0 Mshowa
.02381 .45509 m
.03006 .45509 L
s
[(140)] .01131 .45509 1 0 Mshowa
.02381 .57433 m
.03006 .57433 L
s
[(160)] .01131 .57433 1 0 Mshowa
.125 Mabswid
.02381 .1272 m
.02756 .1272 L
s
.02381 .15701 m
.02756 .15701 L
s
.02381 .18682 m
.02756 .18682 L
s
.02381 .24644 m
.02756 .24644 L
s
.02381 .27624 m
.02756 .27624 L
s
.02381 .30605 m
.02756 .30605 L
s
.02381 .36567 m
.02756 .36567 L
s
.02381 .39548 m
.02756 .39548 L
s
.02381 .42529 m
.02756 .42529 L
s
.02381 .4849 m
.02756 .4849 L
s
.02381 .51471 m
.02756 .51471 L
s
.02381 .54452 m
.02756 .54452 L
s
.02381 .06759 m
.02756 .06759 L
s
.02381 .03778 m
.02756 .03778 L
s
.02381 .00797 m
.02756 .00797 L
s
.02381 .60413 m
.02756 .60413 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02476 .21846 m
.02571 .22947 L
.02667 .2461 L
.02762 .24262 L
.02857 .26264 L
.02952 .25304 L
.03048 .22227 L
.03143 .2317 L
.03238 .22102 L
.03333 .24313 L
.03429 .25113 L
.03524 .23665 L
.03619 .23846 L
.03714 .24146 L
.0381 .23545 L
.03905 .2436 L
.04 .25311 L
.04095 .26826 L
.0419 .25893 L
.04286 .24702 L
.04381 .24327 L
.04476 .23086 L
.04571 .23788 L
.04667 .22436 L
.04762 .21416 L
.04857 .20394 L
.04952 .21631 L
.05048 .22203 L
.05143 .21866 L
.05238 .21242 L
.05333 .21166 L
.05429 .21777 L
.05524 .20472 L
.05619 .21587 L
.05714 .20782 L
.0581 .21626 L
.05905 .21984 L
.06 .21219 L
.06095 .22289 L
.0619 .2267 L
.06286 .20436 L
.06381 .216 L
.06476 .21582 L
.06571 .19659 L
.06667 .19463 L
.06762 .19103 L
.06857 .19402 L
.06952 .19974 L
.07048 .19813 L
.07143 .2195 L
Mistroke
.07238 .20636 L
.07333 .22047 L
.07429 .21991 L
.07524 .21177 L
.07619 .20843 L
.07714 .22628 L
.0781 .21499 L
.07905 .21983 L
.08 .21478 L
.08095 .21645 L
.0819 .2217 L
.08286 .21443 L
.08381 .22015 L
.08476 .21809 L
.08571 .19879 L
.08667 .19272 L
.08762 .1978 L
.08857 .2017 L
.08952 .21926 L
.09048 .21618 L
.09143 .21631 L
.09238 .20802 L
.09333 .21087 L
.09429 .18768 L
.09524 .19637 L
.09619 .18866 L
.09714 .19871 L
.0981 .18158 L
.09905 .18411 L
.1 .16827 L
.10095 .18027 L
.1019 .18431 L
.10286 .17255 L
.10381 .18023 L
.10476 .17089 L
.10571 .15294 L
.10667 .15241 L
.10762 .16295 L
.10857 .1635 L
.10952 .16107 L
.11048 .16862 L
.11143 .17514 L
.11238 .17159 L
.11333 .17058 L
.11429 .17773 L
.11524 .17241 L
.11619 .16787 L
.11714 .15736 L
.1181 .15257 L
.11905 .15032 L
Mistroke
.12 .15333 L
.12095 .13519 L
.1219 .12802 L
.12286 .12092 L
.12381 .12823 L
.12476 .10398 L
.12571 .09693 L
.12667 .10768 L
.12762 .10862 L
.12857 .08669 L
.12952 .0831 L
.13048 .06138 L
.13143 .0621 L
.13238 .05116 L
.13333 .04109 L
.13429 .03518 L
.13524 .04809 L
.13619 .05569 L
.13714 .05782 L
.1381 .06071 L
.13905 .05881 L
.14 .062 L
.14095 .05077 L
.1419 .05552 L
.14286 .05399 L
.14381 .04401 L
.14476 .04666 L
.14571 .06273 L
.14667 .06192 L
.14762 .05578 L
.14857 .04571 L
.14952 .04533 L
.15048 .05473 L
.15143 .06094 L
.15238 .06083 L
.15333 .0694 L
.15429 .07118 L
.15524 .08253 L
.15619 .07314 L
.15714 .06272 L
.1581 .05893 L
.15905 .06135 L
.16 .06441 L
.16095 .0541 L
.1619 .05855 L
.16286 .05901 L
.16381 .06644 L
.16476 .0772 L
.16571 .08908 L
.16667 .09537 L
Mistroke
.16762 .09323 L
.16857 .09502 L
.16952 .10014 L
.17048 .09804 L
.17143 .1075 L
.17238 .09374 L
.17333 .09075 L
.17429 .09418 L
.17524 .09134 L
.17619 .08906 L
.17714 .10653 L
.1781 .10781 L
.17905 .11234 L
.18 .08848 L
.18095 .1027 L
.1819 .11883 L
.18286 .11599 L
.18381 .11953 L
.18476 .12013 L
.18571 .11373 L
.18667 .11004 L
.18762 .11073 L
.18857 .10622 L
.18952 .1039 L
.19048 .1137 L
.19143 .12623 L
.19238 .11766 L
.19333 .11303 L
.19429 .12102 L
.19524 .11473 L
.19619 .11886 L
.19714 .11683 L
.1981 .1007 L
.19905 .0999 L
.2 .10074 L
.20095 .08825 L
.2019 .08253 L
.20286 .08166 L
.20381 .07714 L
.20476 .08622 L
.20571 .07617 L
.20667 .0745 L
.20762 .05856 L
.20857 .0728 L
.20952 .06222 L
.21048 .04572 L
.21143 .04376 L
.21238 .0382 L
.21333 .03892 L
.21429 .05602 L
Mistroke
.21524 .04586 L
.21619 .04658 L
.21714 .05085 L
.2181 .04089 L
.21905 .04731 L
.22 .05051 L
.22095 .05584 L
.2219 .04863 L
.22286 .05784 L
.22381 .05684 L
.22476 .06432 L
.22571 .07101 L
.22667 .07083 L
.22762 .07829 L
.22857 .07362 L
.22952 .07456 L
.23048 .07027 L
.23143 .06136 L
.23238 .05593 L
.23333 .06283 L
.23429 .06722 L
.23524 .05758 L
.23619 .05663 L
.23714 .07279 L
.2381 .0743 L
.23905 .07128 L
.24 .06927 L
.24095 .05567 L
.2419 .03732 L
.24286 .04785 L
.24381 .0514 L
.24476 .04979 L
.24571 .05942 L
.24667 .04515 L
.24762 .04296 L
.24857 .04587 L
.24952 .05012 L
.25048 .04167 L
.25143 .03015 L
.25238 .03084 L
.25333 .0344 L
.25429 .03704 L
.25524 .02698 L
.25619 .0338 L
.25714 .02726 L
.2581 .02566 L
.25905 .01472 L
.26 .01867 L
.26095 .02064 L
.2619 .02703 L
Mistroke
.26286 .03579 L
.26381 .0402 L
.26476 .04122 L
.26571 .03584 L
.26667 .04068 L
.26762 .04441 L
.26857 .0405 L
.26952 .04437 L
.27048 .04732 L
.27143 .05346 L
.27238 .05263 L
.27333 .06322 L
.27429 .04474 L
.27524 .04896 L
.27619 .05571 L
.27714 .06405 L
.2781 .06289 L
.27905 .0668 L
.28 .05621 L
.28095 .05236 L
.2819 .05624 L
.28286 .05358 L
.28381 .04845 L
.28476 .04959 L
.28571 .04626 L
.28667 .04229 L
.28762 .04542 L
.28857 .05514 L
.28952 .05176 L
.29048 .04839 L
.29143 .03878 L
.29238 .03451 L
.29333 .03466 L
.29429 .04154 L
.29524 .06277 L
.29619 .06636 L
.29714 .08298 L
.2981 .0853 L
.29905 .08616 L
.3 .07769 L
.30095 .07296 L
.3019 .08039 L
.30286 .09229 L
.30381 .08738 L
.30476 .07871 L
.30571 .07011 L
.30667 .06386 L
.30762 .07459 L
.30857 .08197 L
.30952 .06781 L
Mistroke
.31048 .07598 L
.31143 .0596 L
.31238 .06161 L
.31333 .03982 L
.31429 .0472 L
.31524 .04013 L
.31619 .04809 L
.31714 .06472 L
.3181 .07635 L
.31905 .09399 L
.32 .09622 L
.32095 .10021 L
.3219 .09842 L
.32286 .09105 L
.32381 .10211 L
.32476 .11753 L
.32571 .12059 L
.32667 .11858 L
.32762 .11244 L
.32857 .11281 L
.32952 .12267 L
.33048 .15229 L
.33143 .15542 L
.33238 .14773 L
.33333 .15899 L
.33429 .14734 L
.33524 .16462 L
.33619 .17893 L
.33714 .15282 L
.3381 .15047 L
.33905 .15591 L
.34 .16315 L
.34095 .17149 L
.3419 .16579 L
.34286 .17229 L
.34381 .15658 L
.34476 .15717 L
.34571 .15051 L
.34667 .1634 L
.34762 .14915 L
.34857 .13947 L
.34952 .14632 L
.35048 .15802 L
.35143 .16872 L
.35238 .16938 L
.35333 .17987 L
.35429 .18496 L
.35524 .18693 L
.35619 .19305 L
.35714 .18185 L
Mistroke
.3581 .17065 L
.35905 .15766 L
.36 .14864 L
.36095 .14511 L
.3619 .13764 L
.36286 .12423 L
.36381 .11386 L
.36476 .09372 L
.36571 .09672 L
.36667 .09432 L
.36762 .10896 L
.36857 .1116 L
.36952 .11218 L
.37048 .10246 L
.37143 .1112 L
.37238 .10626 L
.37333 .09148 L
.37429 .09172 L
.37524 .10313 L
.37619 .10852 L
.37714 .11487 L
.3781 .11484 L
.37905 .12015 L
.38 .11261 L
.38095 .1099 L
.3819 .10654 L
.38286 .09435 L
.38381 .0989 L
.38476 .1071 L
.38571 .10338 L
.38667 .10634 L
.38762 .10498 L
.38857 .10384 L
.38952 .11133 L
.39048 .11399 L
.39143 .10877 L
.39238 .10431 L
.39333 .09126 L
.39429 .09918 L
.39524 .09669 L
.39619 .09202 L
.39714 .09027 L
.3981 .09055 L
.39905 .09368 L
.4 .081 L
.40095 .08567 L
.4019 .08609 L
.40286 .08938 L
.40381 .07888 L
.40476 .08962 L
Mistroke
.40571 .0806 L
.40667 .08845 L
.40762 .10346 L
.40857 .10104 L
.40952 .10011 L
.41048 .10869 L
.41143 .11336 L
.41238 .11017 L
.41333 .10943 L
.41429 .11722 L
.41524 .13498 L
.41619 .13089 L
.41714 .12334 L
.4181 .12623 L
.41905 .124 L
.42 .11371 L
.42095 .1182 L
.4219 .11825 L
.42286 .11141 L
.42381 .11777 L
.42476 .11573 L
.42571 .10548 L
.42667 .10629 L
.42762 .1182 L
.42857 .1118 L
.42952 .10832 L
.43048 .11074 L
.43143 .09872 L
.43238 .10682 L
.43333 .09434 L
.43429 .09669 L
.43524 .10483 L
.43619 .10753 L
.43714 .11548 L
.4381 .10897 L
.43905 .10723 L
.44 .09281 L
.44095 .08047 L
.4419 .09889 L
.44286 .09639 L
.44381 .10109 L
.44476 .11723 L
.44571 .12168 L
.44667 .12585 L
.44762 .13443 L
.44857 .13473 L
.44952 .12319 L
.45048 .12509 L
.45143 .11176 L
.45238 .09567 L
Mistroke
.45333 .08885 L
.45429 .09906 L
.45524 .1041 L
.45619 .08975 L
.45714 .09016 L
.4581 .10814 L
.45905 .10478 L
.46 .11558 L
.46095 .12765 L
.4619 .13162 L
.46286 .14307 L
.46381 .15207 L
.46476 .14557 L
.46571 .15438 L
.46667 .15129 L
.46762 .16012 L
.46857 .17711 L
.46952 .18285 L
.47048 .17318 L
.47143 .18929 L
.47238 .19338 L
.47333 .19207 L
.47429 .19317 L
.47524 .16239 L
.47619 .16018 L
.47714 .1478 L
.4781 .13255 L
.47905 .13517 L
.48 .12931 L
.48095 .13155 L
.4819 .13097 L
.48286 .12214 L
.48381 .10561 L
.48476 .10038 L
.48571 .11871 L
.48667 .12285 L
.48762 .13276 L
.48857 .12509 L
.48952 .12578 L
.49048 .13242 L
.49143 .13301 L
.49238 .13002 L
.49333 .1257 L
.49429 .12395 L
.49524 .12828 L
.49619 .12148 L
.49714 .1281 L
.4981 .11984 L
.49905 .12648 L
.5 .11225 L
Mistroke
.50095 .09913 L
.5019 .10647 L
.50286 .11221 L
.50381 .10703 L
.50476 .11251 L
.50571 .11187 L
.50667 .10988 L
.50762 .10616 L
.50857 .10937 L
.50952 .10199 L
.51048 .11415 L
.51143 .10281 L
.51238 .10103 L
.51333 .10424 L
.51429 .12017 L
.51524 .1286 L
.51619 .13704 L
.51714 .14097 L
.5181 .14255 L
.51905 .14124 L
.52 .14575 L
.52095 .1617 L
.5219 .1581 L
.52286 .14063 L
.52381 .1542 L
.52476 .1741 L
.52571 .17441 L
.52667 .18196 L
.52762 .20339 L
.52857 .20539 L
.52952 .19855 L
.53048 .19394 L
.53143 .20221 L
.53238 .20386 L
.53333 .20492 L
.53429 .22978 L
.53524 .22026 L
.53619 .23237 L
.53714 .24402 L
.5381 .25265 L
.53905 .26143 L
.54 .25464 L
.54095 .25276 L
.5419 .26027 L
.54286 .27271 L
.54381 .27106 L
.54476 .26876 L
.54571 .25885 L
.54667 .26872 L
.54762 .26325 L
Mistroke
.54857 .24941 L
.54952 .25541 L
.55048 .2355 L
.55143 .23436 L
.55238 .24206 L
.55333 .24758 L
.55429 .25432 L
.55524 .25844 L
.55619 .25288 L
.55714 .23874 L
.5581 .22192 L
.55905 .21851 L
.56 .21994 L
.56095 .24228 L
.5619 .24866 L
.56286 .22893 L
.56381 .23821 L
.56476 .25704 L
.56571 .27273 L
.56667 .29655 L
.56762 .30455 L
.56857 .29308 L
.56952 .29864 L
.57048 .30444 L
.57143 .28375 L
.57238 .29268 L
.57333 .30366 L
.57429 .30944 L
.57524 .32262 L
.57619 .32186 L
.57714 .31784 L
.5781 .33741 L
.57905 .34442 L
.58 .35514 L
.58095 .35869 L
.5819 .34694 L
.58286 .34651 L
.58381 .34492 L
.58476 .3234 L
.58571 .32945 L
.58667 .34179 L
.58762 .35163 L
.58857 .34434 L
.58952 .33462 L
.59048 .36453 L
.59143 .34074 L
.59238 .33403 L
.59333 .32275 L
.59429 .34047 L
.59524 .31601 L
Mistroke
.59619 .31714 L
.59714 .31486 L
.5981 .32733 L
.59905 .33854 L
.6 .34522 L
.60095 .35549 L
.6019 .34298 L
.60286 .32427 L
.60381 .32531 L
.60476 .33238 L
.60571 .30616 L
.60667 .30697 L
.60762 .32533 L
.60857 .33191 L
.60952 .32899 L
.61048 .31756 L
.61143 .29046 L
.61238 .29355 L
.61333 .27928 L
.61429 .30105 L
.61524 .32903 L
.61619 .31845 L
.61714 .31887 L
.6181 .3317 L
.61905 .33341 L
.62 .32061 L
.62095 .31797 L
.6219 .3304 L
.62286 .33055 L
.62381 .32432 L
.62476 .31263 L
.62571 .3416 L
.62667 .33923 L
.62762 .34099 L
.62857 .30267 L
.62952 .30975 L
.63048 .30426 L
.63143 .29647 L
.63238 .29545 L
.63333 .28355 L
.63429 .28022 L
.63524 .27719 L
.63619 .28454 L
.63714 .27992 L
.6381 .30804 L
.63905 .30827 L
.64 .29083 L
.64095 .27519 L
.6419 .27207 L
.64286 .25503 L
Mistroke
.64381 .24345 L
.64476 .2465 L
.64571 .24732 L
.64667 .2618 L
.64762 .26746 L
.64857 .26993 L
.64952 .28134 L
.65048 .26201 L
.65143 .26533 L
.65238 .26361 L
.65333 .26316 L
.65429 .26916 L
.65524 .27223 L
.65619 .2619 L
.65714 .26139 L
.6581 .27197 L
.65905 .25019 L
.66 .26758 L
.66095 .26402 L
.6619 .24449 L
.66286 .26318 L
.66381 .24526 L
.66476 .23638 L
.66571 .23629 L
.66667 .22929 L
.66762 .22805 L
.66857 .23005 L
.66952 .22179 L
.67048 .23333 L
.67143 .23431 L
.67238 .22767 L
.67333 .2317 L
.67429 .23071 L
.67524 .22693 L
.67619 .21864 L
.67714 .21055 L
.6781 .23778 L
.67905 .22479 L
.68 .22986 L
.68095 .21929 L
.6819 .2117 L
.68286 .21318 L
.68381 .21318 L
.68476 .20618 L
.68571 .20164 L
.68667 .21179 L
.68762 .20885 L
.68857 .21619 L
.68952 .22459 L
.69048 .22789 L
Mistroke
.69143 .23255 L
.69238 .22449 L
.69333 .21287 L
.69429 .21291 L
.69524 .21137 L
.69619 .23177 L
.69714 .23836 L
.6981 .23896 L
.69905 .2487 L
.7 .26909 L
.70095 .26421 L
.7019 .25943 L
.70286 .25942 L
.70381 .25014 L
.70476 .25191 L
.70571 .25618 L
.70667 .26911 L
.70762 .29141 L
.70857 .29175 L
.70952 .29406 L
.71048 .2797 L
.71143 .28938 L
.71238 .28125 L
.71333 .28991 L
.71429 .28281 L
.71524 .29231 L
.71619 .28997 L
.71714 .29876 L
.7181 .30392 L
.71905 .31296 L
.72 .32273 L
.72095 .33204 L
.7219 .32779 L
.72286 .33875 L
.72381 .34696 L
.72476 .35271 L
.72571 .35772 L
.72667 .32302 L
.72762 .3332 L
.72857 .32197 L
.72952 .33601 L
.73048 .33179 L
.73143 .32629 L
.73238 .31325 L
.73333 .32386 L
.73429 .33058 L
.73524 .33498 L
.73619 .32513 L
.73714 .35311 L
.7381 .34184 L
Mistroke
.73905 .36186 L
.74 .35426 L
.74095 .36915 L
.7419 .34164 L
.74286 .34779 L
.74381 .36362 L
.74476 .34722 L
.74571 .3357 L
.74667 .34143 L
.74762 .35509 L
.74857 .34987 L
.74952 .33306 L
.75048 .30318 L
.75143 .30527 L
.75238 .31769 L
.75333 .3149 L
.75429 .30629 L
.75524 .3192 L
.75619 .32219 L
.75714 .31379 L
.7581 .30333 L
.75905 .29036 L
.76 .29138 L
.76095 .28311 L
.7619 .2805 L
.76286 .29896 L
.76381 .29576 L
.76476 .30253 L
.76571 .28562 L
.76667 .30514 L
.76762 .28678 L
.76857 .29443 L
.76952 .28393 L
.77048 .28556 L
.77143 .28844 L
.77238 .29756 L
.77333 .30664 L
.77429 .31018 L
.77524 .31952 L
.77619 .31091 L
.77714 .31218 L
.7781 .31704 L
.77905 .32159 L
.78 .3206 L
.78095 .3179 L
.7819 .3341 L
.78286 .34248 L
.78381 .35299 L
.78476 .34651 L
.78571 .35446 L
Mistroke
.78667 .35364 L
.78762 .35529 L
.78857 .38235 L
.78952 .38069 L
.79048 .38585 L
.79143 .39479 L
.79238 .39776 L
.79333 .39773 L
.79429 .41781 L
.79524 .42814 L
.79619 .4326 L
.79714 .43859 L
.7981 .43786 L
.79905 .442 L
.8 .45161 L
.80095 .42373 L
.8019 .41422 L
.80286 .41788 L
.80381 .41819 L
.80476 .40988 L
.80571 .38222 L
.80667 .40731 L
.80762 .42502 L
.80857 .43533 L
.80952 .45786 L
.81048 .4641 L
.81143 .48564 L
.81238 .51101 L
.81333 .5336 L
.81429 .51833 L
.81524 .50502 L
.81619 .51194 L
.81714 .512 L
.8181 .51386 L
.81905 .49458 L
.82 .52985 L
.82095 .51691 L
.8219 .53302 L
.82286 .50474 L
.82381 .5045 L
.82476 .47318 L
.82571 .50421 L
.82667 .49881 L
.82762 .52349 L
.82857 .52024 L
.82952 .53205 L
.83048 .52115 L
.83143 .51203 L
.83238 .51405 L
.83333 .51741 L
Mistroke
.83429 .51862 L
.83524 .50932 L
.83619 .50631 L
.83714 .5401 L
.8381 .55632 L
.83905 .56067 L
.84 .57214 L
.84095 .5868 L
.8419 .5825 L
.84286 .57712 L
.84381 .58655 L
.84476 .59385 L
.84571 .60332 L
.84667 .59778 L
.84762 .59757 L
.84857 .59495 L
.84952 .57786 L
.85048 .55236 L
.85143 .55824 L
.85238 .56527 L
.85333 .55343 L
.85429 .5592 L
.85524 .53171 L
.85619 .53027 L
.85714 .49255 L
.8581 .49044 L
.85905 .47985 L
.86 .47453 L
.86095 .46836 L
.8619 .45902 L
.86286 .44372 L
.86381 .44728 L
.86476 .42003 L
.86571 .41027 L
.86667 .40995 L
.86762 .41429 L
.86857 .41592 L
.86952 .44203 L
.87048 .44815 L
.87143 .43011 L
.87238 .42808 L
.87333 .45103 L
.87429 .44168 L
.87524 .45868 L
.87619 .42771 L
.87714 .42511 L
.8781 .43157 L
.87905 .42903 L
.88 .43359 L
.88095 .42728 L
Mistroke
.8819 .4148 L
.88286 .41474 L
.88381 .4139 L
.88476 .44449 L
.88571 .4466 L
.88667 .44779 L
.88762 .41987 L
.88857 .40161 L
.88952 .41152 L
.89048 .42569 L
.89143 .40901 L
.89238 .38551 L
.89333 .40279 L
.89429 .42074 L
.89524 .40155 L
.89619 .40669 L
.89714 .40056 L
.8981 .40113 L
.89905 .40201 L
.9 .41909 L
.90095 .41624 L
.9019 .42778 L
.90286 .44257 L
.90381 .4255 L
.90476 .42708 L
.90571 .41661 L
.90667 .42581 L
.90762 .42662 L
.90857 .40681 L
.90952 .39642 L
.91048 .38379 L
.91143 .38614 L
.91238 .38502 L
.91333 .39254 L
.91429 .3833 L
.91524 .38567 L
.91619 .38074 L
.91714 .38812 L
.9181 .38666 L
.91905 .36249 L
.92 .38201 L
.92095 .37655 L
.9219 .38378 L
.92286 .38653 L
.92381 .3631 L
.92476 .32993 L
.92571 .32672 L
.92667 .31364 L
.92762 .31945 L
.92857 .32797 L
Mistroke
.92952 .31068 L
.93048 .31334 L
.93143 .3373 L
.93238 .32827 L
.93333 .33872 L
.93429 .36267 L
.93524 .34107 L
.93619 .33847 L
.93714 .32898 L
.9381 .34384 L
.93905 .33159 L
.94 .35196 L
.94095 .36528 L
.9419 .36663 L
.94286 .38196 L
.94381 .37105 L
.94476 .39593 L
.94571 .38816 L
.94667 .38389 L
.94762 .37046 L
.94857 .38169 L
.94952 .40604 L
.95048 .41409 L
.95143 .40317 L
.95238 .38738 L
.95333 .38978 L
.95429 .39237 L
.95524 .3748 L
.95619 .39729 L
.95714 .38104 L
.9581 .36798 L
.95905 .38476 L
.96 .38496 L
.96095 .39016 L
.9619 .39304 L
.96286 .36444 L
.96381 .34876 L
.96476 .34656 L
.96571 .34936 L
.96667 .36641 L
.96762 .3683 L
.96857 .35198 L
.96952 .34689 L
.97048 .34742 L
.97143 .34628 L
.97238 .3482 L
.97333 .34803 L
.97429 .36797 L
.97524 .35992 L
.97619 .35027 L
Mistroke
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/A000`40O003h00Oogoo8Goo003oOolQ
Ool00?moob5oo`00ogoo8Goo003oOolQOol00?moob5oo`005Goo00<007ooOol0ogoo2Goo000EOol0
0`00Oomoo`3oOol9Ool001Eoo`800?moo`Yoo`005Goo00<007ooOol0ogoo2Goo000EOol00`00Oomo
o`0kOol00`00Oomoo`3:Ool001Eoo`03001oogoo03]oo`800<]oo`005Goo00<007ooOol0>goo0P00
bgoo000EOol00`00Oomoo`0jOol3003;Ool001Eoo`03001oogoo03Uoo`@00<]oo`005Goo00<007oo
Ool06Woo00<007ooOol077oo0`0000=oo`00Ool01Goo0P00`Woo000EOol2000KOol00`00Oomoo`0B
Ool00`00Oomoo`05Ool01000Ool000000Woo00<007ooOol017oo0P0017oo00<007ooOol0^goo000E
Ool00`00Oomoo`0JOol01000Oomoo`004Goo00<007oo00001Goo0`000goo0P0017oo0`0017oo00<0
07ooOol0^goo000EOol00`00Oomoo`0JOol200000goo0000000@Ool30005Ool30004Ool30002Ool3
0004Ool2002lOol001Eoo`03001oogoo01Yoo`D0011oo`@000Aoo`<000Eoo`800003Ool000000080
00Aoo`800;aoo`005Goo00<007ooOol06Woo1P0000=oo`00Ool01Goo10000Woo1@000Woo10001Woo
1`0017oo0P0047oo0`000Woo0P0017oo0P0097oo0`0017oo0P0017oo0P009Goo0P0017oo0P0017oo
0P008Goo1@000Woo0P0017oo0P0017oo0P000goo000EOol00`00Oomoo`0IOol00`00Ool000040000
0goo00000005Ool00`00Oomoo`03Ool00`00Ool00002000017oo000007oo0P001Woo10000Woo00<0
07ooOol00Woo0P004Goo00@007ooOol0009oo`04001oogoo0002Ool00`00Oomoo`0QOol01000Oomo
o`000Woo00@007ooOol0009oo`04001oogoo000SOol01000Oomoo`000Woo00@007ooOol0009oo`04
001oogoo000ROol01@00Oomoogoo00000Woo00@007ooOol0009oo`04001oogoo0002Ool10001Ool1
Ool001Eoo`03001oogoo01Uoo`03001oo`0000L000Ioo`03001oogoo009oo`03001oogoo00800004
Ool00000Ool20006Ool30003Ool01000Oomoogoo0`003Woo1@0000Eoo`00Oomoo`00009oo`04001o
ogoo000SOol01000Oomoo`000Woo00@007ooOol0009oo`04001oogoo000SOol01000Oomoo`000Woo
00@007ooOol0009oo`04001oogoo000ROol01@00Oomoogoo00000Woo00@007ooOol0009oo`04001o
ogoo0002Ool10001Ool1Ool001Eoo`03001oogoo01Uoo`03001oogoo00=oo`@000Moo`05001oogoo
Ool00002Ool200000goo000000020007Ool30003Ool00`00Oomoo`0200000goo001oo`0=Ool01000
Oomoo`000Woo00@007ooOol0009oo`04001oogoo000SOol30003Ool01000Oomoo`000Woo00@007oo
Ool002Aoo`8000=oo`04001oogoo0002Ool01000Oomoo`008Woo00D007ooOomoo`00009oo`04001o
ogoo0002Ool01000Oomoo`000Woo0@000Goo0Goo000EOol2000JOol00`00Oomoo`03Ool200000goo
001oo`04Ool01000Oomoo`000Woo00<007ooOol00P0000=oo`0000000P002Goo00D007ooOomoo`00
009oo`800003Ool007oo00ioo`03001oo`00009oo`04001oogoo0002Ool01000Oomoo`0097oo00<0
07ooOol00Woo00@007ooOol0009oo`04001oogoo000SOol01000Oomoo`000Woo00@007ooOol0009o
o`04001oogoo000ROol01@00Oomoogoo00000Woo00@007ooOol0009oo`04001oogoo0002Ool10001
Ool1Ool001Eoo`03001oogoo01Uoo`03001oogoo00Aoo`03001oo`0000Ioo`8000=oo`<0009oo`@0
00ioo`03001oo`0000800003Ool007oo00moo`8000=oo`8000Aoo`8002Eoo`<000=oo`8000Aoo`80
02Eoo`8000Aoo`8000Aoo`80029oo`8000Aoo`8000Aoo`8000Aoo`8000=oo`005Goo00<007ooOol0
6Goo00<007ooOol017oo00<007oo00002Woo0P0017oo0P0000=oo`00Ool03Woo100000=oo`00Ool0
^goo000EOol00`00Oomoo`0IOol00`00Oomoo`04Ool00`00Ool0000:Ool00`00Oomoo`03Ool00`00
Oomoo`0@Ool400000goo001oo`0DOol30008Ool00`00Oomoo`2IOol001Eoo`03001oogoo01Uoo`03
001oogoo00Aoo`04001oogoo0008Ool2000HOol2000017oo001oo`005Goo0`0027oo00<007ooOol0
VGoo000EOol00`00Oomoo`0IOol00`00Oomoo`07Ool01000Ool000001Goo0P0067oo0P000goo00<0
07ooOol04goo0`0027oo00@007ooOol009Qoo`005Goo00<007ooOol06Goo00<007ooOol01goo00@0
07oo000000Eoo`03001oogoo01Qoo`04001oogooOol2000=Ool00`00Oomoo`02Ool50008Ool00`00
Oomoo`02002GOol001Eoo`03001oogoo01Uoo`03001oogoo00Moo`@000Eoo`03001oogoo01aoo`80
00]oo`05001oo`00Ool00002Ool300000goo001oo`05Ool00`00Ool00002Ool2002GOol000ioool0
00d000Ioo`005Goo00<007ooOol02Woo00<007ooOol02Goo00<007oo00002Woo0P0000Ioo`00001o
o`00Ool20003Ool00`00Oomoo`09Ool00`00Oomoo`0:Ool01@00Oomoogoo000027oo00<007oo0000
0P0000=oo`0000000P000Woo0`0017oo0`0000=oo`00Ool00`001Goo0P000goo00@007oo000000Qo
o`03001oogoo00Yoo`03001oogoo00Yoo`03001oogoo00Uoo`03001oogoo00Yoo`03001oogoo00Yo
o`03001oogoo00Uoo`03001oogoo00Yoo`03001oogoo00Yoo`03001oogoo00Yoo`03001oogoo00Yo
o`005Goo00<007ooOol067oo00<007ooOol02Goo00P007oo000007oo001oo`008Goo00<007ooOol0
1goo100000=oo`0000000P0017oo00@007ooOomoo`@00006Ool007oo001oo`001Goo0P000goo1000
R7oo000EOol00`00Oomoo`0HOol00`00Oomoo`0;Ool400000goo001oo`0POol00`00Oomoo`07Ool3
000017oo000007oo0P0017oo0P0000aoo`00000007oo001oo`00Ool007oo0005Ool20003Ool40028
Ool001Eoo`03001oogoo01Qoo`03001oogoo00]oo`H0025oo`8000Qoo`04001oo`00Ool200000goo
001oo`04Ool6000027oo001oo`00Ool007oo0005Ool20003Ool40028Ool001Eoo`03001oogoo01Mo
o`03001oogoo00eoo`D0025oo`8000Qoo`05001oogooOol00009Ool40004Ool20002Ool00`00Oomo
o`03Ool20003Ool01000Oomoo`00R7oo000EOol00`00Oomoo`0GOol00`00Oomoo`0=Ool4000ROol2
0008Ool01@00Oomoogoo00002Goo0P001Woo0P000Woo00<007ooOol00goo0P000Woo00D007ooOomo
o`0008Qoo`005Goo00<007ooOol05goo00<007ooOol03goo00<007ooOol08Woo00<007ooOol01Woo
00<007ooOol02goo0P001Woo0P000Woo00<007ooOol00goo1@000goo00<007ooOol0QWoo000EOol2
000HOol00`00Oomoo`0dOol00`00Oomoo`06Ool00`00Oomoo`0;Ool00`00Oomoo`05Ool20002Ool0
0`00Oomoo`02Ool60004Ool00`00Oomoo`25Ool001Eoo`03001oogoo01Moo`03001oogoo03Aoo`03
001oogoo00Ioo`03001oogoo00]oo`03001oogoo00Ioo`05001oogooOol00003Ool40006Ool00`00
Oomoo`25Ool001Eoo`03001oogoo01Moo`03001oogoo03Aoo`03001oogoo00Ioo`03001oogoo00]o
o`03001oogoo00Ioo`05001oogooOol00003Ool00`00Oomoo`07Ool00`00Oomoo`25Ool001Eoo`03
001oogoo01Moo`03001oogoo03Eoo`05001oogooOol00003Ool00`00Oomoo`0HOol01@00Oomoogoo
00002Goo0`00QGoo000EOol00`00Oomoo`0GOol00`00Oomoo`0eOol01@00Oomoogoo00000goo00<0
07ooOol067oo00D007ooOomoo`0000Uoo`<008Eoo`005Goo00<007ooOol05Woo00<007ooOol0=Woo
0P000Woo0P0000=oo`00Ool06Woo00D007ooOomoo`0000Yoo`8008Eoo`005Goo00<007ooOol04goo
00@007ooOol003Qoo`<00005Ool00000Ool0000KOol20002Ool00`00Oomoo`08Ool20025Ool001Eo
o`03001oogoo01=oo`04001oogoo000hOol600000goo001oo`0KOol00`00Ool0000;Ool20025Ool0
01Eoo`8001Aoo`04001oogoo000hOol600000goo001oo`0KOol00`00Ool0000;Ool20025Ool001Eo
o`03001oogoo019oo`@003Yoo`D00003Ool007oo01]oo`03001oo`0000]oo`8008Eoo`005Goo00<0
07ooOol04Woo00@007oo000003Yoo`<00004Ool007oo000LOol00`00Ool0000<Ool00`00Oomoo`23
Ool001Eoo`03001oogoo015oo`800003Ool0000003Yoo`<00004Ool007oo000LOol00`00Ool0000<
Ool00`00Oomoo`23Ool001Eoo`03001oogoo015oo`800003Ool0000003Yoo`07001oo`00Ool007oo
000077oo0`0037oo0P00Q7oo000EOol00`00Oomoo`0AOol20002Ool00`00Oomoo`0hOol00`00Oomo
o`02Ool2000LOol3000=Ool00`00Oomoo`22Ool001Eoo`03001oogoo015oo`80049oo`8001eoo`80
00eoo`03001oogoo089oo`005Goo0P004Woo0P00@Woo0P007Goo0P003Goo00<007ooOol0PWoo000E
Ool00`00Oomoo`0@Ool20013Ool00`00Oomoo`0LOol2000=Ool00`00Oomoo`22Ool001Eoo`03001o
ogoo00Uoo`03001oogoo009oo`03001oo`0004Aoo`03001oogoo01aoo`8000eoo`8008=oo`005Goo
00<007ooOol027oo0P000goo0P0000=oo`00Ool0LGoo0P00Pgoo000EOol00`00Oomoo`08Ool20003
Ool4001bOol2000XOol00`00Oomoo`1HOol001Eoo`03001oogoo00=oo`04001oo`00Ool30003Ool3
001cOol3000WOol00`00Oomoo`1HOol00003Ool0000000<0009oo`8000Aoo`8000Eoo`03001oogoo
00=oo`03001oo`0000H00004Ool00000001eOol00`00Oomoo`0UOol2001IOol000=oo`05001oogoo
Ool00002Ool01000Oomoo`000Woo00<007ooOol00Woo00<007ooOol00goo1P0000Aoo`00001oo`<0
07Eoo`03001oogoo02=oo`H005Moo`000goo00D007ooOomoo`00009oo`04001oogoo0002Ool00`00
Oomoo`02Ool20004Ool600000goo00000004001eOol00`00Oomoo`0SOol6001GOol000=oo`05001o
ogooOol00002Ool01000Oomoo`000Woo00<007ooOol00Woo00<007ooOol00goo1@000Woo1@00MWoo
00<007ooOol00goo0P007Woo0P0000Aoo`00000005Moo`000goo00D007ooOomoo`00009oo`04001o
ogoo0002Ool00`00Oomoo`02Ool00`00Ool00003Ool20002Ool01000Oomoogoo0P00N7oo00<007oo
Ool00goo0P0077oo100000Aoo`00000005Moo`000Woo0P0017oo0P0017oo0P001Goo00<007oo0000
0goo00D007ooOomoo`0000=oo`03001oogoo07Moo`03001oogoo00=oo`8001]oo`D00004Ool00000
001GOol001Eoo`03001oo`00009oo`80085oo`03001oogoo00=oo`8001]oo`D00004Ool00000001G
Ool001Eoo`04001oo`00Ool30021Ool01P00Oomoogoo001oo`8001]oo`800003Ool007oo009oo`80
05Moo`005Goo1`00PGoo00H007ooOomoo`00Ool3000JOol01000Oomoo`0017oo00<007ooOol0EGoo
000EOol60023Ool00`00Oomoo`05000IOol00`00Oomoo`06Ool00`00Oomoo`1EOol001Eoo`H008=o
o`03001oogoo00D001Aoo`03001oogoo009oo`03001oogoo00Moo`03001oogoo05Aoo`005Goo0P00
00Aoo`00000008=oo`03001oogoo00D001Aoo`04001oogooOol20009Ool2001EOol001Eoo`800004
Ool000000023Ool200001Woo00000000Ool001Aoo`04001oogooOol20009Ool2001EOol001Eoo`80
009oo`8008=oo`D0009oo`03001oogoo019oo`80009oo`8000Uoo`8005Eoo`005Goo0P000Woo0P00
Pgoo0`0000Eoo`00Oomoo`0001Aoo`800004Ool000000009Ool2001EOol001Eoo`80009oo`03001o
ogoo08=oo`8000Aoo`03001oogoo015oo`03001oo`0000@000Uoo`03001oo`0005Aoo`005Goo00D0
07ooOomoo`0008Eoo`8000Aoo`03001oogoo015oo`03001oo`0000<000Yoo`03001oo`0005Aoo`00
5Goo00<007ooOol0Qgoo00<007ooOol00goo00<007ooOol04Goo00<007oo00000`0037oo00<007oo
Ool0DWoo000EOol2002>Ool00`00Oomoo`0AOol00`00Ool0000?Ool00`00Oomoo`1BOol001Eoo`03
001oogoo08eoo`03001oogoo00Yoo`03001oogoo009oo`<00003Ool007oo00ioo`8000aoo`03001o
ogoo04Aoo`005Goo00<007ooOol0SGoo00<007oo00002Woo00<007ooOol00Woo0`0000=oo`00Ool0
3Woo0`002goo00<007oo0000A7oo000EOol00`00Oomoo`2=Ool00`00Ool0000:Ool00`00Oomoo`02
Ool3000AOol3000;Ool40013Ool001Eoo`03001oogoo08eoo`03001oo`0000Uoo`8000Aoo`<0015o
o`<000]oo`@004=oo`005Goo00<007ooOol0SGoo0`002Goo0P0017oo0`004Goo00<007oo00002goo
1000@goo000EOol00`00Oomoo`2=Ool30009Ool20004Ool3000DOol00`00Oomoo`07Ool00`00Ool0
00020013Ool001Eoo`03001oogoo08eoo`<000Uoo`8000=oo`04001oo`00000DOol00`00Oomoo`05
Ool01000Ool007oo0`00@goo000EOol2002?Ool20008Ool30003Ool01000Ool0000057oo00<007oo
Ool01Goo0`000Woo0P00@goo000EOol00`00Oomoo`2?Ool00`00Oomoo`06Ool30003Ool01000Ool0
000057oo00<007ooOol01Goo0`0017oo00<007ooOol09Woo00<007ooOol05goo000EOol00`00Oomo
o`2?Ool00`00Oomoo`06Ool30002Ool2000GOol01@00Oomoogoo00000goo0`0017oo00<007ooOol0
9Goo0P006Goo000EOol00`00Oomoo`2?Ool20004Ool200001Goo00000000Ool00`005goo00D007oo
Oomoo`0000=oo`<000Aoo`8002Ioo`8001Uoo`005Goo00<007ooOol0T7oo00<007ooOol00Woo0P00
00=oo`0000001@005goo00<007ooOol00P000goo00<007oo000017oo0P009Woo0P006Goo000EOol0
0`00Oomoo`2@Ool00`00Ool00002Ool:000GOol00`00Ool000030002Ool00`00Ool00005Ool00`00
Oomoo`0TOol2000IOol00003Ool0000000<00003Ool00000008000=oo`8000Eoo`03001oogoo091o
o`03001oo`00009oo`X001Moo`H0009oo`03001oogoo00Ioo`03001oogoo02=oo`@001Moo`000goo
00D007ooOomoo`0000Eoo`04001oogoo0004Ool00`00Oomoo`2@Ool01000Ool007oo1P0000=oo`00
00000P0067oo1@000Woo00<007ooOol01Woo00<007ooOol08goo00<007oo00000P005Woo0003Ool0
0`00Oomoo`02Ool00`00Oomoo`02Ool01000Oomoo`0017oo0P00TGoo00@007oo001oo`@000=oo`04
001oo`00000HOol500000goo00000008Ool00`00Oomoo`0SOol00`00Ool00002000FOol000=oo`03
001oogoo00=oo`05001oogooOol00002Ool00`00Oomoo`02Ool00`00Oomoo`2@Ool00`00Ool00005
0005Ool2000HOol20002Ool01000Ool0000027oo00<007ooOol08goo00<007oo00000P005Woo0003
Ool01@00Oomoogoo00000Woo00@007ooOol0009oo`03001oogoo009oo`03001oogoo095oo`@00003
Ool0000000Eoo`8001Qoo`80009oo`@000Qoo`03001oogoo02=oo`03001oo`00008001Ioo`000Woo
0P0017oo0P0017oo0P001Goo00<007ooOol0TGoo100000=oo`0000007goo0P000Woo100027oo0P00
97oo00<007oo00000P001Goo00@007oo000000eoo`005Goo00<007ooOol0TGoo00D007oo000007oo
008001moo`80009oo`@000Qoo`8002Aoo`03001oo`00008000Eoo`800003Ool0000000aoo`005Goo
00<007ooOol0TGoo00@007ooOol0009oo`03001oogoo01ioo`03001oogoo00@000Qoo`8002Aoo`05
001oo`00Ool00005Ool200000goo0000000<Ool001Eoo`03001oogoo095oo`04001oogoo000SOol0
0`00Oomoo`03000:Ool00`00Oomoo`0ROol01@00Ool007oo00001Goo0P0000=oo`00Ool037oo000E
Ool00`00Oomoo`2DOol00`00Oomoo`0TOol3000:Ool00`00Oomoo`0POol01`00Ool007oo001oo`00
00Eoo`800003Ool007oo00aoo`005Goo0P00UGoo00<007ooOol09Goo0P002Woo00<007ooOol087oo
00<007oo00000goo00<007ooOol00goo0P0000=oo`00Ool037oo000EOol00`00Oomoo`2lOol00`00
Oomoo`09Ool00`00Oomoo`0POol00`00Ool00003Ool00`00Ool00002Ool300000goo001oo`0<Ool0
01Eoo`03001oogoo0<Uoo`03001oogoo01moo`<000=oo`<0009oo`80011oo`005Goo00<007ooOol0
bGoo00<007ooOol07goo0P001Goo0P0000Aoo`000000011oo`005Goo00<007ooOol0bGoo00<007oo
Ool07goo0P001Goo0P0000Aoo`000000011oo`005Goo00<007ooOol0bGoo00D007ooOomoo`0001Io
o`03001oogoo009oo`@000Eoo`800004Ool00000000@Ool001Eoo`03001oogoo0<Uoo`05001oogoo
Ool0000FOol00`00Oomoo`02Ool40005Ool6000@Ool001Eoo`03001oogoo0<Uoo`05001oogooOol0
000FOol00`00Oomoo`02Ool20007Ool6000@Ool001Eoo`800<Yoo`05001oogooOol0000FOol00`00
Oomoo`02Ool00`00Oomoo`06Ool4000BOol001Eoo`03001oogoo0<Yoo`04001oogoo000FOol20003
Ool00`00Oomoo`07Ool3000BOol001Eoo`03001oogoo0<Yoo`04001oogoo000EOol400000goo001o
o`09Ool2000COol001Eoo`03001oogoo0<Yoo`04001oogoo000EOol400000goo001oo`09Ool2000C
Ool001Eoo`03001oogoo0<Yoo`04001oogoo000?Ool00`00Oomoo`03Ool400000goo001oo`09Ool2
000COol001Eoo`03001oogoo0<Yoo`04001oo`00000?Ool20002Ool600000goo001oo`0:Ool00`00
Oomoo`0AOol001Eoo`03001oogoo0<Yoo`03001oo`00008000ioo`80009oo`H00003Ool007oo01io
o`005Goo00<007ooOol0bWoo00D007oo001oo`0000ioo`80009oo`P001moo`005Goo0P00bgoo00D0
07oo001oo`0000ioo`800003Ool0000000<0009oo`8001moo`005Goo00<007ooOol0bWoo00D007oo
001oo`0000ioo`L000=oo`03001oogoo01ioo`005Goo00<007ooOol0bgoo0P0000=oo`00Ool03Goo
1`000goo00<007ooOol07Woo000EOol00`00Oomoo`3;Ool200000goo001oo`0=Ool400000goo0000
0003Ool00`00Oomoo`0NOol001Eoo`03001oogoo0<]oo`800003Ool007oo00eoo`@00003Ool00000
00=oo`03001oogoo01ioo`005Goo00<007ooOol0bgoo0P0000=oo`00Ool037oo1@0000=oo`000000
0goo00<007ooOol07Woo00000goo000000030003Ool30002Ool20005Ool00`00Oomoo`3<Ool00`00
Ool0000=Ool500000goo0000000TOol000=oo`03001oogoo00Aoo`04001oogoo0002Ool00`00Oomo
o`02Ool00`00Oomoo`3<Ool00`00Ool0000=Ool00`00Oomoo`02000WOol000=oo`04001oogooOol5
00001Goo001oogoo000017oo0P00cgoo00<007ooOol02goo00D007ooOomoo`0002Moo`000goo00D0
07ooOomoo`00009oo`04001oogoo0002Ool00`00Oomoo`02Ool00`00Oomoo`3>Ool00`00Oomoo`0;
Ool01@00Oomoogoo00009goo0003Ool00`00Oomoo`02Ool00`00Ool00002Ool01000Oomoo`0017oo
00<007ooOol0cWoo00<007ooOol02goo00<007ooOol0:Goo0002Ool20005Ool20003Ool20005Ool0
0`00Oomoo`3>Ool00`00Oomoo`0;Ool00`00Oomoo`0YOol001Eoo`03001oogoo0<ioo`03001oogoo
00]oo`03001oogoo02Uoo`005Goo00<007ooOol0cWoo00D007ooOomoo`0000Uoo`03001oogoo02Uo
o`005Goo00<007ooOol0cgoo00@007ooOol000Qoo`03001oogoo02Yoo`005Goo00<007ooOol0cgoo
00@007ooOol000Qoo`03001oogoo02Yoo`005Goo0P00d7oo00@007ooOol000Qoo`03001oogoo02Yo
o`005Goo00<007ooOol0cgoo00@007ooOol000Qoo`03001oogoo02Yoo`005Goo00<007ooOol0cgoo
00@007oo000000Qoo`03001oogoo02Yoo`005Goo00<007ooOol0cgoo00@007oo000000Qoo`03001o
ogoo02Yoo`005Goo00<007ooOol0cgoo00<007oo00000P001goo00<007ooOol0:Woo000EOol00`00
Oomoo`3?Ool50007Ool00`00Oomoo`0ZOol001Eoo`03001oogoo0<moo`D00003Ool007oo00Aoo`03
001oogoo02Yoo`005Goo00<007ooOol0cgoo1`001Goo00<007ooOol0:Woo000EOol2003@Ool70005
Ool00`00Oomoo`0ZOol001Eoo`03001oogoo0<moo`04001oo`00Ool30005Ool00`00Oomoo`0ZOol0
01Eoo`03001oogoo0<moo`04001oo`00Ool30005Ool00`00Oomoo`0ZOol001Eoo`03001oogoo0<mo
o`07001oo`00Ool007oo00001Goo00<007ooOol0:Woo000EOol00`00Oomoo`3?Ool01`00Ool007oo
001oo`0000Aoo`8002aoo`005Goo00<007ooOol0cgoo00<007oo00000goo00<007ooOol00Woo00<0
07ooOol0:goo000EOol00`00Oomoo`3EOol00`00Oomoo`02Ool00`00Oomoo`0[Ool001Eoo`03001o
ogoo0=Eoo`03001oogoo009oo`03001oogoo02]oo`005Goo0P00eWoo00<007ooOol00Woo00<007oo
Ool0:goo000EOol00`00Oomoo`3EOol00`00Oomoo`02Ool00`00Oomoo`0[Ool001Eoo`03001oogoo
0=Ioo`03001oogoo008002eoo`005Goo00<007ooOol0eWoo00<007ooOol00P00;Goo000EOol00`00
Oomoo`3FOol00`00Oomoo`02000]Ool001Eoo`03001oogoo0=Ioo`03001oogoo008002eoo`0000=o
o`0000000`0000Aoo`00000000Aoo`8000Eoo`03001oogoo0=Ioo`04001oogoo000^Ool000=oo`05
001oogooOol00002Ool01000Oomoo`000Woo00<007ooOol00Woo00<007ooOol0eWoo00@007ooOol0
02ioo`000goo00D007ooOomoo`00009oo`04001oogoo0002Ool00`00Oomoo`02Ool2003GOol20000
0goo001oo`0]Ool000=oo`04001oogooOol30003Ool01000Oomoo`0017oo00<007ooOol0eWoo0P00
00=oo`00Ool0;Goo0003Ool00`00Oomoo`02Ool00`00Oomoo`02Ool01000Oomoo`0017oo00<007oo
Ool0eWoo0P0000=oo`00Ool0;Goo0002Ool20004Ool30003Ool20005Ool00`00Oomoo`3FOol20000
0goo001oo`0]Ool001Eoo`03001oogoo0=Moo`03001oo`0002ioo`005Goo00<007ooOol0egoo00<0
07oo0000;Woo000EOol00`00Oomoo`3GOol2000_Ool001Eoo`03001oogoo0=Qoo`03001oogoo02eo
o`005Goo0P00ogoo2Woo000EOol00`00Oomoo`3oOol9Ool001Eoo`03001oogoo0?moo`Uoo`005Goo
00<007ooOol0ogoo2Goo003oOolQOol00?moob5oo`00ogoo8Goo003oOolQOol00?moob5oo`00ogoo
8Goo0000\
\>"],
  ImageRangeCache->{{{0, 287}, {176.938, 0}} -> {-84.0883, 59.7822, 3.94216, \
0.629769}}],

Cell[BoxData[
    TagBox[\(\[SkeletonIndicator]  Graphics  \[SkeletonIndicator]\),
      False,
      Editable->False]], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Calendar functions", "Section"],

Cell[TextData[{
  "In the financial world dates are of great importance. So are holiday \
calendars, day counting conventions and rolling conventions. QuantLib for ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " makes the calendar and date tools from QuantLib available."
}], "Text"],

Cell["What day was Christmas Day 2002?", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(DayOfWeek[{2002, 12, 25}]\)], "Input"],

Cell[BoxData[
    \(Wednesday\)], "Output"]
}, Open  ]],

Cell["Was this a business day?", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(BusinessDayQ[{2002, 12, 25}]\)], "Input"],

Cell[BoxData[
    \(False\)], "Output"]
}, Open  ]],

Cell["\<\
No, it wasn't, but that's because BusinessDayQ by default uses the \
trans-european holiday calendar TARGET.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Options[BusinessDayQ]\)], "Input"],

Cell[BoxData[
    \({HolidayCalendar \[Rule] TARGETCalendar}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?TARGETCalendar\)\)], "Input"],

Cell[BoxData[
    \("TARGETCalendar is a holiday calendar representing  the Trans-European \
Automated Real-time Gross Express-settlement  Transfer system calendar. \
http://www.ecb.int/press/00/pr001214_4.htm"\)], "Print"]
}, Open  ]],

Cell["\<\
Instead in Japan, they don't fool around having holidays because of \
Christmas.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(BusinessDayQ[{2002, 12, 25}, 
      HolidayCalendar \[Rule] TokyoCalendar]\)], "Input"],

Cell[BoxData[
    \(True\)], "Output"]
}, Open  ]],

Cell["\<\
So \"10 days later\" mean different thing, depending on, whether you think of \
calendar days, business days in London or business days in Tokyo.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Options[DaysPlus]\)], "Input"],

Cell[BoxData[
    \({HolidayCalendar \[Rule] None}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(DaysPlus[{2002, 12, 20}, 10]\)], "Input"],

Cell[BoxData[
    \({2002, 12, 30}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(DaysPlus[{2002, 12, 20}, 10, 
      HolidayCalendar \[Rule] LondonCalendar]\)], "Input"],

Cell[BoxData[
    \({2003, 1, 8}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(DaysPlus[{2002, 12, 20}, 10, 
      HolidayCalendar \[Rule] TokyoCalendar]\)], "Input"],

Cell[BoxData[
    \({2003, 1, 10}\)], "Output"]
}, Open  ]],

Cell["\<\
When counting days different day count conventions can be used.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \($DayCountConventions\)], "Input"],

Cell[BoxData[
    \({Actual365, Actual360, Thirty360American, Thirty360European, 
      Thirty360Italian, ActualActualISMA, ActualActualAFB, 
      ActualActualISDA}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Options[DaysBetween]\)], "Input"],

Cell[BoxData[
    \({DayCountConvention \[Rule] Actual365}\)], "Output"]
}, Open  ]],

Cell["\<\
So when counting days between two days, it depends on the day counting \
convention:\
\>", "Text"],

Cell[BoxData[
    \(date1 = {2002, 1, 1}; date2 = {2003, 1, 2};\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(DaysBetween[date1, date2]\)], "Input"],

Cell[BoxData[
    \(366\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(DaysBetween[date1, date2, 
      DayCountConvention \[Rule] Thirty360European]\)], "Input"],

Cell[BoxData[
    \(361\)], "Output"]
}, Open  ]],

Cell["\<\
This also affects calculations of the fraction of years between the dates:\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(YearsBetween[date1, date2]\)], "Input"],

Cell[BoxData[
    \(1.0027397260273974`\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(YearsBetween[date1, date2, 
      DayCountConvention \[Rule] Thirty360European]\)], "Input"],

Cell[BoxData[
    \(1.0027777777777778`\)], "Output"]
}, Open  ]],

Cell["\<\
Rolling conventions decides in which direction to change (\"roll\") e.g. a \
settlement date, when it falls on a holiday.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(HolidayQ[{{2003, 1, 1}, {2003, 1, 2}}, 
      HolidayCalendar \[Rule] TokyoCalendar]\)], "Input"],

Cell[BoxData[
    \({True, True}\)], "Output"]
}, Open  ]],

Cell["\<\
In Japan New Years Day is a holiday and January 2nd is also. So if a \
settlement day falls on this date is must be rolled. The default is \
Following.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?Following\)\)], "Input"],

Cell[BoxData[
    \("Following is a rolling convention, which adjust to the first business \
day  after the given holiday."\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(RollDate[{2003, 1, 1}, 
      HolidayCalendar \[Rule] TokyoCalendar]\)], "Input"],

Cell[BoxData[
    \({2003, 1, 6}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(RollDate[{2003, 1, 1}, HolidayCalendar \[Rule] TokyoCalendar, 
      RollingConvention \[Rule] Preceding]\)], "Input"],

Cell[BoxData[
    \({2002, 12, 30}\)], "Output"]
}, Open  ]],

Cell["Also you can check if a certain year is a leap year.", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(LeapYearQ[{1800, 1804, 2000, 2002}]\)], "Input"],

Cell[BoxData[
    \({True, True, True, False}\)], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Error handling", "Section"],

Cell[TextData[{
  "At this stage there is not much error handling in QuantLib for ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  ", but if something goes wrong in the C++ part, it is reported."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(DayOfWeek[{2002, 2, 31}]\)], "Input"],

Cell[BoxData[
    \(QuantLibMma::"error" \(\(:\)\(\ \)\) 
      "\!\(\"Date::Date : day outside month (2) day-range Date::Date : month \
2[1,28]\"\)"\)], "Message"],

Cell[BoxData[
    \($Failed\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Value[EuropeanOption[110, 1, CallOption], 
      BlackScholesModel[100, 0.1, 0.01, \(-0.15\)]]\)], "Input"],

Cell[BoxData[
    \(QuantLibMma::"error" \(\(:\)\(\ \)\) 
      "\!\(\"negative or zero volatility given\"\)"\)], "Message"],

Cell[BoxData[
    \(ReplaceAll::"reps" \(\(:\)\(\ \)\) 
      "\!\({$Failed}\) is neither a list of replacement rules nor a valid \
dispatch table, and so cannot be used for replacing."\)], "Message"],

Cell[BoxData[
    \(Value /. \[InvisibleSpace]$Failed\)], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Stopping the external program", "Section"],

Cell[TextData[{
  "As we now have stopped using the package, you can stop the external \
program, if you wish. Though, it will be stopped automatically anyway, when \
you exit ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " (or quit the kernel)."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Uninstall[QuantLib]\)], "Input"],

Cell[BoxData[
    \("C:\\Program Files\\Microsoft Visual \
Studio\\MyProjects\\QuantLibMma\\Debug\\QuantLibMma.exe"\)], "Output"]
}, Open  ]]
}, Open  ]]
},
FrontEndVersion->"4.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 670}},
WindowSize->{1016, 643},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
CellLabelAutoDelete->True
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 99, 2, 57, "Input"],
Cell[1819, 53, 364, 11, 71, "Text"],

Cell[CellGroupData[{
Cell[2208, 68, 31, 0, 53, "Section"],
Cell[2242, 70, 265, 5, 52, "Text"],
Cell[2510, 77, 491, 10, 52, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[3038, 92, 38, 0, 53, "Section"],
Cell[3079, 94, 206, 6, 33, "Text"],
Cell[3288, 102, 45, 1, 30, "Input"],

Cell[CellGroupData[{
Cell[3358, 107, 46, 1, 30, "Input"],
Cell[3407, 110, 166, 2, 25, "Print"]
}, Open  ]],
Cell[3588, 115, 75, 0, 33, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[3700, 120, 61, 0, 53, "Section"],
Cell[3764, 122, 660, 15, 71, "Text"],
Cell[4427, 139, 742, 17, 71, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[5206, 161, 34, 0, 53, "Section"],
Cell[5243, 163, 89, 2, 33, "Text"],

Cell[CellGroupData[{
Cell[5357, 169, 52, 1, 30, "Input"],
Cell[5412, 172, 216, 3, 44, "Print"]
}, Open  ]],
Cell[5643, 178, 83, 1, 30, "Input"],
Cell[5729, 181, 73, 0, 33, "Text"],

Cell[CellGroupData[{
Cell[5827, 185, 55, 1, 30, "Input"],
Cell[5885, 188, 244, 3, 44, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[6166, 196, 81, 1, 30, "Input"],
Cell[6250, 199, 76, 1, 29, "Output"]
}, Open  ]],
Cell[6341, 203, 68, 0, 33, "Text"],

Cell[CellGroupData[{
Cell[6434, 207, 57, 1, 30, "Input"],
Cell[6494, 210, 319, 5, 29, "Output"]
}, Open  ]],
Cell[6828, 218, 52, 0, 33, "Text"],

Cell[CellGroupData[{
Cell[6905, 222, 52, 1, 30, "Input"],
Cell[6960, 225, 53, 1, 29, "Output"]
}, Open  ]],
Cell[7028, 229, 159, 5, 33, "Text"],

Cell[CellGroupData[{
Cell[7212, 238, 70, 1, 30, "Input"],
Cell[7285, 241, 54, 1, 29, "Output"]
}, Open  ]],
Cell[7354, 245, 105, 3, 33, "Text"],
Cell[7462, 250, 106, 2, 30, "Input"],

Cell[CellGroupData[{
Cell[7593, 256, 62, 1, 30, "Input"],
Cell[7658, 259, 12219, 373, 186, 3765, 264, "GraphicsData", "PostScript", \
"Graphics"],
Cell[19880, 634, 130, 3, 29, "Output"]
}, Open  ]],
Cell[20025, 640, 74, 0, 33, "Text"],
Cell[20102, 642, 147, 3, 30, "Input"],
Cell[20252, 647, 35, 0, 33, "Text"],

Cell[CellGroupData[{
Cell[20312, 651, 81, 1, 30, "Input"],
Cell[20396, 654, 67800, 1360, 242, 21047, 778, "GraphicsData", "PostScript", \
"Graphics"],
Cell[88199, 2016, 137, 3, 29, "Output"]
}, Open  ]],
Cell[88351, 2022, 38, 0, 33, "Text"],
Cell[88392, 2024, 167, 3, 30, "Input"],

Cell[CellGroupData[{
Cell[88584, 2031, 171, 3, 30, "Input"],
Cell[88758, 2036, 118719, 2772, 242, 76516, 2246, "GraphicsData", \
"PostScript", "Graphics"],
Cell[207480, 4810, 137, 3, 29, "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[207666, 4819, 41, 0, 53, "Section"],
Cell[207710, 4821, 50, 0, 33, "Text"],

Cell[CellGroupData[{
Cell[207785, 4825, 38, 1, 30, "Input"],
Cell[207826, 4828, 76, 1, 29, "Output"]
}, Open  ]],
Cell[207917, 4832, 98, 2, 33, "Text"],

Cell[CellGroupData[{
Cell[208040, 4838, 52, 1, 30, "Input"],
Cell[208095, 4841, 152, 2, 25, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[208284, 4848, 106, 2, 30, "Input"],
Cell[208393, 4852, 34411, 1418, 186, 18613, 1218, "GraphicsData", \
"PostScript", "Graphics"],
Cell[242807, 6272, 130, 3, 29, "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[242986, 6281, 37, 0, 53, "Section"],
Cell[243026, 6283, 294, 6, 52, "Text"],
Cell[243323, 6291, 48, 0, 33, "Text"],

Cell[CellGroupData[{
Cell[243396, 6295, 58, 1, 30, "Input"],
Cell[243457, 6298, 43, 1, 29, "Output"]
}, Open  ]],
Cell[243515, 6302, 40, 0, 33, "Text"],

Cell[CellGroupData[{
Cell[243580, 6306, 61, 1, 30, "Input"],
Cell[243644, 6309, 39, 1, 29, "Output"]
}, Open  ]],
Cell[243698, 6313, 132, 3, 33, "Text"],

Cell[CellGroupData[{
Cell[243855, 6320, 54, 1, 30, "Input"],
Cell[243912, 6323, 74, 1, 29, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[244023, 6329, 52, 1, 30, "Input"],
Cell[244078, 6332, 223, 3, 44, "Print"]
}, Open  ]],
Cell[244316, 6338, 104, 3, 33, "Text"],

Cell[CellGroupData[{
Cell[244445, 6345, 107, 2, 30, "Input"],
Cell[244555, 6349, 38, 1, 29, "Output"]
}, Open  ]],
Cell[244608, 6353, 169, 3, 33, "Text"],

Cell[CellGroupData[{
Cell[244802, 6360, 50, 1, 30, "Input"],
Cell[244855, 6363, 64, 1, 29, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[244956, 6369, 61, 1, 30, "Input"],
Cell[245020, 6372, 48, 1, 29, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[245105, 6378, 108, 2, 30, "Input"],
Cell[245216, 6382, 46, 1, 29, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[245299, 6388, 107, 2, 30, "Input"],
Cell[245409, 6392, 47, 1, 29, "Output"]
}, Open  ]],
Cell[245471, 6396, 87, 2, 33, "Text"],

Cell[CellGroupData[{
Cell[245583, 6402, 53, 1, 30, "Input"],
Cell[245639, 6405, 179, 3, 29, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[245855, 6413, 53, 1, 30, "Input"],
Cell[245911, 6416, 72, 1, 29, "Output"]
}, Open  ]],
Cell[245998, 6420, 108, 3, 33, "Text"],
Cell[246109, 6425, 76, 1, 30, "Input"],

Cell[CellGroupData[{
Cell[246210, 6430, 58, 1, 30, "Input"],
Cell[246271, 6433, 37, 1, 29, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[246345, 6439, 111, 2, 30, "Input"],
Cell[246459, 6443, 37, 1, 29, "Output"]
}, Open  ]],
Cell[246511, 6447, 98, 2, 33, "Text"],

Cell[CellGroupData[{
Cell[246634, 6453, 59, 1, 30, "Input"],
Cell[246696, 6456, 53, 1, 29, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[246786, 6462, 112, 2, 30, "Input"],
Cell[246901, 6466, 53, 1, 29, "Output"]
}, Open  ]],
Cell[246969, 6470, 145, 3, 33, "Text"],

Cell[CellGroupData[{
Cell[247139, 6477, 117, 2, 30, "Input"],
Cell[247259, 6481, 46, 1, 29, "Output"]
}, Open  ]],
Cell[247320, 6485, 175, 4, 33, "Text"],

Cell[CellGroupData[{
Cell[247520, 6493, 47, 1, 30, "Input"],
Cell[247570, 6496, 136, 2, 25, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[247743, 6503, 101, 2, 30, "Input"],
Cell[247847, 6507, 46, 1, 29, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[247930, 6513, 138, 2, 30, "Input"],
Cell[248071, 6517, 48, 1, 29, "Output"]
}, Open  ]],
Cell[248134, 6521, 68, 0, 33, "Text"],

Cell[CellGroupData[{
Cell[248227, 6525, 68, 1, 30, "Input"],
Cell[248298, 6528, 59, 1, 29, "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[248406, 6535, 33, 0, 53, "Section"],
Cell[248442, 6537, 215, 5, 33, "Text"],

Cell[CellGroupData[{
Cell[248682, 6546, 57, 1, 30, "Input"],
Cell[248742, 6549, 164, 3, 24, "Message"],
Cell[248909, 6554, 41, 1, 29, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[248987, 6560, 127, 2, 30, "Input"],
Cell[249117, 6564, 124, 2, 24, "Message"],
Cell[249244, 6568, 200, 3, 24, "Message"],
Cell[249447, 6573, 67, 1, 29, "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[249563, 6580, 48, 0, 53, "Section"],
Cell[249614, 6582, 269, 7, 33, "Text"],

Cell[CellGroupData[{
Cell[249908, 6593, 52, 1, 30, "Input"],
Cell[249963, 6596, 129, 2, 29, "Output"]
}, Open  ]]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

